% function [sig] = homog(t, mua, musp, v, R [,bcs] )
%
% analytic homogeneous solution in 3d, semi-infinite, diffusion
% approx. (time units are ps, length are mm).
%
% t = time, can be a vector of times
% mua, musp = scalar values absorption, reduced scattering coefficients
% v = speed of light in medium
% R = radius (S-D distance on surface), can be a vector of radii
%
% bcs = boundary condition and source/detector choices:
%        0 : zero-boundary, source 1/musp inside, kappa.normal deriv det
%        1 : extrap BC, source 1/musp inside, fluence value det
%            mult by kappa/zb to be asymp. same size as 2. (Haskell)
%        2 : extrap BC, source 1/musp inside, kappa.normal deriv det.
%            (Zaccanti)
%        3 : extrap BC, source 1/musp inside, Kienle97 n=1.4 flux combo det.
%            (not implemented yet)
%
% Signals for all pairs t,R are returned in sig: sig(i,j) = time i,
% radius j.
%
% Units of fluence are 1/volume. (differs from Kienle97 by factor
% of c, and Moulton90 by c.Energy)
%
% Note that signals for bcs 0 and 1,2,3 do not agree asymptotically: there
% is an overall factor of (z0+zb)/z0 by which 1,2,3 are larger, due
% to the larger dipole moment, which results from EBC. The
% amplitudes of 1,2,3 have been adjusted to be asymptotically the same.
%
%
% Example:
% s = homog([100 1000], 0.01, 1.0, 0.2, [0 30]);
%
% 3/18/03 improved speed a bit (for .^(3/2) etc), added BCs options 4/18/03
%
% Also see: SLAB

function [sig] = homog(t, mua, musp, v, R, bcs)

if nargin==5
  bcs = 0;      % default
end

c = 0.2997;     % units mm, ps.

z0 = 1/musp;
kap = 1/(3*musp);   % no mua contrib: Hielscher tests as good choice.
lR = length(R);
lt = length(t);

Rmat = kron(reshape(R,[1 lR]), ones(lt ,1));
t = reshape(t,[lt 1]);
tmat = kron(t, ones(1, lR));


switch bcs

 case 0     % dirichlet (zero) BCs, source 1/musp inside, detect
            % normal grad of fluence, mult by kappa.
  prefac = (4*pi*kap)^(3/2)*(v*t).^(5/2);
  sig = z0*exp(-(z0*z0 + Rmat.*Rmat)./(4*kap*v*tmat) - v*mua*tmat) ./...
	kron(prefac, ones(1, lR));

 case 1     % extrap. (partial-current, or `radiation') BCs,
            % source 1/musp inside, detect fluence value, mult by kappa/(z0+zb)
	    % (Haskell's R_{eff}=0.493 for n=1.4 gives z_b = 1.96/musp)...
  
  if abs(c/v - 1.4)<0.2       % tissue n around 1.4
    zb = 1.96/musp;
  elseif abs(c/v - 1.0)<0.1   % air n around 1
    zb = 0.7/musp; % 0.7/musp;
  end
  rmatis = z0*z0 + Rmat.*Rmat;
  rmatos = (2*zb + z0)*(2*zb + z0) + Rmat.*Rmat;
  decay = -v*mua*tmat;
  prefac = (4*pi*kap*v*t).^(3/2);
  sig = (exp(-rmatis./(4*kap*v*tmat) + decay) - ...
	 exp(-rmatos./(4*kap*v*tmat) + decay)) ./ kron(prefac, ones(1, lR));
  % overall amplitude scale to match R->inf limit of case 2...
  sig = sig * kap/zb;
  
 case 2     % extrap. (partial-current, or `radiation') BCs,
            % source 1/musp inside, detect normal deriv mult by kappa.
	    % (Haskell's R_{eff}=0.493 for n=1.4 gives z_b = 1.96/musp)...
  
  if abs(c/v - 1.4)<0.1       % tissue n around 1.4
    zb = 1.96/musp;
  elseif abs(c/v - 1.0)<0.1   % air n around 1
    zb = 0.7/musp; % 0.7/musp;
  end
  rmatis = z0*z0 + Rmat.*Rmat;
  rmatos = (2*zb + z0)*(2*zb + z0) + Rmat.*Rmat;
  decay = -v*mua*tmat;
  prefac = 2 * (4*pi*kap)^(3/2)*(v*t).^(5/2);
  sig = (z0*exp(-rmatis./(4*kap*v*tmat) + decay) + ...
	 (z0+2*zb)*exp(-rmatos./(4*kap*v*tmat) + decay)) ...
	./ kron(prefac, ones(1, lR));
  % overall amplitude scale to match R->inf limit of dirichlet...
  sig = sig;


  
end

%end
