/* slab main: example C calling of slab.c, avoiding matlab
 *
 * reads a simple text file format and writes signal data to text file.
 *
 * usage: slab infile outfile
 *
 * input file format (see slab_in.dat, and fscanf commands in code):
 *
 * tmin tmax dtgate N_radii {R1 R2 ...} N_layers {musp1 musp2 ...}
 * {mua1 mua2 ...} {dz1 dz2 ...} refractive_index N_err_params_to_read
 * [err1 err2 ... ]
 *
 * NOTE: the dz input list is 1 shorter than N_layers since last layer infinite
 * See slab.m for full documentation.
 * 
 * output file format: ntgate N_R signal_array
 * The signal_array is a matrix, rows are dt values. Change the output code
 * below to suit yourself.
 *
 * barnett 4/24/03 finally fixed up 5/19/06
 */

#include <stdio.h>

#define MAX_LAYERS 100      /* should match those in slab.h */
#define MAX_RS     100
#define MAX_SAMPS 10000

#define N_ERR    13          /* max length of err input array */

extern int verb;                      /* global verbosity */
const extern double DEFAULT_ERR[];    /* default error params */


int main(int argc, char **argv)
{
  FILE *fp;
  int N_R, N_l, i, ntgate, j, n_errread;
  double n, R[MAX_RS], mua[MAX_LAYERS], musp[MAX_LAYERS], dz[MAX_LAYERS-1];
  double err[N_ERR], sig[MAX_LAYERS*MAX_SAMPS], tmin, tmax, dtgate;
  char *in, *out;

  if (argc!=3) {
    printf("SLAB C driver, Barnett 2006.      usage: slab infile outfile\n");
    return 1;
  }
  in = argv[1];
  out = argv[2];
  /*printf("in: %s   out: %s\n", in, out);*/

  /* read... */
  fp = fopen(in, "r");
  
  fscanf(fp, "%lf %lf %lf %d", &tmin, &tmax, &dtgate, &N_R);
  for (i=0;i<N_R;++i)
    fscanf(fp, "%lf", R+i);
  fscanf(fp, "%d", &N_l);
  for (i=0;i<N_l;++i)
    fscanf(fp, "%lf", musp+i);
  for (i=0;i<N_l;++i)
    fscanf(fp, "%lf", mua+i);
  if (N_l>1)
    for (i=0;i<N_l-1;++i)
      fscanf(fp, "%lf", dz+i);
  fscanf(fp, "%lf", &n);
  fscanf(fp, "%d", &n_errread);  /* find out how many err params to read */
  for (i=0; i<N_ERR; ++i)
    err[i] = DEFAULT_ERR[i];
  for (i=0; i<n_errread; ++i)    /* read them */
    fscanf(fp, "%lf", err+i);
  fclose(fp);

  /* set global verbosity */
  verb = (int)(0.5 + err[6]);

  ntgate = (int)((tmax-tmin)/dtgate);
  if (verb & 0x01) {
    printf("slab.c v1.5: %d radii, %d layers, %d err params\n", N_R, N_l, n_errread);
    printf("\tntgate = %d, N_R = %d\n", ntgate, N_R);
  }
  if (verb & 0x02)
    for (j=0; j<N_R; ++j)
      printf("\tradius R[%d] = %g\n", j, R[j]);


  /* Do calc.... fills sig array with answer */
  calculate(tmin, dtgate, ntgate, R, N_R, N_l, musp, mua, dz,
	    (dz[N_l-1]==0.0), n, err, sig);

  if (verb & 0x01)
    printf("calculate done, writing to %s\n", out);
  
 /* write output: note sig is stored as a 1D array */
  fp = fopen(out, "w");
  fprintf(fp, "%d %d\n", ntgate, N_R);  /* modify this to suit your needs */
  for (i=0;i<ntgate;++i) {
    for (j=0;j<N_R;++j)
      fprintf(fp, "%.15g \t", sig[j*ntgate + i]);
    fprintf(fp, "\n");
  }
  close(fp);

  return 0;
}
