/* slab.c : multilayer slab fast diffusion forward model, called from MATLAB.
 *
 * see slab.m for documentation and version history.
 *
 * alex barnett, started 11/18/02
 * v1.5 08/01/03
 * v1.6 5/22/05: changed to pointers to structs passed to funcs, struct malloc
 */

#include "slab.h"

/*--------------------------------------------------------------------------*/
int build_slab(struct Multilayer_slab *sys, int N_l, double *musp, double *mua,
		double *dz, int semi_inf, double refr, double *err)
     /* sums up depths, converts to speed-scaled properties, reports
      *
      * barnett 11/20/02, added EBC 4/19/03
      */ 
{
  int N_i, i, bcs;

  /* rescale props to c/n (speed in medium) units, calc kappa = 1/(3.musp) */
  for (i=0; i<N_l; ++i) {
    sys->cmua[i] = LIGHTSPEED*mua[i]/refr;
    sys->ckap[i] = LIGHTSPEED/(3*musp[i]*refr);
  }
  sys->kap_top = 1/(3*musp[0]);

  /* boundary condition and det type choose... */
  bcs = (int)(err[8] + 0.5);    /* nearest integer */

  /* release depth (could depend on BCS) */
  sys->z0 = 1/musp[0];
  
  /* worst-case diffusivity ckap_max, range... */
  sys->ckap_max = 0.0;
  sys->ckap_min = 1e12; /* acts as infinity */
  sys->cmua_min = 1e12;
  for (i=0; i<N_l; ++i) {
    sys->ckap_max = max(sys->ckap_max, sys->ckap[i]);
    sys->ckap_min = min(sys->ckap_min, sys->ckap[i]);
    sys->cmua_min = min(sys->cmua_min, sys->cmua[i]);
  }

  sys->N_l = N_l;
  sys->semi_inf = semi_inf;
  N_i = N_l - semi_inf;         /* number of interfaces (excluding z=0) */
  if (verb & 0x01)
    mexPrintf("\n\tbuild_slab: N_l=%d, semi_inf=%d, z0=%g\n", N_l, semi_inf,
	   sys->z0);

  /* if at least one interface, find depths */
  if (N_i>0) {
    /* check valid */
    for (i=0; i<N_i; ++i)
      if (dz[i]<0.0) {
	mexPrintf("slab.c build_slab: dz[%d] = %g < 0\n", i, dz[i]);
	return 1;
      }
    /* sum up depths */
    sys->z[0] = dz[0];
    for (i=1; i<N_i; ++i) {
      sys->z[i] = sys->z[i-1] + dz[i];
    }
    if (verb & 0x01) {
      /* report */
      mexPrintf("\tlayer\tfrom z\t\tto z\t\tcmua\t\tckap\n");
      mexPrintf("\t0\t0\t\t%-16.4g%-16.4g%-16.4g\n",
		sys->z[0], sys->cmua[0], sys->ckap[0]);
      if (N_i>=1)
	for (i=1;i<N_i;++i)
	  mexPrintf("\t%d\t%-16.4g%-16.4g%-16.4g%-16.4g\n", i, sys->z[i-1],
		 sys->z[i], sys->cmua[i], sys->ckap[i]);
      /* final layer... can be semi_inf */
      if (semi_inf)
	mexPrintf("\t%d\t%-16.4ginf\t\t%-16.4g%-16.4g\n", N_l-1, sys->z[N_i-1],
	       sys->cmua[N_l-1], sys->ckap[N_l-1]);
    }
  } else
    if (verb & 0x01) {
      /* report. N_i=0 means z array never used */
      mexPrintf("\tlayer\tfrom z\t\tto z\t\tcmua\t\tckap\n");
	mexPrintf("\t0\t0\t\tinf\t\t%-16.4g%-16.4g\n",     sys->cmua[0],
	       sys->ckap[0]);
    }

  /* boundary condition and det type choose... */
  if (bcs>=ZBC && bcs<N_BCS) {
    sys->bcs = bcs;
    if (verb & 0x01)
      mexPrintf("\tBCs = %s\n", bcs_names[bcs]);
    /* choose z_b... */
    if (fabs(refr-1.0)<0.1)
      sys->zb = 0.71 / musp[0]; /* Haskell94 */
    else if (fabs(refr-1.4)<0.1)
      sys->zb = 1.96 / musp[0]; /* Haskell94 */
    else {
      mexPrintf("\n\tbuild_slab: z_b unknown for refr=%f!\n", refr);
      sys->zb = 0.0;
    }
    /* set zb=0 if ZBC */
    if (bcs==ZBC)
      sys->zb = 0.0;
    if ((bcs!=ZBC) && (verb & 0x01))
      mexPrintf("\t\tEBC has zb=%f\n", sys->zb);
  } else
    mexErrMsgTxt("slab.c build_slab: invalid BCs, err[8]!"); 
  
  if (err[11]>0.0) {
    /* Override the minimum mu_s' found, using err(12) input param... */
    if (verb & 0x01)
      mexPrintf("\n\t\tckap_max=%g overridden by err(12) to %g", sys->ckap_max,
		LIGHTSPEED/(3*err[11]*refr));
    sys->ckap_max = LIGHTSPEED/(3*err[11]*refr);
  }
  if (err[12]>0.0) {
    /* Override the maximum mu_s' found, using err(13) input param... */
    if (verb & 0x01)
      mexPrintf("\n\t\tckap_min=%g overridden by err(13) to %g", sys->ckap_min,
		LIGHTSPEED/(3*err[12]*refr));
    sys->ckap_min = LIGHTSPEED/(3*err[12]*refr);
  }

  return 0;
}


/* macros for z-domain of a given layer rel to -zb... includes zb for EBCs
   Changed to s being pointer-to-struct, barnett 5/22/05 */
#define Z_MIN(s, l) \
((l==0) ? 0.0 : s->z[l-1]+s->zb)
/* the following is a hack to work for semi-inf: can't be too large
   or get integer overflow in init_FD... */
#define Z_MAX(s, l) \
((s->semi_inf && l==s->N_l-1) ? LARGE_DEPTH : s->z[l]+s->zb*(l==s->N_l-1 ? 2:1))


/*--------------------------------------------------------------------------*/
int init_FD(struct Onedim_FD_system *o, struct Multilayer_slab *sys,
	    struct SingleSD *meas, double *err)
     /* Chooses 1d FD lattice h, size, material properties.
      *
      * z range discretized (always Dirichlet BCs) is:
      *          -sys.zb to zbox                        (semi-inf)
      *          -sys.zb to sys.z[sys.N_l - 1] + sys.zb (finite thickness)
      *
      * first cell goes from -sys.zb to -sys.zb + h.   Etc.
      * nz is number of cells not number of interior nodes.
      *
      * Ideas:
      * could return vector of derivative of cell props for each model param.
      *
      * 4/19/03 Extrap BCs added (zb).
      */
{
  double frac, ckap_max, z_box, err_zbox, sigma_tmin, sigma_tmax, thick;
  int i, l, i_min, i_max, fall_off;
 
  if (verb & 0x01)
    mexPrintf("\n\tinit_FD:\n");

  err_zbox = err[0]; /* how to extract FD-related errs from passed-in */
  o->err_dt = err[2]; /* just pass straight to onedim_FD, even if <0 */
  /* CONST_VAR_CN override inputs... */
  o->t0 = err[9];
  o->dt0 = err[10];

  /* Decide 1d lattice constant... */
  if (err[1]>0.0) {
    /* calc h, using diffusivity of top layer only (sigma_tmin < dz[0] ?) */
    sigma_tmin = sqrt(2 * sys->ckap[0] * meas->tmin);
    o->h = sqrt(err[1] / 2.3) * sigma_tmin; /* empirical factor, fwd Euler */
    if (verb & 0x01)
      mexPrintf("\t\tz-lattice: err_h=%g gives h=%g\n", err[1], o->h);
  } else {
    o->h = -err[1];
    if (verb & 0x01)
      mexPrintf("\t\tz-lattice: user chose h=%g\n", o->h);
  }

  /* Decide system depth, use to tweak h if depth has strict value... */
  if (sys->semi_inf) {
    /* semi-infinite in z direction: calc z_box (called L in sla)... */
    z_box = sqrt(sys->ckap_max * meas->tmax *
		 log((2+4*log(1/err_zbox))/err_zbox));

    /* round up in units of h... lower model jitter than rounding down h ? */
    o->nz = (int)(1.0 + (z_box + sys->zb) / o->h);
    z_box = -sys->zb + o->nz * o->h;
    if (verb & 0x01)
      mexPrintf("\t\tz-box:   err_zbox=%g gives z_box(L)=%g\n",
		err_zbox, z_box);
    
  } else {
    /* finite thick: adjust h downwards so bottom surface on a node... */
    thick = sys->z[sys->N_l - 1]+2*sys->zb;
    o->nz = (int)(1.0 + thick / o->h);
    o->h = thick / o->nz;
    if (verb & 0x01)
      mexPrintf("\t\t           bottom interface tweaks, h=%g\n", o->h);
  }
  /* report, check doesn't exceed lattice allocation */
  if (verb & 0x01)
    mexPrintf("\t\tnz=%d\n", o->nz);
  if (o->nz > MAX_ZNODES)
    mexErrMsgTxt("slab.c init_FD: nz exceeds limit!"); 


  /* Build cells: kap_h, mua_h... would be easy to find derivs too */
  /* In this section, z is meas relative to lowest z (=-zb)... */
  for (i=0; i<o->nz; ++i) {
    o->mua_h[i] = 0.0;  /* zero them first */
    o->kap_h[i] = 0.0;   /* this will collect c.kappa */
    o->kapi_h[i] = 0.0;   /* this will collect INVERSE of c.kappa */
  }
  for (l=0; l<sys->N_l; ++l) {
    /* add cell contribs due to each layer : note that kap_h is inverse */
    i_min = (int)(Z_MIN(sys, l)/o->h);
    i_max = (int)(Z_MAX(sys, l)/o->h);
    if (i_max>=o->nz) {
      i_max = o->nz;
      fall_off = 1;
    } else
      fall_off = 0;
    /* mexPrintf("layer %d: i_min=%d, i_max=%d\n", l, i_min, i_max); */
    if (i_max==i_min) {
      /* whole layer within 1 cell */
      frac = (Z_MAX(sys, l) - Z_MIN(sys, l)) / o->h;
      o->mua_h[i_min] += frac * sys->cmua[l];
      o->kap_h[i_min] += frac * sys->ckap[l];
      o->kapi_h[i_min] += frac / sys->ckap[l];  /* inverse */
    } else {
      /* bottom cell frac */
      frac = i_min + 1 - Z_MIN(sys, l)/o->h;
      o->mua_h[i_min] += frac * sys->cmua[l];
      o->kap_h[i_min] += frac * sys->ckap[l];
      o->kapi_h[i_min] += frac / sys->ckap[l];   /* inverse */
      /* intermediate full cells (may be zero of them) */
      for (i=i_min+1; i<i_max; ++i) {
	o->mua_h[i] += sys->cmua[l];
	o->kap_h[i] += sys->ckap[l];
	o->kapi_h[i] += 1.0 / sys->ckap[l];  /* inverse */
      }
      if (!fall_off) {
	/* top cell frac */
	frac = Z_MAX(sys, l)/o->h - i_max;
	o->mua_h[i_max] += frac * sys->cmua[l];
	o->kap_h[i_max] += frac * sys->ckap[l];
	o->kapi_h[i_max] += frac / sys->ckap[l];  /* inverse */
      }
    }
  }
  for (i=0; i<o->nz; ++i)
    o->kapi_h[i] = 1.0 / o->kapi_h[i];   /* re-invert inv c.kappa */


  if (verb & 0x20) {
    mexPrintf("\tcell properties:\n");
    for (i=0; i<o->nz; ++i)
      mexPrintf("\ti=%d : z[i]=%f,\tmua=%f,\tkap=%f,\tkapi=%f\n", i,
		-sys->zb+i*o->h, o->mua_h[i], o->kap_h[i], o->kapi_h[i]);
  }

  /* decide on optical mode freq to kill via first few euler steps, currently
     it's a Courant # of 1/2 given the top layer diffusivity (crude) */
  o->dt_euler_kill_hf = o->h * o->h / (4*sys->ckap[0]);
  /* Only used for 1d; For 3D, this gets overwritten later. */
  /* this must be the same for all modes m, for proper cancellation */

  /* decide default mu_shift. (This may be modified in 3D case, once
     k_0 known, for CNvar. Also in 2D case it defines whole mode decay). */
  o->mu_shift = sys->cmua_min;

  return 0;
}



/*--------------------------------------------------------------------------*/
void swap(double **a, double **b)
     /* swaps pointers to doubles. Usage : swap(&f, &g); */
{
  double *ptr = *a;
  *a = *b;
  *b = ptr;
}


/*--------------------------------------------------------------------------*/
void f_euler(const struct Onedim_FD_system *o, double *f, double *g, double dt)
     /* single step of forward Euler, takes f -> g.
      * o->nz is # interior cells, not nodes.
      * Note, uses mue_h as effective absorption in each cell.
      *
      * 7/12/03: changed to use kapi_h which is cell-inv-averaged c.kappa
      */
{
  int i;
  double dt_h2 = dt/(o->h*o->h);

  /* interior... */
  for (i=1; i<o->nz; ++i)
    g[i] = dt_h2*(o->kapi_h[i-1]*f[i-1] + o->kapi_h[i]*f[i+1]) +
      (1.0 - (dt*(o->mue_h[i-1]+o->mue_h[i])/2 +
	      dt_h2*(o->kapi_h[i-1]+o->kapi_h[i])))*f[i];
  /* endpoints... simple enforce Dirichlet */
  g[0] = 0.0;
  g[o->nz] = 0.0;
}


/*--------------------------------------------------------------------------*/
int build_LU(const struct Onedim_FD_system *o, double dt,
	     double *D, double *E)
     /* builds LDL^T decomp of symm +ve tridiag for single implicit step of dt,
      * filling D, E.
      * See f_euler for info on matrix used. Endpoints are included as dofs.
      * (they are uncoupled to body of matrix)
      *
      * 7/12/03: changed to use kapi_h which is cell-inv-averaged c.kappa
      * 8/1/03: bug fixed for final E element.
      */
{
  int i, info, n = o->nz + 1;           /* matrix order */
  double dt_h2 = dt/(o->h*o->h);
  
  /* fill tridiag... interior */
  for (i=1; i<o->nz; ++i)
    D[i] = 1.0 + dt*(o->mue_h[i-1]+o->mue_h[i])/2 +
		   dt_h2*(o->kapi_h[i-1]+o->kapi_h[i]);
  for (i=1; i<o->nz-1; ++i)
    E[i] = -dt_h2*o->kapi_h[i];
  
  /* endpoints... */
  D[0] = 1.0;
  E[0] = 0.0;
  E[o->nz-1] = 0.0;
  D[o->nz] = 1.0;

  /* prefactorize tridiag... */
  dpttrf_(&n, D, E, &info);

  if (info!=0)
    mexPrintf("build_LU: problem in dpttrf, info=%d\n",info);

  return info;
}


/*--------------------------------------------------------------------------*/
int mu_shift(struct Onedim_FD_system *o, double k)
/* sets mu_shift in a 1d system struct. k = lowest transverse wavenumber.
 *
 * barnett 4/25/03
 *
 * 7/12/03: changed to use kapi_h which is cell-inv-averaged c.kappa
 */
{
  int i;
  double mue_i, D[MAX_ZNODES], E[MAX_ZNODES], W[MAX_ZNODES];
  double work[5*MAX_ZNODES];
  double ih2 = 1/(o->h*o->h), dummy = 0.0, abstol = 1e-9; /* eig tolerance */
  char jobz = 'N', range = 'I';
  int ldz = o->nz - 1, found;    /* z not referenced anyway */
  int info, n = o->nz - 1, iwork[MAX_ZNODES], idummy, il = 1, ih = 1;

  /* fill matrix -A as tridiag... NOTE: missing end (dirichlet) dofs! */
  for (i=0; i<o->nz-1; ++i)
    D[i] = (o->mua_h[i] + o->mua_h[i+1] + (2*ih2 + k*k) *
      (o->kapi_h[i]+o->kapi_h[i+1]))/2;
  for (i=0; i<o->nz-2; ++i)
    E[i] = -ih2 * o->kapi_h[i+1];
  
  /* find lowest eigenvalue of -A... eigenvalue range i=il to ih */
  dstevx_(&jobz, &range, &n, D, E, &dummy, &dummy, &il, &ih, &abstol, &found,
	  W, &dummy, &ldz, work, iwork, &idummy, &info);
  if (found!=(ih-il+1) || info!=0) {
    mexPrintf("mu_shift: dstevx info=%d, found=%d !", info, found);
    return 1;
  }

  /* if (verb & 0x01)
    for (i=0;i<=(ih-il);++i)
    mexPrintf("W[%d] = %g\n", i, W[i]); */

  o->mu_shift = W[0];
  return 0;
}



/*--------------------------------------------------------------------------*/
void append_lat_out(const struct Onedim_FD_system *o, double *u)
     /* save time-slice of lattice data, preceded by lattice spacing */
{
  int i;

  fprintf(fp_lat_out, "%g ", o->h); /* lattice spacing */
  for (i=0; i<o->nz; ++i)
    fprintf(fp_lat_out, "%g ", u[i]);
  fprintf(fp_lat_out, "\n");
}


/*--------------------------------------------------------------------------*/
int detection(double *u, double t, int templ_i, double *templ_val,
	      struct Onedim_signal *uzt)
/* Detection model for a 1d system. Writes current detected signal u
 * into the uzt function, with time=t, and current uzt node i.
 * Protects against overrunning uzt arrays.
 * The detection template previously set up in onedim_FD (where det models
 * are defined) consists of 3 node weights (templ_val[0..2]) and a starting
 * node index (templ_i). Redundancy of 3 weights (sometimes 2 needed) is
 * acceptable.
 *
 * barnett 4/19/03
 */
{
  if (uzt->nt<MAX_TNODES) {
    uzt->u[uzt->nt] = templ_val[0] * u[templ_i] + templ_val[1] * u[templ_i+1]
      + templ_val[2] * u[templ_i+2];
    uzt->t[uzt->nt] = t;
    /* increment current node for uzt... */
    ++uzt->nt;
    return 0;
  } else
    return 1;
}


/*--------------------------------------------------------------------------*/
int c_n_steps(double *t, double t_max, double dt, double *f, double *g,
	      const struct Onedim_FD_system *o, struct Onedim_signal *uzt,
	      double alpha, int det_i, double *det_v)
/* performs several Crank-Nicholson (or general implicit-explicit) time-steps,
 * updating uzt, t, nt, as goes. alpha = 0 explicit, 1 implicit, 1/2 C-N.
 * uzt->nt counts uzt nodes. input f. output is in f (g) if s is even (odd).
 *
 * barnett 12/6/02
 *         12/20/02 added option for variable explicit->implicit fraction.
 *         4/19/03 converted to detection(), changed args (removed i).
 *         4/26/03 changed factorization to pos def symm tridiag.
 */
{
  int j, s;      /* s counts steps */
  double D[MAX_ZNODES], E[MAX_ZNODES]; 
  int nrhs = 1, ldb = o->nz + 1; /* other pass-by-reference args for LAPACK */
  int info, n = o->nz + 1;
  double dt_fe = (1.0-alpha)*dt;   /* forward time-step part */
 
  /* factorize */
  build_LU(o, alpha*dt, D, E);

  for (s=0; (*t)<=t_max; (*t)+=dt, ++s) {

    /* forward half-step... */
    if (dt_fe != 0.0) {
      f_euler(o, f, g, dt_fe);
    } else {
      /* just copy, so that swap leave us in right place... */
      for (j=0; j<=o->nz; ++j)
	g[j] = f[j];
    }
    swap(&f, &g);                  /* swap local copies of pointers */

    /* backward half-step : tridiag solve... input/output is in f */
    dpttrs_(&n, &nrhs, D, E, f, &ldb, &info);
    if (info!=0)
      mexPrintf("problem in dpttrs, info=%d @ t = %f", info, t);

    detection(f, *t+dt, det_i, det_v, uzt); /* det u(z,t) at time *t+dt */
    if (o->lat_out)
      append_lat_out(o, f); 
  }

  return s;
}


/*--------------------------------------------------------------------------*/
int onedim_FD(struct Onedim_signal *uzt,
	      const struct Multilayer_slab *s,
	      const struct SingleSD *m,
	      struct Onedim_FD_system *o, double k, int o_verb)
/* Time-evolution of 1d system (corresponding to single radial mode).
 * Transverse wavenumber k is used to compute mu_e from mu_a.
 * Finite-difference (FD) method is used.
 * Initial condition is a point-source at depth z=z0.
 * Fills time-node array in uzt with desired measured signal.
 * o_verb is flag to write basic mode info to Matlab stdout
 *
 * barnett 11/21/02, added Extrap BCs and general detection model 4/19/03.
 * 7/21/03, added gradient EBC meas.
 */
{
  int i, ns, ns_fe, ns_c, det_i;            /* i counts time nodes in uzt */
  double w1[MAX_ZNODES], w2[MAX_ZNODES];  /* work arrays */
  double *u = w1, *v = w2;                /* movable pointers to arrays */ 
  double t_stop, z0, dt, t, a, dtot, dt_max, min_mue = 1e12, max_mue = -1e12;
  double alpha, depth_over_h, det_v[3], frac, dt0, t0, eps_kill;

  /* mu_shift(o, k);*/ /* regenerate mu_shift for each 1D system ? */
  uzt->mu_shift = o->mu_shift; /* define the mu_shift for this mode */
  /* uzt->mu_shift += s->ckap_min*k*k;*/   /* quadratic mu_shift change ok */
  /* uzt->mu_shift = o->mu_shift - (random()/2.1475e+9)*0.001; */
  /* printf("mu_sh[k=%g] = %g\n", k, uzt->mu_shift); */

  /* make mu_e from mu_a... keeping track of range found. Note kap not inv */
  for (i=0; i<o->nz; ++i) {
    o->mue_h[i] = o->mua_h[i] - uzt->mu_shift + k*k * o->kap_h[i];
    min_mue = min(min_mue, o->mue_h[i]);
    max_mue = max(max_mue, o->mue_h[i]);
  }

  /* estimate necessary time for contribution... with factor adj by hand */
  t_stop = min( log(10/m->err_trunc)/(k*k*s->ckap_min) , m->tmax);
  if (o_verb)
    mexPrintf("\t\tonedim_FD: t_stop=%g, mu_sh=%g, %.3g<mue<%.3g\n",
	      t_stop, uzt->mu_shift, min_mue, max_mue);
  if (t_stop<m->tmin) {
    uzt->nt = 0;
    if (o_verb)
      mexPrintf("\t\tt_stop < t_min, no contrib.\n");
    return 0;
  }

  /* Initial Condition... lambda (hat) function of h-size, fixed area */
  for (i=0; i<=o->nz; ++i)
    u[i] = 0.0;
  depth_over_h = (s->z0+s->zb)/o->h;  /* depth from z = -zb */
  i = (int)(depth_over_h);
  u[i+1] = depth_over_h - i;
  u[i]   = 1.0 - u[i+1];
  u[0]   = 0.0;           /* Dirichlet BCs apply to IC too (if z0+zb < h) */
  u[i] /= o->h;           /* make unit area under IC (indep of h) */
  u[i+1] /= o->h;
  if (verb & 0x40)        /* dump initial condition vector */
    for (i=0; i<=o->nz; ++i)
      mexPrintf("i=%d\tz=%g\tui=%g\tmue_h=%g\n", i, (i*o->h) - s->zb,
		u[i], o->mue_h[i]);

  /* Set up measurement (detection) vector: 3 values and a start index */
  depth_over_h = s->zb / o->h;  /* depth of meas z=0 relative to z=-zb */
  if (s->bcs==1) {
    /* value (linear-interp in z) at z=0, multipled by kappa/zb */
    i = (int)(max(depth_over_h, 0.0));
    frac = depth_over_h - i;
    det_i = i;
    det_v[0] = (1-frac) * s->kap_top / s->zb;
    det_v[1] = frac     * s->kap_top / s->zb;
    det_v[2] = 0.0;
  } else if (s->bcs==0) {
    /* use first-order surface gradient at z=0, mult by kappa */
    det_i = 0;
    det_v[0] = 0.0;
    det_v[1] = s->kap_top / o->h;
    det_v[2] = 0.0;
  } else if (s->bcs==2) {
    /* gradient (linear-interp in z) at z=0, mult by kappa */
    i = (int)(max(depth_over_h - 0.5, 0.0)); /* ensure not fall off array */
    frac = depth_over_h - 0.5 - i;  /* f goes negative if zb<h/2, ok */
    det_i = i;
    det_v[0] = (frac-1)   * s->kap_top;
    det_v[1] = (1-2*frac) * s->kap_top;
    det_v[2] = frac       * s->kap_top;
  } else {
    /* lin comb of value and gradient at z=0 (Kienle97, Eq.8) */
    mexErrMsgTxt("onedim_FD: bcs=3 det vec not yet implemented!");
    return 1;
  }
  if (o_verb)
    mexPrintf("\t\tdet meas: det_i=%d, det_v=(%g,%g,%g)\n", det_i,
	      det_v[0], det_v[1], det_v[2]);

  /* 'clear' the uzt array... */
  uzt->nt = 0;

  if (o->lat_out)
    fp_lat_out = fopen("test.1dlat.dat","w");

  if (o->method==F_EULER) {
    /* .......................... Forward Euler, const dt .................. */

    dt = o->dt_euler_kill_hf;
    if (o_verb)
      mexPrintf("\t\t\tf_euler (max courant#=1/2): dt=%g\n", dt);

    /* timestep loop */
    for (t=0.0, i=0; t<t_stop; t+=dt, ++i) {
      f_euler(o, u, v, dt);
      swap(&u, &v);                                /* output is now in u */
      detection(u, t+dt, det_i, det_v, uzt); /* meas u(z,t) at time t+dt */
      if (o->lat_out)
	append_lat_out(o, u); 
    }
    if (o_verb)
      mexPrintf("\t\t\t\tdone %d const-dt FE steps\n", i);


  } else if (o->method==B_EULER) {
    /* ..................... Implicit-Explicit, const dt .................. */

    dt = 2.8 * o->err_dt * m->tmin;  /* ! need to modify in conj with alpha */
    if (o_verb)
      mexPrintf("\t\t\tb_euler : dt=%g\n", dt);

    t = 0.0;
    i = 0;
    alpha = 1/(1 + pow(1e-6, dt/m->tmin));
    if (o_verb)
      mexPrintf("\t\t\t\talpha = %g\n", alpha);    /* euler mixing param */
    ns = c_n_steps(&t, t_stop, dt, u, v, o, uzt, alpha, det_i, det_v);
    if (ns%2)
      swap(&u, &v);               /* output is now in u */
    if (o_verb)
      mexPrintf("\t\t\t\tdone %d const-dt expl-impl steps\n", ns);


  } else if (o->method==C_N) {
    /* .......................... Crank-Nic, const dt, no kill .......... */

    if (o->err_dt>0.0)
      dt = 2.0 * m->tmin * sqrt(o->err_dt);  /* adjust */
    else
      dt = -o->err_dt;
    t = 0.0;
    i = 0;
    if (o_verb)
      mexPrintf("\t\t\tC-N : dt=%g\n", dt);
    /* Cranck-Nicholson: alpha=1/2... gets you to t_stop */
    ns = c_n_steps(&t, t_stop, dt, u, v, o, uzt, 0.5, det_i, det_v);
    if (ns%2)
      swap(&u, &v);               /* output is now in u */
    if (o_verb)
      mexPrintf("\t\t\t\tdone %d steps.\n", ns);


  } else if (o->method==C_N_VAR) {
    /* .......................... Crank-Nic, variable dt .................. */

 /* C-N dt/t, const for evolution */
    if (o->err_dt>0.0)
      dtot = sqrt(2.8 * o->err_dt);
    else
      dtot = -o->err_dt;

    /* how many FE before worth crossing over... */    
    ns_fe = (int)(CN_SLOWER_FE / dtot);
    /* ns_fe Euler steps to kill high modes (including BC-induced) */
    dt = o->dt_euler_kill_hf; 
    /*dt = o->dt_euler_kill_hf/(1+k*k);*/  /* inconsistent scheme demo */

    if (o_verb)
      mexPrintf("\t\t\tC-N var : dt/t=%g, dt_kill=%g, ns_fe=%d\n",
	     dtot, dt, ns_fe);
    for (t=0.0, i=0; i<ns_fe; t+=dt, ++i) {
      f_euler(o, u, v, dt);
      swap(&u, &v);                /* output is now in u */
      detection(u, t+dt, det_i, det_v, uzt); /* meas u(z,t) at time t+dt */
      if (o->lat_out)
	append_lat_out(o, u); 
    }

    /* C-N, using dt based on current t (not tmin): need to kill hf as go */
    dt = dtot*t;
    while (t<=t_stop) {
      /* just 1 CN step... */
      ns = c_n_steps(&t, t, dt, u, v, o, uzt, 0.5, det_i, det_v);
      if (ns!=1)
	mexPrintf("\t\t\tc_n_steps didn't do 1 step!\n");
      if (ns%2)
	swap(&u, &v);               /* output is now in u */
      dt = dtot*t;
    }
    if (o_verb)
      mexPrintf("\t\t\t\tdone %d var-dt C-N steps\n", uzt->nt-ns_fe);


  } else if (o->method==C_N_CONST_VAR) {
    /* .......................... Crank-Nic, const-then-var dt ............. */

    if (o->err_dt>0.0)
      dtot = 2.0 * m->tmin * sqrt(o->err_dt);  /* adjust */
    else
      dtot = -o->err_dt;
    ns_fe = 0;  /* no FE steps */
    /* decide if to override const CN segment params... */
    if (o->dt0==0.0 && o->t0==0.0) {
      /* compute dt_c and t0 given beta and lattice params... */
      eps_kill = 1e-9;     /* fixed allowed err from CNconst HF mode decay */
      dt0 = 4*o->dt_euler_kill_hf/(dtot*log(1/eps_kill));
      t0 = dt0/dtot;
      if (t0 > m->tmin) {
	/* limit t0 to tmin and recompute max dt_c possible... */
	t0 = m->tmin;
	dt0 = sqrt(4*t0*o->dt_euler_kill_hf/log(1/eps_kill));
      }
    } else {
      /* compute the missing parameter, if necessary, or override both... */
      if (o->dt0==0.0) {
	t0 = o->t0;
	dt0 = t0*dtot;
      } else if (o->t0==0.0) {
	dt0 = o->dt0;
	t0 = dt0/dtot;
      } else {
	/* override both... careful since can have HF errors! */
	dt0 = o->dt0;
	t0 = o->t0;
      }
    }
    if (o_verb)
      mexPrintf("\t\t\tC-N const-var : ns_fe=%d, dt0=%g, t0=%g, dt/t=%g\n",
		ns_fe, dt0, t0, dtot);

    /* ns_fe Euler steps to kill high modes */
    dt = o->dt_euler_kill_hf;
    for (t=0.0, i=0; i<ns_fe; t+=dt, ++i) {
      f_euler(o, u, v, dt);
      swap(&u, &v);                /* output is now in u */
      detection(u, t+dt, det_i, det_v, uzt); /* meas u(z,t) at time t+dt */
      if (o->lat_out)
	append_lat_out(o, u); 
    }

    /* Cranck-Nicholson: alpha=1/2... gets you to t_stop */
    ns = c_n_steps(&t, t0, dt0, u, v, o, uzt, 0.5, det_i, det_v);
    if (ns%2)
      swap(&u, &v);               /* output is now in u */
    if (o_verb)
      mexPrintf("\t\t\t\tdone %d const dt=%g steps.\n", ns, dt0);
    ns_c  = ns;

    /* C-N, using dt based on current t (not tmin) */
    dt = dtot*t;
    while (t<=t_stop) {
      /* just 1 CN step... */
      ns = c_n_steps(&t, t, dt, u, v, o, uzt, 0.5, det_i, det_v);
      if (ns!=1)
	mexPrintf("\t\t\tc_n_steps didn't do 1 step!\n");
      if (ns%2)
	swap(&u, &v);               /* output is now in u */
      dt = dtot*t;
    }
    if (o_verb)
      mexPrintf("\t\t\t\tdone %d var-dt C-N steps\n", uzt->nt - ns_c);
  }

  if (o->lat_out)
    fclose(fp_lat_out);

  return 0;
}


/* how many ps to enclose meas t range with in onedim_dummy */
#define SAFETY_TIME_SPILL 100

/*--------------------------------------------------------------------------*/
int onedim_dummy(struct Onedim_signal *uzt,
	      const struct Multilayer_slab *s,
	      const struct SingleSD *m,
	      struct Onedim_FD_system *o, double k, int o_verb)
/* Places a function in uzt which is 1, constant in time, over required
 * time range. Also sets mode decay rate correct for 2d calc mode.
 *
 * barnett 12/5/02
 */
{
  uzt->nt = 2; /* two points both with value 1 */
  uzt->t[0] = m->tmin - SAFETY_TIME_SPILL;
  uzt->t[1] = m->tmax + SAFETY_TIME_SPILL;
  uzt->u[0] = 1;
  uzt->u[1] = 1;

  /* decay rate for mode wavenumber k, using top 'layer' properties... */
  uzt->mu_shift = s->cmua[0] + s->ckap[0] * k*k;

  return 0;
}



/*--------------------------------------------------------------------------*/
int add_as_samples(double *sig, const struct SingleSD *m,
		   const struct Onedim_signal *uzt, double *a)
     /* Interpolates a non-uniform ordered set of (t,u) values onto a
      * regular time-grid
      *	of samples, adding a multiple (a[k]) of this to sig array,
      * for each radius index k in sig array.
      * If the regular grid goes beyond the last t, pads with zeros quietly.
      *
      * Need to include exp. and cubic interp. options!
      *
      * barnett 11/24/02, vectorized for N_R>1 12/01/02
      * 12/5/02: include multiplicative exponential decay mue_typ.
      * 4/25/03: mue_typ renamed mu_shift.
      */
{
  int i, j, k;
  double t, frac, u_interp;

  j = 0;
  /* j is index in uzt array; i is index in regularly-spaced sig array */ 
  for (i=0; i<m->ntgate; ++i) {
    t = m->tmin + i*m->dtgate;       /* sample time is at start of timegate */
    /* search for interval in uzt data in which t lies... */
    while (j<uzt->nt-1 && uzt->t[j+1]<t) {
      ++j;
    }
    if (j>=uzt->nt-1) {
      if (verb & 0x10)
	mexPrintf("slab.c add_as_samples : j=%d overrun nt, i=%d\n", j, i);
      /* return 1; */
    } else {
      /* linear interpolation (is O(dt^2) accurate) */
      frac = (t - uzt->t[j])/(uzt->t[j+1] - uzt->t[j]);
      u_interp = frac*uzt->u[j+1] + (1.0-frac)*uzt->u[j];
      /* exponential decay AFTER interpolation... */
      if (uzt->mu_shift!=0.0)
	u_interp *= exp(-uzt->mu_shift * t);
      for (k=0; k<m->N_R; ++k)
	sig[k*m->ntgate + i] += a[k] * u_interp;
      if (verb & 0x10)
	mexPrintf("i=%d, j=%d, frac=%f, u[j]=%g, u[j+1]=%g\n",i,j,frac,uzt->u[j],
	       uzt->u[j+1]);
    }
  }

  return 0;
}



/*--------------------------------------------------------------------------*/
int calculate(double tmin, double dtgate, int ntgate, double *R, int N_R,
	      int N_l,
	      double* musp, double* mua, double* dz, int semi_inf,
	      double refr, double* err, double* sig)

     /* all inputs apart from sig which is output.
      *
      * see slab.m for docs.
      */
{
  struct Multilayer_slab *sys; 
  struct SingleSD *meas;
  struct Onedim_FD_system *oned;
  struct Onedim_signal *uzt;
  double a[MAX_RS], l_box, dk, k_max, sigma_tmax, k_m, R_max, mode_decay;
  int i, j, m, m_min, m_max, N_m, calc_mode;
  char oned_file[] = "test.1d.dat";
  char twod_file[] = "test.2d.dat";
  FILE *fp;

  /* debug from Chris Gaal @ mathworks support... always better to point
     to struct, since when pass, doesn't duplicate, etc, from Adi Rangan */
  sys = malloc(sizeof(struct Multilayer_slab));
  meas = malloc(sizeof(struct SingleSD));
  oned = malloc(sizeof(struct Onedim_FD_system));
  uzt = malloc(sizeof(struct Onedim_signal));
 
  build_slab(sys, N_l, musp, mua, dz, semi_inf, refr, err);

  /* set up meas struct */
  for (i=0; i<N_R; ++i)
    meas->R[i] = R[i];
  meas->N_R = N_R;
  meas->tmin = tmin;
  meas->dtgate = dtgate;
  meas->ntgate = ntgate;
  meas->tmax = tmin + ntgate*dtgate;  /* maximum time we care about */
  meas->err_lbox = err[3];
  meas->err_trunc = err[4];

  /* encode things from err[] into onedim struct... */
  oned->lat_out = (verb & 0x100); /* tell onedim_FD to output lat FD data */
  
  /* decide calculational mode (1d, 2d, 3d, etc)... */
  calc_mode = min(1, ROUND(err[5]));
  if (verb & 0x01)
    mexPrintf("\n\tcalc mode: %s\n", calc_mode_label[calc_mode-CALC_MODE_OFFSET]);

  /* find R_max */
  R_max = 0.0;
  for (i=0; i<N_R; ++i)
    R_max = max(R_max, R[i]);

  if (calc_mode!=TWO_D) {
    /* initialize the 1d u(z,t) method... */
    if (err[7]==0) {
      oned->method = B_EULER;
      if (verb & 0x01)
	mexPrintf("\ttime-step scheme 0: const dt BE (mixed implicit-explicit)\n");
    } else if (err[7]==1) {
      oned->method = C_N_VAR;
      if (verb & 0x01)
	mexPrintf("\ttime-step scheme 1: variable dt C-N (initial const dt FE)\n");
    } else if (err[7]==2) {
      oned->method = F_EULER;
      if (verb & 0x01)
	mexPrintf("\ttime-step scheme 2:  const dt FE (max Courant# = 1/2)\n");
    } else if (err[7]==3) {
      oned->method = C_N;
      if (verb & 0x01)
	mexPrintf("\ttime-step scheme 3:  const dt CN (no kill)\n");
    } else if (err[7]==4) {
      oned->method = C_N_CONST_VAR;
      if (verb & 0x01)
	mexPrintf("\ttime-step scheme 4:  const-then-var CN\n");
    } else
      mexPrintf("\tinvalid time-step, err[7]!");
    init_FD(oned, sys, meas, err);
  }



  /* zero signal bins ready for summing fourier modes... */
  for (i=0; i<ntgate; ++i)
    for (j=0; j<N_R; ++j)
      sig[j*ntgate + i] = 0.0;



  if (calc_mode==ONE_D) {
    /* u(z,t) for k=zero-mode only...................................... */

    /* Time evolve 1d (k=0): sys etc as pointers for speed */
    onedim_FD(uzt, sys, meas, oned, 0.0, (verb & 0x02) || (verb & 0x80));
    
    /* file dump for diagnosis (including exp decay in uzt) */
    fp = fopen(oned_file, "w");
    for (i=0; i<uzt->nt; ++i)
      fprintf(fp, "%.12g %.12g\n", uzt->t[i],
	      uzt->u[i]*exp(-uzt->t[i]*uzt->mu_shift));
    fprintf(fp, "\n");
    fclose(fp);
    
    /* unit contrib (for each R value) */
    for (j=0; j<N_R; ++j)
      a[j] = 1.0;

    add_as_samples(sig, meas, uzt, a);
    
  } else {
    /* other calc modes (2d, 3d...) ................................... */
    
    /* use top-layer properties in 2d */
    if (calc_mode==TWO_D)
      sys->ckap_max = sys->ckap[0];

    /* find box radius lbox (called a) */
    l_box = R_max + sqrt(sys->ckap_max * meas->tmax * log(4/meas->err_lbox));

    /* find k_max (assume slowest spread) */
    k_max = sqrt(log(1 / meas->err_trunc) / (sys->ckap_min*meas->tmin));

    /* find # modes */
    N_m = (int)(1 + l_box * k_max / PI);

    if (verb & 0x01)
      mexPrintf("\txy plane: abox=%g, kmax=%g, N_m=%d\n", l_box, k_max, N_m);

    if (N_m > MAX_MODES) {
      mexPrintf("slab.c calculate : Exceeded max # radial modes!\n");
      return 1;
    }

    /* compute Euler initial timestep for CNvar (needed N_m for this)... */
    k_max = J0_ZEROS[N_m] / l_box; /* k_max given by actual N_m */
    oned->dt_euler_kill_hf = 1/(sys->ckap_max*(4/(oned->h*oned->h) + k_max*k_max));
    /*oned->dt_euler_kill_hf = 1/(sys->ckap_max*(4/(0.1*0.1) + k_max*k_max));*/
    if (verb & 0x01)
      mexPrintf("\tk_max adj dt_kill to %g\n", oned->dt_euler_kill_hf);


    /* compute mu_shift for CNvar... sets oned->mu_shift using k_0 */
    if (oned->method==C_N_CONST_VAR) {
      mu_shift(oned, J0_ZEROS[0] / l_box);
      if (verb & 0x01)
	mexPrintf("\toverall mu_shift = %g\n", oned->mu_shift);
    }

    if (calc_mode==THREE_D_SINGLE_MODE) {
      /* single radial fourier mode only (subtract 1 from Matlab ordering) */
      m_min = ROUND(err[5]) - 1;
      m_max = m_min+1;
      if (verb & 0x01)
	mexPrintf("\tonly mode %d contrib\n", m_min);
    } else {
      /* all modes */
      m_min = 0;
      m_max = N_m;
      if (verb & 0x01)
	mexPrintf("\tsumming all modes\n");
    }
    
    if (calc_mode==THREE_D_SINGLE_MODE)
      fp = fopen(oned_file, "w");
    else if (calc_mode==TWO_D)
      fp = fopen(twod_file, "w");
    


    /* loop over modes.......................................... */
    for (m=m_min; m<m_max; ++m) {

      /* transverse wavenumber of mode m */
      k_m = J0_ZEROS[m] / l_box;
      if (verb & 0x02)
	mexPrintf("\t\tk[%d] = %g\n", m, k_m);

      /* compute weights a using bessels */
      for (j=0; j<N_R; ++j) {
	a[j] = j0(k_m * R[j]) / (PI * SQ(l_box * J1_AT_J0_ZEROS[m]));
	if (verb & 0x08)
	  mexPrintf("\t\t\ta(R[%d]=%g) = %g\n", j, R[j], a[j]);
      }

      if (calc_mode==TWO_D)
	/* dummy evolution (constant, setting mu_shift for correct decay */
	onedim_dummy(uzt, sys, meas, oned, k_m, 0);
      else
	/* Time-evolve 1d: sys etc are as pointers for speed */
	onedim_FD(uzt, sys, meas, oned, k_m, (verb & 0x02) || (verb & 0x80) && m==0);
      

      
      if (calc_mode==THREE_D_SINGLE_MODE) {
	/* if single mode contrib only, file dump for diagnosis */
	for (i=0; i<uzt->nt; ++i)
	  fprintf(fp, "%.12g %.12g\n", uzt->t[i], uzt->u[i]);
	fprintf(fp, "\n");
      } else if (calc_mode==TWO_D) {
	/* output the components due to each mode (columns sweep time) */
	for (j=0; j<N_R; ++j) {
	  for (i=0; i<ntgate; ++i)
	    fprintf(fp, "%.18g ", a[j]*exp(-uzt->mu_shift * (tmin+i*dtgate)));
	  fprintf(fp, "\n");
	}
	fprintf(fp, "\n");
      }    


      /* add contrib to sig array */
      add_as_samples(sig, meas, uzt, a);
    }

    if (calc_mode==THREE_D_SINGLE_MODE || calc_mode==TWO_D)
      fclose(fp);
    
  }

  if (0) {  /* option to clip */
  /* clip negative sig values, only if in full summing mode... */
  if (calc_mode==THREE_D) {
    for (i=0; i<ntgate; ++i)
      for (j=0; j<N_R; ++j)
	sig[j*ntgate + i] = max(0.0, sig[j*ntgate + i]);
  }
  }

  /* dump all signals to stdout if you want... */
  if (verb & 0x04) {
    mexPrintf("\nOutput signals...\n");
    for (j=0; j<N_R; ++j) {
      mexPrintf("\n\tradius R[%d] = %g:\n", j, R[j]);
      for (i=0; i<ntgate; ++i)
	mexPrintf("\t\ts(t[%d]=%g) = %g\n", i, tmin+dtgate*i,
		  sig[j*ntgate + i]);
    }
  }

  free(sys);
  free(meas);
  free(oned);
  free(uzt);

  return 0;
}



#ifdef MATLAB_MEX_FILE
/* ................. From here on is Matlab-C interface .............. */


/* ------------- matlab-C interface args ------------- */
/* Input */
#define	TMIN_IN	 prhs[0]
#define	TMAX_IN	 prhs[1]
#define	DTSAM_IN prhs[2]
#define	R_IN	 prhs[3]
#define	MUA_IN   prhs[4]
#define	MUSP_IN	 prhs[5]
#define DZ_IN    prhs[6]
#define N_IN     prhs[7]
#define ERR_IN   prhs[8]
/* Output */
#define	SIG_OUT	 plhs[0]


/* ------------------- "main" : I/O & arg error checking & arg defaults -----
 * see matlab External API interfaces docs.
 */
void mexFunction(int nlhs, mxArray *plhs[], 
		 int nrhs, const mxArray *prhs[])
{ 
  double tmin, tmax, dtgate, *R, *musp, *mua, *dz, n, *err, *err_args, *sig;
  int N_R, N_l, ntgate, i, j, semi_inf;
    
  /* Check for proper number of arguments */
  if (nrhs < 7 || nrhs > 9) 
    mexErrMsgTxt("slab.c: 7-9 input arguments required."); 
  else if (nlhs > 1)
    mexErrMsgTxt("slab.c: too many output arguments."); 
 
    
  /* get arguments & check their sizes.... */
  tmin = mxGetScalar(TMIN_IN);
  tmax = mxGetScalar(TMAX_IN);
  dtgate = mxGetScalar(DTSAM_IN);

  /* number of Radii */
  N_R = mxGetN(R_IN);
  if (N_R > MAX_RS)
    mexErrMsgTxt("slab.c: too many R values requested!");
  if (mxGetM(R_IN)!=1)
    mexErrMsgTxt("slab.c: R must be 1 row vector"); 
  R = mxGetPr(R_IN);

  /* number of layers */
  N_l = mxGetN(MUSP_IN);
  if (N_l > MAX_LAYERS)
    mexErrMsgTxt("slab.c: too many layers!");

  /* absorption coefficients */
  if (mxGetM(MUA_IN)!=1)
    mexErrMsgTxt("slab.c: mua must be 1 row vector"); 
  mua = mxGetPr(MUA_IN);

  /* reduced scattering coefficients */
  if (mxGetM(MUSP_IN)!=1)
    mexErrMsgTxt("slab.c: musp must be 1 row vector"); 
  if (mxGetN(MUSP_IN)!=N_l)
    mexErrMsgTxt("slab.c: length of musp must equal mua"); 
  musp = mxGetPr(MUSP_IN);

  /* thicknesses */
  if (N_l==1) {
    if (!(mxGetN(DZ_IN)==0 || (mxGetM(DZ_IN)==1 && mxGetN(DZ_IN)==1)))
      mexErrMsgTxt("slab.c: 1 layer, so dz must be empty or 1-by-1"); 
  } else {
    if (mxGetM(DZ_IN)!=1)
      mexErrMsgTxt("slab.c: dz must be 1 row vector"); 
    if (mxGetN(DZ_IN) != N_l-1 && mxGetN(DZ_IN) != N_l)
      mexErrMsgTxt("slab.c: length of dz must be equal to or 1 less than musp");
  }
  /* set semi-infinite flag, must be 0 or 1 */
  semi_inf = (mxGetN(DZ_IN)==N_l-1);
  dz = mxGetPr(DZ_IN);

  /* refractive index */
  n = DEFAULT_REFR;
  if (nrhs > 7 && mxGetM(N_IN)==1)
    n = mxGetScalar(N_IN);

  /* error tolerances */
  err = mxMalloc(N_ERR*sizeof(double));
  for (i=0; i<N_ERR; ++i)
    err[i] = DEFAULT_ERR[i];
  if (nrhs > 8) {
    if (mxGetM(ERR_IN)!=1 || mxGetN(ERR_IN)>N_ERR)
      mexErrMsgTxt("slab.c: err must be 1 row vector of length <=N_ERR"); 
    err_args = mxGetPr(ERR_IN); /* maybe shorter incoming err argument array */
    /* copy into true err array... */
    for (i=0; i<mxGetN(ERR_IN); ++i)
      err[i] = err_args[i];
  }

  /* set global verbosity */
  verb = (int)(0.5 + err[6]);

  if (verb & 0x01)
    mexPrintf("slab.c v1.5: %d input args, %d output args\n", nrhs, nlhs);
  /* how many output time samples (row vector length)... */
  ntgate = (int)((tmax-tmin)/dtgate + 1.0);
  if (verb & 0x01)
    mexPrintf("\tntgate = %d, N_R = %d\n", ntgate, N_R);
  if (verb & 0x02)
    for (j=0; j<N_R; ++j)
      mexPrintf("\tradius R[%d] = %g\n", j, R[j]);

  /* Create a matrix for the return argument : vertical=fast direction */ 
  SIG_OUT = mxCreateDoubleMatrix(ntgate, N_R, mxREAL);
  sig = mxGetPr(SIG_OUT);

  /* Do calc.... fills sig array with answer */
  calculate(tmin, dtgate, ntgate, R, N_R, N_l, musp, mua, dz, semi_inf, n,
	    err, sig);

  return;
}

#endif
