/* slab.h: Global defines and variables for slab.c
 *
 * note that some #defines specific to routines in slab.c happen there.
 *
 * barnett 7/28/03
 */

/* external LAPACK library functions */
extern void dstevx_(char *, char *, int *, double *, double *, double *, \
		    double *, int *, int *, double *, int *, double *, \
		    double *, int *, double *, int *, int *, int *);
extern void dpttrf_(int *, double *, double *, int *);
extern void dpttrs_(int *, int *, double *, double *, double *, \
		    int *, int *);


/* limits */
#define MAX_ZNODES 10000
#define MAX_TNODES 200000 /* 200000 fails between 100000 and 125000 */
#define MAX_LAYERS 100
#define MAX_RS     100
#define MAX_SAMPS 10000


#define DEFAULT_REFR 1.4

/* Default values used if no err array provided... */
#define N_ERR    13          /* length of err input array */
const double DEFAULT_ERR[] = {
  1e-6,   /* 0:zbox  (z)          matlab index 1 */
  1e-2,   /* 1:h     (z)                       2 */
  -0.05,  /* 2:dt    (z)   fixed beta=0.05     3 */
  1e-6,   /* 3:lbox  (r)                       4 */
  1e-12,  /* 4:trunc (r)                       5 */
  -1,     /* 5:full 3d                         6 */
  0,      /* 6:not verbose - no output at all  7 */
  4,      /* 7:default t-step: CN const-var    8 */
  1,      /* 8:BCs: Haskell's EBC              9 */
  0,      /* 9:t0 override for C_N_CONST_VAR  10 */
  0,      /* 10:dt0 override for same         11 */
  0,      /* 11:minimum mu_s' override        12 */
  0       /* 12:maximum mu_s' override        13 */
};


/* ...... from raymond s. norris (rayn@mathworks.com) : */
#ifndef MATLAB_MEX_FILE
  /* compilation as linux executable... */
  #include <stdio.h>
  #include <stdlib.h>
  #undef  printf
  #define mexErrMsgTxt printf
  #define mexPrintf printf
#else
  /* compilation as matlab MEX file... */
  /* can play around with how want to output... when as MEX file */
  #undef  printf
  #define printf mexPrintf
  #include "mex.h"
#endif

#include <math.h>

/* Bessel zeros and coeffs... generated from MATLAB by bessel_zeros/jzeros.m */
#include "bessel_zeros/j0.h"

#define LIGHTSPEED 0.2997   /* in mm/ps */

int verb; /* global verbosity */

FILE *fp_lat_out; /* global filepointer for debug writing FD lattice file */

/* calc mode labels... */
static char *calc_mode_label[] = {
  "2d, top layer props, no gradient",
  "full 3d (default)",
  "1d u(z,t) (all R's are duplicate)",
  "3d contrib from only single mode"};
#define CALC_MODE_OFFSET   -2       /* keep equal to lowest calc mode */
#define TWO_D              -2
#define THREE_D            -1
#define ONE_D               0
#define THREE_D_SINGLE_MODE 1

/* rounding float to int conversion */
#define ROUND(x) ((x)>=0?(long)((x)+0.5):(long)((x)-0.5))

#ifndef SQ
#define SQ(x) ((x)*(x))
#endif

#ifndef max
#define max(a,b)	((a) > (b) ? (a) : (b))
#endif

#ifndef min
#define min(a,b)	((a) < (b) ? (a) : (b))
#endif

#ifndef PI
#define PI 3.141592653589793
#endif


#define LARGE_DEPTH 1e6   /* can't be too large, since get int overflow */

/* objects */
struct Multilayer_slab {               /* system */
  int N_l;
  double cmua[MAX_LAYERS];             /* speed-rescaled absorption */
  double ckap[MAX_LAYERS];             /* speed-rescaled diffusivity */
  double ckap_max, ckap_min;           /* range of ckap */
  double cmua_min;                     /* range of cmua */
  double z[MAX_LAYERS];                /* finishing depths of each layer */
  int semi_inf;  /* if true, z goes from 0 to N_l-2,  otherwise 0 to N_l-1 */
  double z0;                           /* Initial Condition release depth */
  int bcs;                             /* BCs and det model choice */
  double zb;                           /* z_b for EBC usage (0 for ZBC) */
  double kap_top;                      /* unscaled (raw) kappa of top layer */
};

struct SingleSD {                      /* measurement */
  double R[MAX_RS];
  int N_R;
  double tmin;
  double tmax;
  double dtgate;
  int ntgate;
  double err_lbox, err_trunc;        /* errors relevant to summing modes */
};

struct Onedim_FD_system {             /* 1d system */
  int method;
  int nz;                         /* # z-nodes including ONE bdry not other */
  double h;                       /* lattice spacing */
  double mua_h[MAX_ZNODES];       /* speed-rescaled absorption of cells */
  double mue_h[MAX_ZNODES];       /* speed-rescaled, mu-shifted abs of cells */
  double kap_h[MAX_ZNODES];       /* speed-rescaled cell-avg diffusivity */
  double kapi_h[MAX_ZNODES];       /* speed-rescaled cell-inv-avg diffu. */
  int opts;
  double dt_euler_kill_hf;        /* launch material initial smoothing */
  double mu_shift;                /* mu shift */
  double err_dt;                  /* err(3) passed to onedim_FD, see slab.m */
  double t0;                      /* err(10) passed to onedim_FD, see slab.m */
  double dt0;                     /* err(11) passed to onedim_FD, see slab.m */
  int lat_out;                    /* flag to request 1d lattice file output */
};

/* 1d FD methods */
#define F_EULER 1
#define B_EULER 2
#define C_N     3
#define C_N_VAR 4
#define C_N_CONST_VAR 5

/* 1d BC and det choices */
#define ZBC     0
#define EBCV    1
#define EBCF    2
#define EBCK    3
#define N_BCS   4
static char *bcs_names[] = {
  "0 (ZBC): Zero BC (Dirichlet at z=0), det gradient",
  "1 (EBCV): Extrap. BC (Dir. at z=-zb), det value (Haskell94)(R3)",
  "2 (EBCF): Extrap. BC (Dir. at z=-zb), det gradient (Martelli1999,R1)",
  "3 (EBCK): Extrap. BC (Dir. at z=-zb), det lin comb (Kienle97)(R2)"};


#define CN_SLOWER_FE 7          /* how many times a C-N step is than FE
				   (find by minimizing overall runtime) */

struct Onedim_signal {          /* 1d function object, piecewise-linear
				   with overall exponential decay factor. */
  int nt;                       /* number of nodes */
  double t[MAX_TNODES];         /* times of nodes */
  double u[MAX_TNODES];         /* values at nodes */
  double mu_shift;              /* exponential decay rate */
};


