
/* cfg2inp.y, written Jon Stott.   Modified 5/29/02 Alex Barnett to
 * include multiple sources in .inp format, rendering it incompatible with
 * pre-May02 .inp format.
 */

%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <getopt.h>
#include "tMCimg.h"

double tmpmus, tmpmua, tmpn, tmpg;
double tmppos[3], tmpdir[3], tmprad, tmpna;
struct Config *config = NULL;
 
void init_config(void);
int  checkconfig(void);
void add_tissue(double mus, double mua, double g, double n);
void add_detector(double r[3], double d[3], double rad, double na);
void   add_source(double r[3], double d[3], double rad, double na);
int  dump_config(const struct Config *, const char *);
int  parseconfig(struct Config *, FILE *);

#define YYDEBUG 0

extern void parsefile(FILE *);
extern int yylex();

void yyerror(const char *s)
{
  extern int linenumber;
  extern char *yytext;

  fprintf(stderr, "%s\n", (s == NULL) ? "" : s);
  fprintf(stderr, "Error on line %d near text \"%s\"\n", 
          linenumber, (yytext == NULL) ? "" : yytext);

  return;
}
%}

%union {
  double      real;
  int         integer;
  const char *string;
  char        character;
  int        *ipointer;
  double     *rpointer;
}

%token NPHOTONS
%token SEED
%token SEGFILE
%token FREQUENCY
%token CLIGHT

%token STIME
%token NTIME
%token DTIME

%token TISSUETYPE

%token MUS
%token MUA
%token G
%token N

%token SOURCE
%token DETECTOR
%token SDPOS
%token SDDIR
%token SDRAD
%token SDNA

%token DX
%token DY
%token DZ

%token IMGX
%token IMGY
%token IMGZ

%token NXSTEP
%token NYSTEP
%token NZSTEP

%token <string>   FILENAME
%token <real>     REALNUMBER
%token <integer>  DECNUMBER

%type  <rpointer> vector
%type  <ipointer> range
%type  <real>     number
  
%%

input:	/* empty */
	| input statement
	| input error { exit(1); }

statement: option
	| tissue
        | source
	| detector

option:   NPHOTONS  DECNUMBER { config->NT        = $2; }
	| SEED      DECNUMBER { config->seed      = (long)$2; }
	| SEGFILE   FILENAME  { strncpy(config->segFile, $2, FILENMLEN); }
	| FREQUENCY number    { config->frequency = $2; }
	| CLIGHT    number    { config->v0        = $2; }
	| STIME     number    { config->Tmin      = $2; }
	| DTIME     number    { config->stepT     = $2; }
	| NTIME     number    { config->nTstep    = (int)$2; }
	| IMGX      range     { config->Ixmin=($2)[0]; config->Ixmax=($2)[1]; }
	| IMGY      range     { config->Iymin=($2)[0]; config->Iymax=($2)[1]; }
	| IMGZ      range     { config->Izmin=($2)[0]; config->Izmax=($2)[1]; }
	| DX        number    { config->xstep = $2; }
	| DY        number    { config->ystep = $2; }
	| DZ        number    { config->zstep = $2; }
	| NXSTEP    number    { config->nxstep=(int)$2; }
	| NYSTEP    number    { config->nystep=(int)$2; }
	| NZSTEP    number    { config->nzstep=(int)$2; }

tissue:   TISSUETYPE '{' { tmpmus = -1; tmpmua = -1; 
                           tmpn   = -1; tmpg   = -1; }
	| tissue tissue_prop 
	| tissue '}' { if ((tmpmus == -1) || (tmpmua == -1) ||
		           (tmpg   == -1) || (tmpn   == -1)) {
                         yyerror("Incompletely initialized tissue type\n");
			 YYABORT;
		       }

		       add_tissue(tmpmus, tmpmua, tmpg, tmpn); 
		     }

detector: DETECTOR '{' { tmppos[0] = tmppos[1] = tmppos[2] = -998;
                         tmpdir[0] = tmpdir[1] = tmpdir[2] = 0.0;
                         tmprad = -1.0;
                         tmpna  =  0.0; }
        | detector sd_prop
        | detector '}' { add_detector(tmppos, tmpdir, tmprad, tmpna); }

source:   SOURCE '{'   { tmppos[0] = tmppos[1] = tmppos[2] = -998;
                         tmpdir[0] = tmpdir[1] = tmpdir[2] = 0.0;
                         tmprad = 0.0;
                         tmpna  = 0.0; }
        | source sd_prop
        | source '}'   { add_source(tmppos, tmpdir, tmprad, tmpna); }

tissue_prop: MUS number { tmpmus = $2; }
	   | MUA number { tmpmua = $2; }
	   | G   number { tmpg   = $2; }
	   | N   number { tmpn   = $2; }

sd_prop:  SDRAD number  { tmprad = $2; }
	| SDNA  number  { tmpna  = $2; }
	| SDPOS vector  { memcpy(tmppos, (double *)$2, sizeof(tmppos)); 
                          free((double *)$2); }
	| SDDIR vector  { memcpy(tmpdir, (double *)$2, sizeof(tmpdir));
                          free((double *)$2); }

vector:  '[' number number number ']' {
                double *rvec = (double *)malloc(3*sizeof(double));
                assert(rvec != NULL);

                rvec[0] = $2; rvec[1] = $3; rvec[2] = $4;
                $$ = rvec;
              }

range: DECNUMBER DECNUMBER {
		static int drange[2];

		drange[0] = $1;  drange[1] = $2;
		$$ = drange;
	      }

number:   REALNUMBER { $$ =         $1; }
        | DECNUMBER  { $$ = (double)$1; /* type promotion */ }

%%

static struct option lopt[3] = {
  { "output", 1, NULL, 'o' },
  { "help",   0, NULL, 'h' },
  { NULL,     0, NULL,  0  }
};

static void usage(const char *s)
{
  printf("SYNTAX: %s [--output file] [--help]\n", s);

  return;
}

int main(int argc, char *argv[])
{
  extern char *optarg;
  extern int optind, optopt;
  char *output_file = "-";
  FILE *fp;  
  int n;

  while ((n = getopt_long(argc, argv, "o:h", lopt, NULL)) != -1)
    switch (n)
      {
      case 'o':
        output_file = strdup(optarg);
        break;
      case 'h':
        usage(argv[0]);
        return 0;
      case ':':
        printf("Error - missing argument\n");
        usage(argv[0]);
        return 1;
      case '?':  /* unrecognized character */
      default:
        printf("Error - unknown flag '%c'\n", optopt);
        usage(argv[0]);
        return 1;
      }

  if (optind == argc)
    {
      usage(argv[0]);
      return 1;
    }
  else
    do
      {
        struct Config cfg;

        if ((fp = fopen(argv[optind++], "rt")) == NULL) 
          {
            perror("Error opening file");
            printf("Can't open file \"%s\"\n", argv[1]);
            return 1;
          } 

        config = &cfg;

        memset(config, 0, sizeof(struct Config));

        init_config();

        if (parseconfig(&cfg, fp))
          return 1;

        fclose(fp);

        if (checkconfig() != 0)
          return 1;

        if (dump_config(config, output_file))
          return 1;

        config = NULL;
      }
    while (optind < argc);

  return 0;
}

int parseconfig(struct Config *p, FILE *fp)
{
  double dv = 0;

  assert(p  != NULL);
  assert(fp != NULL);

  config = p;

  init_config();                  /* Initialize the memory */

  parsefile(fp);                  /* Read and parse the .inp file */

  /* Compute the various derived quantities */

  config->frequency *= 1.0e6;     /* Convert from MHz */
  config->k0 = 2*PI*config->frequency / config->v0;

  config->Tmax = config->Tmin + config->nTstep * config->stepT;

  config->nIxstep = config->Ixmax - config->Ixmin + 1;
  config->nIystep = config->Iymax - config->Iymin + 1;
  config->nIzstep = config->Izmax - config->Izmin + 1;

  if (config->zstep > dv) dv = config->zstep;
  if (config->ystep > dv) dv = config->ystep;
  if (config->xstep > dv) dv = config->xstep;

  /* At least one length must be specified */
  if (dv <= 0) {
    printf("Voxel size <= 0");
    return 2;
  }

  config->dl = dv;

  /* Fill in any missing lengths (implicitly the largest one) */

  if (config->xstep <= 0.0) config->xstep = config->dl;
  if (config->ystep <= 0.0) config->ystep = config->dl;
  if (config->zstep <= 0.0) config->zstep = config->dl;

  /* Calculate min and max photon lengths from the min and max
     propagation times in vacuuo */

  config->Lmax  = config->Tmax  * config->v0;
  config->Lmin  = config->Tmin  * config->v0;
  config->stepL = config->stepT * config->v0;

  if (checkconfig() != 0)         /* Sanity check the final configuration */
    return 2;

  return 0;                       /* Done */
}

void init_config(void)
{
  int i;
#if (YYDEBUG)
  extern int yydebug;

  yydebug = 1;
#endif

  /* Set default configuration */

  config->v0 = C_VACUUM;

  config->xstep = config->ystep = config->zstep = -1;

  config->segFile[0] = '\0';

  config->Ntissue = 0;

  for (i = 0; i < MAXTISSUE; i++)
    {
      config->tmus[i] = MUS_AIR;
      config->tmua[i] = MUS_AIR;
      config->tg[i]   =   G_AIR;
      config->tn[i]   = IDX_AIR;
    }

  config->nDets = 0;

  for (i = 0; i < MAXDET; i++)
    {
      config->detLoc[i][0] = -1;
      config->detLoc[i][1] = -1;
      config->detLoc[i][2] = -1;

      config->detDir[i][0] = -1;
      config->detDir[i][1] = -1;
      config->detDir[i][2] = -1;

      config->detRad[i]    = -1;
      config->detNA[i]     = -1;
    }

  config->nSrcs = 0;

  for (i = 0; i < MAXDET; i++)
    {
      config->srcLoc[i][0] = -1;
      config->srcLoc[i][1] = -1;
      config->srcLoc[i][2] = -1;

      config->srcDir[i][0] = -1;
      config->srcDir[i][1] = -1;
      config->srcDir[i][2] = -1;

      config->srcRad[i]    = -1;
      config->srcNA[i]     = -1;
    }

  return;
}

void add_tissue(double mus, double mua, double tg, double tn)
{
  int n = ++config->Ntissue;

  config->tmus[n] = mus;
  config->tmua[n] = mua;
  config->tg[n]   = tg;
  config->tn[n]   = tn;

  return;
}

/* Add a new detector entry based on the supplied detector parameters */

void add_detector(double r[3], double d[3], double rad, double na)
{
  double nrm = sqrt(d[0]*d[0] + d[1]*d[1] + d[2]*d[2]);
  int      n = config->nDets;

  assert((n >= 0) && (n < MAXDET));

  if ((r[0] == -998) && (r[1] == -998) && (r[2] == -998)) 
    {
      yyerror("Detector position not defined\n");
      exit(1);
    }
  else
    {
      config->detLoc[n][0] = r[0];
      config->detLoc[n][1] = r[1];
      config->detLoc[n][2] = r[2];
    }

  if (nrm > 0)
    {
      /* Turn direction into a unit vector */

      config->detDir[n][0] = d[0]/nrm;
      config->detDir[n][1] = d[1]/nrm;
      config->detDir[n][2] = d[2]/nrm;
    }
  else
    {
      config->detDir[n][0] = 0;
      config->detDir[n][1] = 0;
      config->detDir[n][2] = 0;
    }

  if (rad <= 0)
    {
      yyerror("Detector radius non-positive or not defined");
      exit(1);
    }
  else
    config->detRad[n] = rad;

  if (na < 0)
    {
      yyerror("Detector NA must be non-negative");
      exit(1);
    }
  else if (na > PI/2)
    {
      yyerror("Source NA too large");
      exit(1);
    }
  else
    config->detNA[n] = rad;

  ++config->nDets;

  return;
}

void add_source(double r[3], double d[3], double rad, double na)
{
  double nrm = sqrt(d[0]*d[0] + d[1]*d[1] + d[2]*d[2]);
  int      n = config->nSrcs;

  assert((n >= 0) && (n < MAXDET));

  if ((r[0] == -998) && (r[1] == -998) && (r[2] == -998)) 
    {
      yyerror("Source position not defined\n");
      exit(1);
    }
  else
    {
      config->srcLoc[n][0] = r[0];
      config->srcLoc[n][1] = r[1];
      config->srcLoc[n][2] = r[2];
    }

  if (nrm > 0)
    {
      /* Turn direction into a unit vector */

      config->srcDir[n][0] = d[0]/nrm;
      config->srcDir[n][1] = d[1]/nrm;
      config->srcDir[n][2] = d[2]/nrm;
    }
  else
    {
      yyerror("Source direction zeros or not defined\n");
      exit(1);
    }

  if (rad < 0)
    {
      yyerror("Source radius negative or not defined");
      exit(1);
    }
  else
    config->srcRad[n] = rad;

  if (na < 0)
    {
      yyerror("Source NA must be non-negative");
      exit(1);
    }
  else if (na > PI/2)
    {
      yyerror("Source NA too large");
      exit(1);
    }
  else
    config->srcNA[n] = rad;

  ++config->nSrcs;

  return;
}

#define PARSE_ERROR(s) { printf("Input error: %s\n", (s)); return 1; }
#define PARSE_WARN(s)  { printf("Warning: %s\n", (s)); }

int checkconfig(void)
{
  int i;

  if (config->NT < 1) 
    PARSE_ERROR("Total number of photons must be a postitive number");

  if ((config->frequency > 1.0e10) || (config->frequency < -1.0e10))
    printf("Warning: abnormally large frequency reqested (%f MHz)\n",
           config->frequency * 1e-6);

  if ((config->nTstep <= 0) || (config->stepT <= 0))
    PARSE_ERROR("Non-causal time steps not allowed");

  if ((config->xstep <= 0) || (config->ystep <= 0) || (config->zstep <= 0))
    PARSE_ERROR("xstep, ystep, and zstep must be strictly positive");

  if ((config->xstep != config->ystep) || (config->xstep != config->zstep))
    PARSE_ERROR("xstep == ystep == zstep required in this version");

  if ((config->Ixmin < 0) || (config->nIxstep > config->nxstep))
    PARSE_ERROR("Illegal image size (X dimension)");

  if ((config->Iymin < 0) || (config->nIystep > config->nystep))
    PARSE_ERROR("Illegal image size (Y dimension)");

  if ((config->Izmin < 0) || (config->nIzstep > config->nzstep))
    PARSE_ERROR("Illegal image size (Z dimension)");

  if (config->Ntissue >= MAXTISSUE)
    PARSE_ERROR("Too many tissue types");

  for (i = 1; i <= config->Ntissue; i++)
    {
      if (config->tmus[i] <= 0)
        PARSE_ERROR("Non-physical scattering coefficient");
      if (config->tmua[i] <  0)
        PARSE_ERROR("Non-physical absorption coefficient");
      if ((config->tg[i] < -0.5) || (config->tg[i] > 1.0))
        PARSE_ERROR("Non-physical anisotropy coefficient");
      if (config->tn[i] <= 0)
        PARSE_ERROR("Non-physical index of refraction");

      if (config->tn[i] < 1.0)
        PARSE_WARN(
          "Index of refraction < 1, I hope you know what you're doing!\n");
    }

  if (config->nSrcs <= 0)
    PARSE_ERROR("No Source Specified");

  for (i = 0; i < config->nDets; i++)
    if ((config->detLoc[i][0] < 0) || 
        (config->detLoc[i][1] < 0) || 
        (config->detLoc[i][2] < 0) || 
        (config->detLoc[i][0] >= config->nxstep*config->xstep) ||
        (config->detLoc[i][1] >= config->nystep*config->ystep) ||
        (config->detLoc[i][2] >= config->nzstep*config->zstep))
      PARSE_WARN("Detector located outside of image volume");

  return 0;
}

int dump_config(const struct Config *config, const char *output)
{
  FILE *fp;
  int i;

  fp = stdout;

  if ((output[0] != '-') || (output[1] != '\0'))
    {
      fp = fopen(output, "w");

      if (fp == NULL)
	{
	  perror("Can't write output file");
	  return 1;
	}
    }

  fprintf(fp, "%d\n", config->NT);
  fprintf(fp, "%ld\n", config->seed);
  fprintf(fp, "%f\n", config->frequency);
  /* removed 8 numbers from here specifying single source */
  fprintf(fp, "%e %e %d\n", config->Tmin, config->stepT, config->nTstep);
  fprintf(fp, "%s\n", config->segFile);

  /* Add 1 for legacy support */
  fprintf(fp, "%f %d %d %d\n",
	  config->xstep, config->nxstep, config->Ixmin + 1,
	  config->Ixmax + 1);
  fprintf(fp, "%f %d %d %d\n", config->ystep, config->nystep,
	  config->Iymin + 1, config->Iymax + 1);
  fprintf(fp, "%f %d %d %d\n", config->zstep, config->nzstep,
	  config->Izmin + 1, config->Izmax + 1);

  fprintf(fp, "%d\n", config->Ntissue);
  for (i = 1; i <= config->Ntissue; i++)
    fprintf(fp, "%f %f %f %f\n", config->tmus[i],
	    config->tg[i], config->tmua[i], config->tn[i]);

  /* added following source loc, rad (only # 0) (not NA or dir) */
  fprintf(fp, "%d %f\n", config->nSrcs, config->srcRad[0]);

  for (i = 0; i < config->nSrcs; i++)
    fprintf(fp, "%f %f %f\n",
	    config->srcLoc[i][0], config->srcLoc[i][1], config->srcLoc[i][2]);
  /* end of added part */

  /* note only det rad of 0 is output ... */
  fprintf(fp, "%d %f\n", config->nDets, config->detRad[0]);

  for (i = 0; i < config->nDets; i++)
    fprintf(fp, "%f %f %f\n",
	    config->detLoc[i][0], config->detLoc[i][1], config->detLoc[i][2]);

  if (fp != stdout)
    fclose(fp);

  return 0;
}
