% function [y, y_uncorr_sigma] = forward(x, fwd)
%
% compute y (col vec of signals) given param col vec x using forward model.
% y_uncorr_sigma is standard deviation of uncorrelated noise from
% fwd model (only non-zero for MC).
%
% fwd is struct defining forward model:
%
% fwd.model:
%   0 : Dummy forward model (very fast).
%                 The dummy has correct size & S-D structure of
%                 signal y vector.
%                 It also is ill-conditioned in a similar way to
%                 the layered inference problem.
%        fwd.M = desired output vector size.
%
%   1 : Homogeneous, semi-infinite.
%        fwd.R = vector of S-D radii; its length defines N_d (N_s = 1).
%
%   2 : General layered slab (slab.mexglx)
%
%   [3 : FDTD without gradient info (tFDimg).]
%        fwd.geom = filename of .geom file, defining geom, N_s, N_d.
%
%   [4 : Monte-Carlo photon migration (tMCimg).]
%
% fwd.tmin, .dtsam, .ntsam, give the time window and sample spacing.
%           Samples are returned as value in middle of each
%           sampling timegate.
%
% fwd.precon_x : column vector of factors by which incoming x has been scaled.
%                (note that this is in reduced x space)
%
% barnett 4/24/02 added fix_x 6/3/02, added dummy 6/12/02
% 7/9/02 added node_specific filespace use (fwd.work_name), 7/19/02 y_forward_sigma added.
% 7/29/02 dummy improved, 8/1/02 fwd.len_sc added.
% totally rewritten 1/25/03, removing fix_x feature.

function [y, y_uncorr_sigma] = forward(x, fwd)

global A_DUMMY;   % used by dummy, for speed.

diary off; diary on;   % hack to get diary to update !

switch fwd.model
        
 case 0
    % use dummy.....................................................................*
    
    M = fwd.sys.nsrc * fwd.sys.ndet * fwd.sys.nt;
    N = numel(x);
    if isempty(A_DUMMY)
        % A Incarnation 1 was sinusoids:
        % scale = 1.5; % 1.0 (for m_decay=M/8) gives close to singular in x-space; larger nicer.
        % A = (1 + cos( (scale*pi/M) * reshape(kron(1:N,1:M),[M N])))/2; % (1+sin)/2
        %
        % A Incarnation 2 was nonoverlapping top-hat functions...
        % A = [kron(eye(N),ones(floor(M/N),1));kron(ones(rem(M,N),1),[zeros(1,N-1),1])];
        %
        % New pattern where x_n's control (overlapping /\ shaped) regions of m:
        p = floor(M/N);  % chunk size of division of M
        a = zeros(M+p, 1);   % size of col vec to wrap around.
        a(1:2*p,1) = [1:p, p:-1:1]' / p;   % add the /\ function - too much overlap bad.
        aa = repmat(a,[N,1]);
        A = reshape(aa(1:N*M),[M,N]);      % wrap around inside A.
        % figure; imagesc(A);
         
        % Decay of A in M and in N directions...
        n_decay = 1.5; % 3.5/N,  decay of effects of params x_n : makes ill-posed.
        m_decay = M/11; % use to simulate S-D signal dynamic range
        A = A .* reshape(kron(exp(-(1:N)/n_decay),exp(-(1:M)/m_decay)),[M N]);

        A_DUMMY = A;
    else
        A = A_DUMMY;
    end
    % rearrange components of x to be closer to mua,musp layer structure...
    rx(N:-2:2) = x(N/2 + 1:N);
    rx(N-1:-2:1) = x(1:N/2);
    x = rx';
    % non-lin transform of x
    p = 1.0; % tightness of nonlin bend - can't hit 0.
    q = 0.0; % amount of linear from 0 to 1.
    tx = q*x + (1-q)*(1 + (1-exp(-p*(x-1)))/p);
    
    x1 = 0.8*ones(N,1);  % offset cannot get too large or y goes negative for usual x.
    y = 600*A * x1 + 5000*A*(tx-x1); % make y col vec from Ax, with scale & offset.
    
    y_forward_sigma = zeros(size(y));

    
else
    % use correct forward model.....................................................*
    
    % if needed, rescale x back to original units... (undo preconditioning)
    if isfield(fwd,'precon_x')
        x = x ./ fwd.precon_x;
    end
    
    % if needed, expand x back to its full size using info in fwd...
    if isfield(fwd,'fix_x')
        No = numel(x) + numel(fwd.fix_x.dims);
        flag = zeros(No,1);
        flag(fwd.fix_x.dims) = 1;
        free = find(flag==0);
        x_full = zeros(No,1);
        x_full(free) = x;
        x_full(fwd.fix_x.dims) = fwd.fix_x.vals;
        x = x_full;
    end
    
        % FDTD no gradient info, using external C program tFDimg.
        
        % ------- Write config file from x param vector -------------------------------
        % read cfgnot...
        cfgnot = sprintf('%s.cfgnot', fwd.name);
        if ~exist(cfgnot, 'file')
            disp(sprintf('forward: file %s not found!', cfgnot));
            exit(1);
        end
        fid = fopen(cfgnot, 'r');
        txt = fread(fid)';
        % append tissue property text lines... choose refractive index, and g, here!
        n_tiss = length(x)/2;
        for t=1:n_tiss
            tissue_line = sprintf('tissue { mua %.12f mus %.12f n 1 g 0 }\n',x(t),x(t+n_tiss));
            txt = [txt, tissue_line];
        end
        fclose(fid);
        % choose workspace file head name...
        if isfield(fwd,'work_name')
            head = fwd.work_name;
        else
            head = fwd.name;
        end
         % write cfg...
        fid = fopen([head, '.cfg'], 'w');
        fwrite(fid, txt');
        fclose(fid);
        
        % call FDTD (silently) and load -------- note tFDimg options! -----------
        dt = 2.6; % for musp lower bound 0.4 per mm, 2mm voxel size.
        if isfield(fwd,'len_sc')
            dt = dt * (fwd.len_sc)^2;
        end
        if isfield(fwd,'forobin')
            cmd = sprintf('tFDimg ernt %f %s > %s.FD.log', dt, head, head);
        else
            cmd = sprintf('tFDimg ent %f %s > %s.FD.log', dt, head, head);
        end
        system(cmd);
        fid = fopen([head, '.sig'], 'r');
        n = fwd.sys.nsrc * fwd.sys.ndet * fwd.sys.nt;
        dt = fscanf(fid, '# %f', 1);    % read in main timestep in ps
        [yfd, count] = fscanf(fid, '%f', n);
        fclose(fid);
        if count~=n
            disp('forward: read wrong number of signals from .sig file!');
        end
        
        y = yfd;
        if isfield(fwd,'len_sc')    % correct overall amplitude for voxel size
            y = y / (fwd.len_sc)^3;
        end
        y_forward_sigma = zeros(size(y));
            
            
    case 1
        % Monte Carlo, no gradient info, using external C program tMCimg.
        
        n_tiss = length(x)/2;

        % ------- Write config file from x param vector -------------------------------
        
        % call tMCimg (silently)... only if needed (all .his files present)
        % (really ought to check if .cfg had same musp as x vector - too fancy).
        % choose workspace file head name...
        if isfield(fwd,'work_name')
            head = fwd.work_name;
        else
            head = fwd.name;
        end
        if fwd.sys.nsrc > 1
            need_tmcimg = 0;
            for s=1:fwd.sys.nsrc
                hisname = sprintf('%s.his%d', head, s-1);
                disp(['looking for ',hisname]);
                need_tmcimg = need_tmcimg | ~exist(hisname,'file');
            end
        else
            need_tmcimg = ~exist(sprintf('%s.his', head),'file');
        end
        if need_tmcimg
            disp('tMCimg needed to be run, since his files not all present.');
            % read cfgnot...
            cfgnot = sprintf('%s.cfgnot', fwd.name);
            if ~exist(cfgnot, 'file')
                disp(sprintf('forward: file %s not found!', cfgnot));
                exit(1);
            end
            fid = fopen(cfgnot, 'r');
            txt = fread(fid)';
            % append tissue property text lines... choose refractive index, and g, here!
            for t=1:n_tiss
                tissue_line = sprintf('tissue { mua %.12f mus %.12f n 1 g 0 }\n',x(t),x(t+n_tiss));
                txt = [txt, tissue_line];
            end
            fclose(fid);
            % write cfg...
            fid = fopen([head, '.cfg'], 'w');
            fwrite(fid, txt');
            fclose(fid);
            
            cmd = sprintf('tMCimg -n %s > %s.log', head, head);
            disp(['not executing for safety: ', cmd]);
            %system(cmd); since erases .his* files.
        end
        
        % extract just the mua part of x...
        expt_mua = x(1:n_tiss);
        
        sys = load_cfg(head);
        % loop over sources, loading in each ----------------------------------------
        for s=1:sys.nsrc
            if sys.nsrc > 1
                [his, status] = load_his(head, sys, s); % his file for src #
            else
                [his, status] = load_his(head, sys);
            end
            [ydt, mc_var_ydt] = absorb_muaderiv(sys, his, expt_mua, 0); % don't want derivs
            y(s,:,:) = ydt;
            y_forward_sigma(s,:,:) = sqrt(mc_var_ydt);
        end
        y = reshape(permute(y,[3 2 1]),[numel(y),1]);
        y_forward_sigma = reshape(permute(y_forward_sigma,[3 2 1]),[numel(y),1]);

        if numel(y)~=fwd.sys.nsrc * fwd.sys.ndet * fwd.sys.nt
            disp('forward: MC, fwd.sys doesnt agree with size of y vector!');
        end

        
    otherwise
        disp(sprintf('forward: fwd.model %d not implemented',fwd.model));
        exit(1);
    end
    
end

% unpack into (s,d,t) ordering, then choose subsets in s,d,t axes...
ysdt = permute(reshape(y, fwd.sys.nt, fwd.sys.ndet, fwd.sys.nsrc), [3 2 1]);
ysdt_fs = permute(reshape(y_forward_sigma, fwd.sys.nt, fwd.sys.ndet, fwd.sys.nsrc), [3 2 1]);

detsums(1,:) = sum(sum(ysdt,1),3);
if isfield(fwd,'choose')
    if isfield(fwd.choose,'s')
        ysdt = ysdt(fwd.choose.s,:,:);
        ysdt_fs = ysdt_fs(fwd.choose.s,:,:);
    end
    if isfield(fwd.choose,'d')
        ysdt = ysdt(:,fwd.choose.d,:);
        ysdt_fs = ysdt_fs(:,fwd.choose.d,:);
    end
    if isfield(fwd.choose,'t')
        ysdt = ysdt(:,:,fwd.choose.t);
        ysdt_fs = ysdt_fs(:,:,fwd.choose.t);
    end
end
detsums(2,:) = sum(sum(ysdt,1),3);

% pack back to original vector ordering...
y = reshape(permute(ysdt,[3 2 1]),[numel(ysdt),1]);
y_forward_sigma = reshape(permute(ysdt_fs,[3 2 1]),[numel(ysdt),1]);

% end
