% function [I, Ierr] = physics_of_tissue(z_f, phys)
%
% compute the surface intensity expectation value
% at detectors, due to sources, given
% optical and geometric parameters passed in through z_f (= z_full).
% phys may contain other non-variable geometric parameters, or info
% on how to interpret z_f.
% This is the bare forward physical model,
% before calibration, convolution with IRF, binning into timegates, etc.
%
% outputs.
%   I              : col vec of surface intensity signals, length
%                    Ns.Nd.ntsam.
%   Ierr           : struct returning things about numerical/physics error
%         Ierr.var : is variance of uncorrelated noise from
%                    model (only non-zero for MC, or maybe early times in
%                    diffusion models?).
%
% inputs.
%   z_f            : physical parameters of tissue, in physical
%                    units.
%                    How these are interpreted depends on the model.
%   phys           : struct specifying model and containing all
%                    remaining geometry, filename, parameter info.
% Things used from phys are...
%
% phys.model : physical model to be used.
%
%   model = 0 : Dummy model, very fast.
%                 The dummy has correct size & S-D structure of
%                 signal f vector.
%                 It also is ill-conditioned in a similar way to
%                 the layered inference problem.
%                 phys.M = desired output vector size.
%
%           1 : Homogeneous, semi-infinite.
%                 z_f = [mua musp R], where R = vector of S-D dists in mm.
%                          R's length defines Nd (Ns = 1).
%                 mua = absorption coeff (in per mm)
%                 musp = scattering coeff (in per mm)
%                 phys.Nl, phys.Ndz, phys.Nd, phys.Ns needed or checked.
%
%           2 : General layered slab (slab.mexglx; see slab.m)
%                 z_f = [mua musp dz R], where
%                 mua = vector of Nl absorption coeffs (in per mm)
%                 musp = vector of Nl scattering coeffs (in per mm)
%                 R = vector of S-D dists in mm.
%                          its length defines Nd (Ns = 1).
%                 dz = vector of Ndz layer thicknesses in mm (see
%                 slab.m).
%                 phys.Nl, phys.Ndz, phys.Nd, phys.Ns needed or checked.
%
%          3 : FDTD without gradient info (tFDimg).
%                 phys.geom = full filename of geom file, defining geom, Ns, Nd.
%                 z_f = [mua musp], where
%                 mua = vector of Nl absorption coeffs (in per mm)
%                 musp = vector of Nl scattering coeffs (in per mm)
%                 These do not include CSF, which is inserted
%                 between layers 2,3.
%                 phys.Nl, phys.tmin, phys.ntsam, phys.dtsam all needed.
%                 phys.work_name, phys.len_sc optional.
%
%          [4 : Monte-Carlo photon migration (tMCimg).]
%
% phys.tmin, .dtsam, .ntsam :  give the time window and sample spacing.
%                              I samples are returned as signal in
%                              middle of each
%                              sampling timegate (models 1,2), or
%                              integrated over timegate (models
%                              3,4, normalization pending).
%
% phys.Ns, .Nd : are checked for consistency against model.
%
% phys.bcs : boundary conditions, 0=ZBC(dirichlet), 1=EBC(close to robin).
%
% phys.n : refractive index of all tissues.
%
% phys.musp_min, musp_max : min(max) mu_s', is used in slab.c to prevent this
%                 changing when finite differencing in
%                 meas_jacobean used. Bit of a hack.
%
% phys.norm : if 1, normalization hack so max value is 1.
%
% barnett 1/27/03, musp_min hack 5/24/03
% 12/16/04

function [I, Ierr] = physics_of_tissue(z_f, phys)

global A_DUMMY;   % used by dummy, for speed.

%diary off; diary on;   % hack to get diary to update ! saves sdtout

switch phys.model
        
 case 0
  % use dummy....000000000000000000000000000000000000000000000000000000000000
  
  M = phys.M;
  N = length(z_f);  % currently just fake dependence on everything

  if isempty(A_DUMMY)
    % A Incarnation 1 was sinusoids:
    % scale = 1.5; % 1.0 (for m_decay=M/8) gives close to singular in x-space; larger nicer.
    % A = (1 + cos( (scale*pi/M) * reshape(kron(1:N,1:M),[M N])))/2; % (1+sin)/2
    %
    % A Incarnation 2 was nonoverlapping top-hat functions...
    % A = [kron(eye(N),ones(floor(M/N),1));kron(ones(rem(M,N),1),[zeros(1,N-1),1])];
    %
    % New pattern where x_n's control (overlapping /\ shaped) regions of m:
    p = floor(M/N);  % chunk size of division of M
    a = zeros(M+p, 1);   % size of col vec to wrap around.
    a(1:2*p,1) = [1:p, p:-1:1]' / p;   % add the /\ function - too much overlap bad.
    aa = repmat(a,[N,1]);
    A = reshape(aa(1:N*M),[M,N]);      % wrap around inside A.
    % figure; imagesc(A);
				       
    % Decay of A in M and in N directions...
    n_decay = 1.5; % 3.5/N,  decay of effects of params x_n : makes ill-posed.
    m_decay = M/11; % use to simulate S-D signal dynamic range
    A = A .* reshape(kron(exp(-(1:N)/n_decay),exp(-(1:M)/m_decay)),[M N]);
    
    A_DUMMY = A;
  else
    A = A_DUMMY;
  end
  % rearrange components of x to be closer to mua,musp layer structure...
  rx(N:-2:2) = x(N/2 + 1:N);
  rx(N-1:-2:1) = x(1:N/2);
  x = rx';
  % non-lin transform of x
  p = 1.0; % tightness of nonlin bend - can't hit 0.
  q = 0.0; % amount of linear from 0 to 1.
  tx = q*x + (1-q)*(1 + (1-exp(-p*(x-1)))/p);
  
  x1 = 0.8*ones(N,1);  % offset cannot get too large or y goes negative for usual x.
  I = 600*A * x1 + 5000*A*(tx-x1); % make y col vec from Ax, with scale & offset.
  
  Ierr.var = zeros(size(I));
  
  
 case {1,2}
  % slabs...11111111111111111111111111111111111111111111111111111111111111
  %         22222222222222222222222222222222222222222222222222222222222222
  
  % intensity signal evaluation times come in MIDDLE of sample windows...
  tmin = phys.tmin + phys.dtsam/2;
  tmax = tmin + phys.dtsam*(phys.ntsam - 0.5);
  
  if phys.Ns~=1
    disp('models 1,2 must have Ns=1!');
  end
  
  if phys.model==1
    if length(z_f)~=2+phys.Nd
      disp('model 1: z_f has wrong length!');
    end
    extras = [0,0,0,0,0,0,0,0,phys.bcs];  % last number gives EBC or ZBC
    I = slab_homog_inf(tmin, tmax, phys.dtsam, z_f(3:2+phys.Nd), ...
		       z_f(1), z_f(2), [], phys.n, extras);
  else
    if length(z_f)~=2*phys.Nl+phys.Ndz+phys.Nd
      disp('model 2: z_f has wrong length!');
    end
    % unpack z_f into mua, musp, dz, and R...
    R = z_f(2*phys.Nl+phys.Ndz+1:2*phys.Nl+phys.Ndz+phys.Nd);

    % set default slab.c params here: :::::::::::::::::::::::::::::::::::::
    verb = 0;   % 129 for basics + first mode
    if isfield(phys,'musp_min')
      extras = [1e-6,-0.25,-0.05,1e-6,1e-12,-1,verb,1,phys.bcs,300,0.1,...
		phys.musp_min, phys.musp_max];
    else
      extras = [1e-6,-0.25,-0.05,1e-6,1e-12,-1,verb,1,phys.bcs];  % CNvar
    end
    % :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
    
 %   I = slab(tmin, tmax, phys.dtsam, R, z_f(1:phys.Nl), ...
	%     z_f(phys.Nl+1:2*phys.Nl), z_f(2*phys.Nl+1:2*phys.Nl+phys.Ndz));
    I = slab(tmin, tmax, phys.dtsam, R, z_f(1:phys.Nl), ...
	     z_f(phys.Nl+1:2*phys.Nl), z_f(2*phys.Nl+1:2*phys.Nl+phys.Ndz),...
	     phys.n, extras);
  end
  
  I = reshape(I,[numel(I) 1]); % puts detectors slower, time samples faster

  Ierr.var = zeros(size(I));
  
  
 case 3
  % tFDimg...3333333333333333333333333333333333333333333333333333333333333
  
  % read geom file (compatible with cfgnot file)...
  fid = fopen(phys.name, 'r');
  if fid==-1
    disp(sprintf('physics_of_tissue, model 3: file %s not found!', phys.name));
    exit(1);
  end
  txt = fread(fid)';
  
  % append timegate info... time units convert ps to seconds.
  tgate_line = sprintf('\nstart_time=%g\ngate_width=%g\nngate=%d\n\n', ...
		       phys.tmin*1e-12, phys.dtsam*1e-12, phys.ntsam);
  txt = [txt, tgate_line];
  
  % insert CSF into tissue properties...
  mua = [z_f(1:2) phys.CSF.mua z_f(3:phys.Nl)];
  musp = [z_f(1+phys.Nl:2+phys.Nl) phys.CSF.musp z_f(3+phys.Nl:2*phys.Nl)];
  phys.Nl = phys.Nl + 1; % local change only, for CSF
  
  % append tissue property text lines... g=0 (mu_s=mu_s'), diffusion approx!
  for t=1:phys.Nl
    tissue_line = sprintf('tissue { mua %.12f mus %.12f n %f g 0 }\n',...
			  mua(t), musp(t), phys.n);
    txt = [txt, tissue_line];
  end
  fclose(fid);
  
  % choose workspace file head name... (only needed for parallel nodes)
  if isfield(phys,'work_name')
    head = phys.work_name;
  else
    head = sprintf('%s.work', phys.name);
  end
  
  % write working cfg file...
  fid = fopen([head, '.cfg'], 'w');
  fwrite(fid, txt');
  fclose(fid);
        
  % call FDTD (silently) and load -------- note tFDimg options! -----------
  dt = 2.6; % for musp lower bound 0.4 per mm, 2mm voxel size (standard)
  if isfield(phys,'len_sc')
    dt = dt * (phys.len_sc)^2;
  end
  if phys.bcs==1
    cmd = sprintf('tFDimg ernt %f %s > %s.FD.log', dt, head, head); % EBC (robin)
  else
    cmd = sprintf('tFDimg ent %f %s > %s.FD.log', dt, head, head); % ZBC (dirich.)
  end
  disp(cmd);
  system(cmd);
  disp('done.');
  %  disp('physics_of_tissue, model 3: tFDimg not executed correctly!');
  %end
  
  % read back the resulting data...
  fid = fopen([head, '.sig'], 'r');
  n = phys.Ns*phys.Nd*phys.ntsam;  % signal length
  dt = fscanf(fid, '# %f', 1);    % read in main timestep in ps - check it??
  [I, count] = fscanf(fid, '%f', n); % read in signal
  fclose(fid);
   if count~=n
    disp('forward: read wrong number of signals from .sig file!');
  end
  I = reshape(I,[numel(I) 1]); % ordering: srcs slowest, time samples fastest
  if isfield(phys,'len_sc') 
    I = I / (phys.len_sc)^3;   % correct overall amplitude for voxel size
  end
  Ierr.var = zeros(size(I));  % dummy for now

  
 case 4
  % tMCimg...4444444444444444444444444444444444444444444444444444444444444
  
  
  
end

% overall normalization hack
if isfield(phys, 'norm')
  if phys.norm==1
    I = I / max(max(I));
  end
end

% end
