/*********************************************************************
 *
 * tFDimg      :      3d finite-difference time-domain photon diffusion.
 *                    Based on tMCimg.
 * output files:      .sig : multiple-source (s,d,t) signal file.
 *
 *        compile options: RF_SUPPORT (not implemented).
 *
 * barnett 3/12/02
 * 3/14/02      coded ADI, with no air-box-edge contact allowed.
 * 3/23/02      basic forward-Euler method, with sparse matrix.
 * 4/18/02      integrated with Jon's new Config struct with flags in it.
 * 4/29/02  first-order Robin BCs for tissue-air boundaries (see toyFDTD).
 * 5/3/02 v.2: continuously-chosen dt, correct interpolation into timegates.
 * 5/16/02 smoothed out kink when new minimum musp takes over in choosing dt.
 * 7/25/02 v.3: new S&D models for better rough surface, cut some 3d F bits.
 * 5/28/03 extract_sig now handles dt(FD) > dt(timegates). (2pt does not).
 * 11/17/04 Removed Jon's code for FPU exception settings, fixed memory leaks.
 *          Option for single-precision calc (still double for 2pt, system)
 *
 *********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include "tMCimg.h"

/* floating point exception control... (glibc 2.2) */
#include <fenv.h>

/* Set precision: use double or float for timestepping and sparse vectors
   (everything else stays double)
   barnett 11/17/04 */
#define _SINGLEPREC


#ifdef _SINGLEPREC
#define real float
#else
#define real double
#endif


/* Jon's... */
#ifdef _WIN32
/* For some reason, windows adds an extra underscore */
#define snprintf _snprintf
#else
# include <unistd.h>
#endif


/*********************************************************************
 Function prototypes
*/

/* lapack */
extern void dgttrs_(char *, int *, int *, double *, double *, double *, \
		double *, int *, double *, int *, int *);
extern void dgttrf_(int *, double *, double *, double *, double *, int *,\
		    int *);
/* Jon's parse code */
extern int parseconfig(struct Config *, FILE *);

/* in this file */
int loadconfig(struct Config *p, char *filehead);
int loadsegments(const struct Config *p, TISSUE ****tt);
void fpusetup(void);
int categorize_node(TISSUE ***tt, int x, int y, int z, int nx, int ny, int nz);
int touch_material_cyc(TISSUE ***tissueType, int d, int a, int b, int c, \
		       int na, int nb);
double ***alloc_3d_array(int, int, int);
int ***alloc_3d_int_array(int, int, int);
void icyc(int d, int *a, int *b, int *c, int x, int y, int z);
TISSUE Tgetcyc(TISSUE ***T, int d, int a, int b, int c);
real dgetcyc(real ***A, int d, int a, int b, int c);
double get_edge_prop_cyc(TISSUE ***tissueType, double *p, int d, int a, \
			 int b, int c, int na, int nb);
int extract_sig(const struct Config* sp, real* F, int nn, \
		double** sig_of_s, double t_beg, double dt, \
		int* ncdv, int** cdv, real** dv);
int save_sigbinary(const struct Config *, double ***, char *);
int save_sig(const struct Config *system, double ***sig, double courant,
	     char *filehead);
int save2pt(const char *, const double *, const double *, int, const char *);

void rescale_system(struct Config *system);
double kappa(const struct Config *, int);
void record_fluence_in_2pt(const struct Config *, int, real *, int ***,
			   double *, double *);
void sparse_mv_dir(real **A, int **cA, int *ncA, real *F, real *G, int nn,\
	   real scalar);
void sparse_mv(real **A, int **cA, int *ncA, real *F, real *G, int nn,\
	   real scalar);
double musp(const struct Config *system, int ti);
int sparse_optode_vec(int *cov, real *ov, double *loc, double *dir,
		      double rad, const struct Config *system,
		      TISSUE ***tissueType, int ***ni,
		      int nn, double optode_depth, int grad_flag);
real sparse_dense_inner(int nca, int* ca, real *a, int nb, real *b);
void densify(int nca, int* ca, real *a, int nb, real *b);
void axpy3(double a, double *x, double *y);
double nrm3(double *y);
void copy3(double *x, double *y);
void cross3(double *a, double *b, double *c);
TISSUE tissue_penetrate(double *r, double *move, double *launch, double *dir,
			double depth, const struct Config *system,
			TISSUE ***tissueType);
void beamlet_dump(const struct Config *sp, real *ov_dense, int ***ni,\
		  double *r, double depth_correct_fac);


/* methods, BCs ... */
#define FORWARD_EULER 0
#define ADI           1    /* obsolete, for now */

/* tissue-air BCs... */
#define DIRICHLET      0
#define CRUDE_ROBIN    1   /* obsolete */
#define CRUDE_FO_ROBIN 2

/* tissue-box-face BCs... */
#define BOX_CANT_TOUCH    0
#define BOX_CRUDE_ROBIN   1
#define BOX_PML           2  /* never coded since not worth it */

/* max number of non-zero elements of Euler A matrix, per row... */
#define NNZR   7
/* node flags */
#define NOT_A_DOF  -1
#define NOT_IN_BOX -2

/* node categories */
#define NODE_AIR    0
#define NODE_BDRY   1
#define NODE_TIS    2

/* Macros to find node numbers given xyz index, with array bound checking...
 * assumes that xyz is a valid node itself (ie in the box). */
/* finds for node one left in direction d... */
#define GET_N_M(x, y, z, nx, ny, nz, ni, d) \
(d==0 ? (z>0 ? ni[x][y][z-1] : NOT_IN_BOX) : \
(d==1 ? (x>0 ? ni[x-1][y][z] : NOT_IN_BOX) : \
(y>0 ? ni[x][y-1][z] : NOT_IN_BOX)))
/* finds for node one right in direction d... */
#define GET_N_P(x, y, z, nx, ny, nz, ni, d) \
(d==0 ? (z<nz ? ni[x][y][z+1] : NOT_IN_BOX) : \
(d==1 ? (x<nx ? ni[x+1][y][z] : NOT_IN_BOX) : \
(y<ny ? ni[x][y+1][z] : NOT_IN_BOX)))





/* STANDARD METHOD PARAMETERS: also see individual routines ............ */

/* interpolation weight per unit time, in extract_sig... */
#define TIME_WEIGHT 1e12

/* standard optode depth in voxels - can override in cmd line */
#define OPTODE_DEPTH 1.5

/* standard courant number, just below stability limit of 1/6 for Euler */
#define COURANT 0.14

/* max number of nodes involved in each S,D sparse vector... */
#define MAX_NCOV 1000



/*********************************************************************
 MAIN
 *********************************************************************/

int main(int argc, char *argv[])
{
  struct Config system;		  /* System configuration, constructed from
				   * the .cfg file */
  double *II, *JJ;		  /* 2pt TIME-FLUENCE - double, for 2pt out */
  double ***sig;                           /* output signal by (s,d,t) */
  double dt, max_kv, kv;
  double syska[MAXTISSUE];                /* system's tissue type kappas */
  double ka_m, ka_p, mua_m, mua_p, inv_n, A_refl, t, this_dt;
  int a, b, c, d, i, j, k, n, s, ti, na, nb, nc, x, y, z, tg;
  int nx, ny, nz, nmax;            /* system size: # voxels in each dim */
  char *opts, *filehead;
  char twoptfile[FILENMLEN];      /* filename for 2pt */
  TISSUE ***tissueType;		  /* storage of segmentation input */

  /* Euler stuff... */
  int ***ni, *xni, *yni, *zni;    /* node index, loc of given node index */
  real *F, *G, *dummyv;                /* FLUENCE as column vector */
  real **A;                     /* sparse A matrix values... */
  int **cA;                       /* columns locations of nz els of A */
  int *ncA;                     /* number of nonzero col values for a row */
  double Euler_prefac, time_step = 0.0, d_ext, courant;
  int nn;
  int *ncdv, **cdv, ncsv, *csv;                /* sparse S,D vectors... */
  real **dv, *sv;
  double optode_depth = OPTODE_DEPTH;

  /* cmd line defaults... */
  int vary_courant = 0;
  int method = FORWARD_EULER;
  int bcs = DIRICHLET;
  int box_bcs = BOX_CANT_TOUCH;
  int grad_optode = 1;

  /* Set up FPU exception control... (see man fesetexceptflag) */
  feenableexcept(FE_DIVBYZERO | FE_INVALID | FE_OVERFLOW);

  /* Initialize the configuration structure's memory */
  memset(&system, 0, sizeof(struct Config));

  /* cmd line defaults... */
  system.flags.gen_twopt = 1;          /* default: do write 2pt file */


  /* --------------------------- READ COMMAND LINE --------------------- */

  i = 1;
  if (argc < 2 || argc > 4) {
    printf("\ntFDimg v3, 5/28/03: FDTD time evolution diffusion approx\n\n");
    printf("Included: ");
#if (RF_SUPPORT)
    printf("RF support (not implemented in FD), ");
#endif
    printf("Only Forward-Euler timestepping currently supported.\n");
    printf("v.3 has new circular beam gradient model for S,D optodes.\n\n");
    /* help for options... */
    printf("USAGE:: tFDimg [-opts] [params] input_file (.cfg assumed)\n\n");
    printf("opts: if string contains...\n");
    printf("\tn - inhibit .2pt file calc and dump\n");
    printf("\t[e - use forward-Euler timestepping. Has no effect]\n");
    printf("\tc - read fixed Courant number for dt from params (default %g)\n",
	   COURANT);
    printf("\tt - read fixed dt in ps from params (can't use with c)\n");
    printf("\to - read optode_depth from params (in voxels, default %g)\n",
	   OPTODE_DEPTH);
    printf("\tb - use single beam dump rather than gradient optode model\n");
    printf("\tr - use crude FO Robin BCs rather than default Dirichlet\n");
    printf("\nCurrent recommended options are none, giving the command:\n"
	   "    tFDimg input_file\n"
	   "However, usage of t to fix the timestep (once known) is advised\n"
	   "\nNote the params are read in the following order : cto\n\n");
    printf("The .sig file has the following format:\n");
    printf("    # followed by the dt used in picoseconds\n");
    printf("    time-sequences of signal for each S-D pair (S is outer loop)\n");
    return 1;
  } else if (argc>2) {
    /* search through cmd line opts word to find meaningful chars... */
    opts = argv[1];
    /* printf("opts = %s\n",opts); */
    i = 2; /* argv counter, to params, or filename if no params given */
    if (strchr(opts,'n')!=NULL) {
      system.flags.gen_twopt = 0;
      printf("No 2pt will be generated.\n");
    }
    if (strchr(opts,'e')!=NULL)
      method = FORWARD_EULER;
    if (strchr(opts,'r')!=NULL)
      bcs = CRUDE_FO_ROBIN;
    if (strchr(opts,'b')!=NULL)
      grad_optode = 0;
    if (strchr(opts,'c')!=NULL) {
      sscanf(argv[i++],"%lf",&courant);
      vary_courant = 0;
      printf("Courant number will be fixed at %g\n",courant);
    }
    if (strchr(opts,'t')!=NULL) {
      sscanf(argv[i++],"%lf",&time_step);
      vary_courant = 0;
      printf("dt will be fixed at %g\n",time_step);
    }
    if (strchr(opts,'o')!=NULL) {
      sscanf(argv[i++],"%lf",&optode_depth);
      printf("optode_depth will be fixed at %g\n", optode_depth);
    }
  }
  filehead = argv[i]; /* pointer to filehead name */
  /* report method, BCs... */
  if (method==ADI)
    printf("FDTD method: naive (symmetric) ADI.\n");
  if (method==FORWARD_EULER)
    printf("FDTD method: simple forward-Euler.\n");
  if (bcs==DIRICHLET)
    printf("tissue-air BCs: Dirichlet.\n");
  if (bcs==CRUDE_ROBIN)
    printf("tissue-air BCs: crude 2nd-order Robin (normals aligned with axes) - OBSOLETE, may be unstable\n");
  if (bcs==CRUDE_FO_ROBIN)
    printf("tissue-air BCs: crude FO Robin (normals aligned with axes).\n");
  if (box_bcs==BOX_CANT_TOUCH)
    printf("tissue-box-face BCs: none (can't touch!)\n");
  if (box_bcs==BOX_CRUDE_ROBIN)
    printf("tissue-box-face BCs: crude Robin (normals aligned with axes).\n");
  if (box_bcs==BOX_PML)
    printf("tissue-box-face BCs: perfectly-matched layer.\n");
  if (grad_optode==1)
    printf("gradient optode model will be used.\n");

  
  /* ------------------------- LOAD ------------------------------- */

  /* read in system configuration struct, and rescale to voxel units... */
  if ((n = loadconfig(&system, filehead)) != 0)
    exit(n);
  rescale_system(&system);

  /* allocate and load segmentation... */
  if ((n = loadsegments(&system, &tissueType)) != 0)
    exit(n);



  /* ------------------------ ARRAY ALLOCATION ------------------------- */

  /* 2pt time-fluence... */
  if (system.flags.gen_twopt) {
    II = (double *)malloc(system.nIxstep * system.nIystep *
			  system.nIzstep * system.nTstep  * sizeof(double));
    if (II == NULL) {
      fprintf(stderr, "ERROR: malloc() failed, can't allocate %lu bytes\n",
	      (unsigned long)system.nIzstep * system.nIxstep *
	      system.nIystep * system.nTstep  * sizeof(double));
      exit(1);
    }
    
#if (RF_SUPPORT)
    JJ = (double *)malloc(system.nIzstep * system.nIxstep *
			  system.nIystep * system.nTstep  * sizeof(double));
    if (JJ == NULL)  {
      fprintf(stderr, "ERROR: malloc() failed, can't allocate %lu bytes\n",
	      (unsigned long)system.nIzstep * system.nIxstep *
	      system.nIystep * system.nTstep  * sizeof(double));
      
      exit(1);
    }
#else
    JJ = NULL;
#endif
  }

  /* abbreviations for system sizes in voxels... */
  nx = system.nxstep;
  ny = system.nystep;
  nz = system.nzstep;
  nmax = max(max(nx,ny),nz);

  /* FDTD arrays... */
  /* node indices */
  ni = alloc_3d_int_array(nx+1, ny+1, nz+1);
  
  /* S,D vectors... */
  csv = (int *)malloc(MAX_NCOV*sizeof(int));
  sv = (real *)malloc(MAX_NCOV*sizeof(real));
  ncdv = (int *)malloc((system.nDets)*sizeof(int));
  cdv = (int **)malloc((system.nDets)*sizeof(int *));
  dv = (real **)malloc((system.nDets)*sizeof(real *));
  for (d=0;d<system.nDets;++d) {
    cdv[d] = (int *)malloc(MAX_NCOV*sizeof(int));
    dv[d] = (real *)malloc(MAX_NCOV*sizeof(real));
  }

  /* Output signal array... */
  sig = alloc_3d_array(system.nSrcs, system.nDets, system.nTstep);
  printf("%d sources, %d detectors, %d timegates.\n",system.nSrcs,\
	 system.nDets, system.nTstep);
  printf("first timegate starts at %f ps.\n", system.Tmin*1e12);



  /* ------------------- PRELIM SETUP for all FDTD ------------------------ */

  /* kappa by tissue index... */
  syska[0] = 0.0; /* explicit since don't want *_AIR from tMCimg.h */
  for (ti=1; ti<=system.Ntissue; ++ti) {
    syska[ti] = kappa(&system, ti);
    printf("tissue %d: kappa = %g voxels, mua = %g per voxel\n",ti,\
	   syska[ti],system.tmua[ti]);
  }

  /* find min diffusion time across voxel (max kappa * speed in medium)... */
  max_kv = 0.0;
  for (i=0; i<nx; ++i)
    for (j=0; j<ny; ++j)
      for (k=0; k<nz; ++k) {
	ti = (int)tissueType[i][j][k]; /* int since used as array index */
	if ((kv = syska[ti]*system.v0/system.tn[ti]) > max_kv)
	  max_kv = kv;
      }
  printf("max kappa/n found is %f\n",max_kv/system.v0);

  if (0) {
    /* recalculate an effective smooth (wrt musp's) max_kv: */
    max_kv = 0.0;
    for (ti=1; ti<=system.Ntissue; ++ti) {
      max_kv += pow( syska[ti]*system.v0/system.tn[ti], 4);
    }
    max_kv = pow(max_kv, 0.25);
    printf("effective max kappa/n is %f\n",max_kv/system.v0);
  }




  if (time_step==0.0) {
    courant = COURANT;
    /* time-step stability criterion (computed in voxel-units)... */
    dt = courant / max_kv;
    /* printf("dt = %g, max_kv = %g\n",dt,max_kv); */
    /* find next smallest divisor of tgate... */
    printf("Courant r = %f: gives dt = %f ps  (%f steps per tgate)\n", \
	   courant, dt*1e12, system.stepT/dt);
  } else {
    /* use dt from cmd line. */
    dt = time_step * 1e-12;
    printf("dt = %g ps (%f steps per tgate): max courant = %f\n",dt*1e12, \
	   system.stepT/dt, dt * max_kv);
  }

  /* mixed BC reflectivity fudge-factor... should dep on outer tissue n */
  /* A_refl = 1.0 * SQR(system.tn[1]);  total hack-guess for now */
  /* This is a linear approx between 0.71/musp @ n=1 to 1.96/musp @ n=1.4 */
  A_refl = 1.065 + (system.tn[1] - 1.0)*4.69; /* Anand help, 05/09/03 */






  /* ========================== EULER setup ============================ */

  if (method==FORWARD_EULER) {
    /* prefactor * dt is of order unity, (...since close to instability) */
    Euler_prefac = system.v0; /* includes speed of light */

    nn = 0;                   /* nn will be the # nodes (dofs) */
    for (i=0; i<=nx; ++i)
      for (j=0; j<=ny; ++j)
	for (k=0; k<=nz; ++k)	  /* Deciding which nodes get to be dofs : */
	  if (bcs==DIRICHLET && box_bcs==BOX_CANT_TOUCH) {
	  /* either exclude nodes on tissue-air bdry... */
	    if (categorize_node(tissueType, i,j,k, nx,ny,nz)!=NODE_TIS)
	      ni[i][j][k] = NOT_A_DOF;
	    else
	      ni[i][j][k] = nn++;
	  } else {
	    /* or include them... */
	    if (categorize_node(tissueType, i,j,k, nx,ny,nz)==NODE_AIR)
	      ni[i][j][k] = NOT_A_DOF;
	    else
	      ni[i][j][k] = nn++;
	  }
    printf("vector length nn = %d.\n\n",nn);

    /* allocate vectors and sparse A... */
    F = (real*)malloc(nn*sizeof(real));
    G = (real*)malloc(nn*sizeof(real));
    A = (real**)malloc(nn*sizeof(real*));
    cA = (int**)malloc(nn*sizeof(int*));
    ncA = (int*)malloc(nn*sizeof(int));
    for (n=0; n<nn; ++n) {
      A[n] = (real*)malloc(NNZR*sizeof(real));
      cA[n] = (int*)malloc(NNZR*sizeof(int));
    }

    /* allocate and fill locations given node index... */
    xni = (int*)malloc(nn*sizeof(int));
    yni = (int*)malloc(nn*sizeof(int));
    zni = (int*)malloc(nn*sizeof(int));
    for (i=0; i<=nx; ++i)
      for (j=0; j<=ny; ++j)
	for (k=0; k<=nz; ++k)
	  if ((n=ni[i][j][k])!=-1) {
	    xni[n] = i;
	    yni[n] = j;
	    zni[n] = k;
	  }

  /* compute detection vectors... */
    for (d=0;d<system.nDets;++d) {
      printf("Det %d: ", d);
      ncdv[d] = sparse_optode_vec(cdv[d], dv[d], system.detLoc[d],
				system.detDir[d], system.detRad[d],
				&system, tissueType, ni, nn,
				optode_depth, grad_optode);
    }


    /* -------------------------- Build A (inc prefac) ------------------- */
    printf("building A...\n");

    /* zero all column-indices & vals, so ones unwritten below still valid */
    for (n=0; n<nn; ++n)
      for (i=0; i<NNZR; ++i) {
	A[n][i] = 0.0;
	cA[n][i] = 0;
      }

    if (box_bcs!=BOX_CANT_TOUCH)
      printf("box-face touching BCs not implemented for forward-Euler!\n");

    if (bcs==DIRICHLET) {
      /* .....................basic Dirichlet.............................. */

      /* will fail (seg fault) if any node dof is on box faces */
      for (n=0; n<nn; ++n) {
	x = xni[n];
	y = yni[n];
	z = zni[n];
	i = 0; /* index of which nonzero column of A we're at, for this row */

	/* find average inverse refractive index for current node... */
	inv_n = 0.5 * (1.0/get_edge_prop_cyc(tissueType, system.tn, \
					     1, y, z, x, ny, nz) + \
		       1.0/get_edge_prop_cyc(tissueType, system.tn, \
					     1, y, z, x-1, ny, nz));
	/* NOTE: only valid for homogeneous n! */

	/* offdiag els in x */
	ka_p = get_edge_prop_cyc(tissueType, syska, 1, y, z, x, ny, nz);
	ka_m = get_edge_prop_cyc(tissueType, syska, 1, y, z, x-1, ny, nz);
	if ((j=ni[x-1][y][z])!=NOT_A_DOF) {
	  cA[n][i] = j;
	  A[n][i++] = inv_n * ka_m;
	}
	if ((j=ni[x+1][y][z])!=NOT_A_DOF) {
	  cA[n][i] = j;
	  A[n][i++] = inv_n * ka_p;
	}
      	/* offdiag els in y */
	ka_p = get_edge_prop_cyc(tissueType, syska, 2, z, x, y, nz, nx);
	ka_m = get_edge_prop_cyc(tissueType, syska, 2, z, x, y-1, nz, nx);
	if ((j=ni[x][y-1][z])!=NOT_A_DOF) {
	  cA[n][i] = j;
	  A[n][i++] = inv_n * ka_m;
	}
	if ((j=ni[x][y+1][z])!=NOT_A_DOF) {
	  cA[n][i] = j;
	  A[n][i++] = inv_n * ka_p;
	}
	/* offdiag els in z */
	ka_p = get_edge_prop_cyc(tissueType, syska, 0, x, y, z, nx, ny);
	ka_m = get_edge_prop_cyc(tissueType, syska, 0, x, y, z-1, nx, ny);
	if ((j=ni[x][y][z-1])!=NOT_A_DOF) {
	  cA[n][i] = j;
	  A[n][i++] = inv_n * ka_m;
	}
	if ((j=ni[x][y][z+1])!=NOT_A_DOF) {
	  cA[n][i] = j;
	  A[n][i++] = inv_n * ka_p;
	}
	/* diag el... can use ka_p and ka_m from the last (z) dir */
	/* any direction can be used to find current node mua... */
	mua_p = get_edge_prop_cyc(tissueType, system.tmua, 1, y, z, x, ny, nz);
	mua_m = get_edge_prop_cyc(tissueType, system.tmua, 1, y, z, x-1,
				  ny, nz);
	cA[n][i] = n;
	A[n][i++] = -inv_n*(3*(ka_p+ka_m) + (mua_p+mua_m)/2);
	/* printf("2ka[%d] = %f\n",n,ka_p+ka_m); */
	
	/* finally record number of non-zero cols of A for this row... */
	ncA[n] = i;
      }

    } else if (bcs==CRUDE_FO_ROBIN) {
      /* ..........Crude FO Robin - surf normal in axes dirs only.......... */

      /* will fail if any node dof is on box faces */
      for (n=0; n<nn; ++n) {
	x = xni[n];
	y = yni[n];
	z = zni[n];
	i = 0; /* index of which nonzero column of A we're at, for this row */
	/* zero diagonal (diagonal [i=0] is special, written to >1 times) */
	cA[n][0] = n;
	A[n][i++] = 0.0;

	/* if bdry node then FO Robin takes over from 7-point template... */
	if (categorize_node(tissueType, x,y,z, nx,ny,nz)==NODE_BDRY) {

	  /* sum in values from adjacent nodes it will depend on... */
	  for (d=0; d<3; ++d) {
	    icyc(d,&na,&nb,&nc,nx,ny,nz);
	    icyc(d,&a,&b,&c,x,y,z);
	    /* in Plus direction */
	    if (d==0)
	      j = categorize_node(tissueType, x,y,z+1, nx,ny,nz);
	    else if (d==1)
	      j = categorize_node(tissueType, x+1,y,z, nx,ny,nz);
	    else
	      j = categorize_node(tissueType, x,y+1,z, nx,ny,nz);

	    if (j==NODE_TIS) {
	      ka_p = get_edge_prop_cyc(tissueType, syska, d, a, b, c, na, nb);
	      d_ext = 2*A_refl*ka_p;
	      cA[n][i] = GET_N_P(x, y, z, nx, ny, nz, ni, d);
	      A[n][i++] = d_ext/(1+d_ext);
	    }
	    /* in Minus direction */
	    if (d==0)
	      j = categorize_node(tissueType, x,y,z-1, nx,ny,nz);
	    else if (d==1)
	      j = categorize_node(tissueType, x-1,y,z, nx,ny,nz);
	    else
	      j = categorize_node(tissueType, x,y-1,z, nx,ny,nz);

	    if (j==NODE_TIS) {
	      ka_m = get_edge_prop_cyc(tissueType,syska, d, a, b, c-1, na, nb);
	      d_ext = 2*A_refl*ka_m;
	      cA[n][i] = GET_N_M(x, y, z, nx, ny, nz, ni, d);
	      A[n][i++] = d_ext/(1+d_ext);
	    }

	  }
	  /* if (i==1)
	     printf("disconnected node at xyz = %d %d %d.\n",x,y,z);
	     should be just edges and corners of tissue. */

	  /* finally record number of non-zero cols of A for this row... */
	  ncA[n] = -i; /* use -ve to signify no identity part needed here */

	} else {
	  
	  /* loop over 3 directions: d=0,1,2 is z(xy), x(yz), y(zx). */
	  for (d=0; d<3; ++d) {
	    
	    icyc(d,&na,&nb,&nc,nx,ny,nz);
	    icyc(d,&a,&b,&c,x,y,z);
	    /* it's always inside tissue... */
	    ka_m = get_edge_prop_cyc(tissueType, syska, d, a, b, c-1, na, nb);
	    ka_p = get_edge_prop_cyc(tissueType, syska, d, a, b, c, na, nb);
	    mua_m = get_edge_prop_cyc(tissueType, system.tmua, d, \
				      a, b, c-1, na, nb);
	    mua_p = get_edge_prop_cyc(tissueType, system.tmua, d, \
				      a, b, c, na, nb);
	    /* find average inverse refractive index for current node... */
	    inv_n = (1.0/get_edge_prop_cyc(tissueType, system.tn, \
					   d, a, b, c, na, nb) + \
		     1.0/get_edge_prop_cyc(tissueType, system.tn, \
					   d, a, b, c-1, na, nb)) / 2;
	    cA[n][i] = GET_N_M(x, y, z, nx, ny, nz, ni, d);
	    A[n][i++] = ka_m * inv_n;
	    A[n][0] -= (ka_m + ka_p + (mua_m+mua_p)/6) * inv_n;
	    cA[n][i] = GET_N_P(x, y, z, nx, ny, nz, ni, d);
	    A[n][i++] = ka_p * inv_n;
	  }


	  /* finally record number of non-zero cols of A for this row... */
	  ncA[n] = i;
	}
      }
    } else
      printf("bcs not implemented!\n");
  }


  /* ==================== ADI setup =============================== */

  if (method==ADI) {

    printf("\nADI method no longer currently supported - exiting!\n\n");
    exit(1);

  }

  free(xni);
  free(yni);
  free(zni);


  /* ====================== TIME_EVOLUTION CALCULATION =================== */

  /* --------------------------- loop over sources --------------------- */
  for (s=0; s<system.nSrcs; ++s) {

    printf("\nSrc %d: ", s);

    /* put initial source condition vector in F, strength NT ... */
    ncsv = sparse_optode_vec(csv, sv, system.srcLoc[s], system.srcDir[s],
			     system.srcRad[s], &system, tissueType, ni, nn,
			     optode_depth, grad_optode);
    for (i=0;i<ncsv;++i)
      sv[i] *= system.NT;
    densify(ncsv, csv, sv, nn, F);

    /*for (i=0;i<nn;++i)
      F[i] = random()/2e9; */ /* randomize fluence array - makes faster! */

    /*for (i=0;i<nn;++i)
      printf("F[%d] = %g\n", i, F[i]); */


    /* for this src, clear det signals for all timegates... */
    for (d=0; d<system.nDets; ++d)
      for (tg=0; tg<system.nTstep; ++tg)
	sig[s][d][tg] = 0.0;

    /* --------------------------- loop over time ----------------------- */
    for (t=0.0; t < system.Tmin+system.stepT*system.nTstep; ) {

      /* choose varying timestep here... */
      this_dt = (vary_courant && (t < 1.1e-10)) ? dt/2 : dt;

      printf("t = %g, this_dt = %g\n", t, this_dt);
      /* evolve one step... */
      if (bcs==DIRICHLET)
	sparse_mv_dir(A, cA, ncA, F, G, nn, (real)(Euler_prefac * this_dt));
      else
	sparse_mv(A, cA, ncA, F, G, nn, (real)(Euler_prefac * this_dt));
      /* swap fluence arrays... after this, fluence is back in F */
      dummyv = G;
      G = F;
      F = dummyv;

      /* include signals for each detector from this timestep... */
      tg = extract_sig(&system, F, nn, sig[s], t, this_dt, ncdv, cdv, dv);

      /* if this dt takes you out of current timegate... */
      if (t + this_dt > (tg+1)*system.stepT+system.Tmin) {
	printf("t = %.3f ns (this_dt=%.3f ps)\n",1e9*t,1e12*this_dt);
	/* 2pt file is a snapshot at end of timegate, rather than average... */
	if (system.flags.gen_twopt)
	  record_fluence_in_2pt(&system, tg, F, ni, II, JJ);
      }

      /* increment time... */
      t += this_dt;
    }
    /* ---------- finished time loop ------------- */


    /* save a 2pt file for the current source... */
    if (system.flags.gen_twopt) {
      /* choose name... */
      if (system.nSrcs == 1) {
	printf("Single 2pt file = %d x %d x %d x %d\n",
	       system.nIxstep, system.nIystep, system.nIzstep, system.nTstep);
	/* Save fluences to disk - Jon's routine */
	save2pt(filehead, II, JJ,
		       system.nIxstep * system.nIystep *
		       system.nIzstep * system.nTstep, "2pt.FD");
      } else {
	sprintf(twoptfile, "2pt.FD%d", s); /* suffix */
	printf("2pt file = %d x %d x %d x %d\n",
	       system.nIxstep, system.nIystep, system.nIzstep, system.nTstep);
	/* Save fluences to disk - Jon's routine */
	save2pt(filehead, II, JJ,
		       system.nIxstep * system.nIystep *
		       system.nIzstep * system.nTstep, twoptfile);
      }
    }
    
  }
  /* ================== */

  /* write out all signals... including dt used in ps */
  save_sig(&system, sig, dt*1e12, filehead);
  /* save_sigbinary(&system, sig, filehead); */

  return 0;
}
/************************************************************************
 end of MAIN
 ************************************************************************/







/* --------------------- FORMULA FOR MUSP, KAPPA ------------------------- */
double kappa(const struct Config *system, int ti)
     /* computes kappa given ti, a tissue index */
{
  /* kappa = 1 / (3.mus' + alpha.mua) ... use alpha = 0, see Arridge(1999)
    and Hielscher PMB (1998) numerical tests */
  return 1.0 / (3*musp(system, ti));
}

double musp(const struct Config *system, int ti)
  /* mus' = (1-g)mus ... */
{
  assert(system->tg[ti] < 1);
  return (1.0 - system->tg[ti]) * system->tmus[ti];
}


/* -------------------- sparse vector, matrix utilities --------------- */
void sparse_mv_dir(real **A, int **cA, int *ncA, real *F, real *G, int nn,
	   real scalar)
     /* Does g = (1 + scalar.A)f.
      * Identity and scalar factor included 4/22/02
      *
      * 11/15/04: removing moving pointers to track down Custo SIGFPE bug
      * which was floating underflow anyway. It's no slower.
      *
      * 11/17/04: always use all 7 col entries, unroll loops. Dirichlet only.
      */
{
  int      n;
  real     Gn;
  real*    An;
  int*     cAn;

  for (n=0; n<nn; ++n) {
    An = A[n];
    cAn = cA[n];
    Gn = An[0]*F[cAn[0]] + An[1]*F[cAn[1]] + An[2]*F[cAn[2]] + 
      An[3]*F[cAn[3]] + An[4]*F[cAn[4]] + An[5]*F[cAn[5]] + An[6]*F[cAn[6]];
    G[n] = F[n] + scalar*Gn;
  }
}

/* -------------------- sparse vector, matrix utilities --------------- */
void sparse_mv(real **A, int **cA, int *ncA, real *F, real *G, int nn,
	   real scalar)
     /* Does g = (1 + scalar.A)f.
      * Identity and scalar factor included 4/22/02
      * effect of negative ncA is to zero the identity part for that dof,
      * and ignore the scalar - this is for FO_ROBIN bcs.
      * this is a bad hack, since puts a branch inside the loop!
      *
      * 11/15/04: removing moving pointers to track down Custo SIGFPE bug
      * which was floating underflow anyway. It's no slower.
      *
      * 11/17/04: always use all column 7 entries, unroll loops.
      */
{
  int      n;
  real     Gn;
  real*    An;
  int*     cAn;
  
  /* printf("\t\t\t sparse_mv: scalar = %g\n", scalar); */

  for (n=0; n<nn; ++n) {
    An = A[n];
    cAn = cA[n];
    Gn = An[0]*F[cAn[0]] + An[1]*F[cAn[1]] + An[2]*F[cAn[2]] + 
      An[3]*F[cAn[3]] + An[4]*F[cAn[4]] + An[5]*F[cAn[5]] + An[6]*F[cAn[6]];
    if (ncA[n]>0)
      G[n] = F[n] + scalar*Gn;
    else
      G[n] = Gn;
  }
}




real sparse_dense_inner(int nca, int* ca, real *a, int nb, real *b)
     /* returns (sparse) a dot (dense) b, checking if array bounds of
      * b are exceeded
      *
      * barnett 7/26/02
      */
{
  int i;
  real ans = 0.0;

  for(i=0;i<nca;++i) {
    if (ca[i]<nb && ca[i]>=0)
      ans += a[i] * b[ca[i]];
    else
      printf("sparse_dense_inner : ca[%d]=%d exceeds nb=%d !\n", i, ca[i], nb);
  }
  return ans;
}

void densify(int nca, int* ca, real* a, int nb, real* b)
     /* copies sparse a into dense b */
{
  int i;

  for (i=0;i<nb;++i)
    b[i] = 0.0;

  for(i=0;i<nca;++i) {
    if (ca[i]<nb && ca[i]>=0)
      b[ca[i]] = a[i];
    else
      printf("densify : ca[%d]=%d exceeds nb=%d !\n", i, ca[i], nb);
  }
}


/* --------------------------------------- 3-vector utilities ------------ */
#define INNER3(a, b) \
(a[0]*b[0] + a[1]*b[1] + a[2]*b[2])

void axpy3(double a, double *x, double *y)
     /* output is in y */
{
  int i;
  for (i=0;i<3;++i)
    y[i] += a * x[i];
}

double nrm3(double *y)
{
  int i;
  double nrm = sqrt(INNER3(y, y));

  for (i=0;i<3;++i)
    y[i] /= nrm;
  return nrm;
}

void copy3(double *x, double *y)
     /* x -> y */
{
  y[0] = x[0];
  y[1] = x[1];
  y[2] = x[2];
}

void cross3(double *a, double *b, double *c)
     /* output is in c */
{
  c[0] = a[1]*b[2] - a[2]*b[1];
  c[1] = a[2]*b[0] - a[0]*b[2];
  c[2] = a[0]*b[1] - a[1]*b[0];
}



/* ----------------------------- TISSUE PENETRATE -------------------------- */

#define BCK_STEP 0.1
#define FWD_STEP 0.01
#define MAX_STEPS 10000

TISSUE tissue_penetrate(double *r, double *move, double *launch, double *dir,
			double depth, const struct Config *sp,
			TISSUE ***tissueType)
     /* find point r a depth below tissue surface from launch along dir.
      * Return index if hit tissue, otherwise 0 ('air' or false).
      * Put amount moved along dir vec, until hit surface, in move.
      *
      * barnett 7/26/02 modified from tMCimg6:move_src_just_inside
      */
{
  TISSUE ti;
  int n_bck = 0, n_fwd = 0, in_vol, left_vol;
  
  *move = 0.0;
  copy3(launch, r);
  ti = GET_TISSUE_V(r, sp, tissueType);

  /* Step back to outside tissue, even if means stepping outside system */
  while (IN_TISSUE(ti) && n_bck<MAX_STEPS) {
    axpy3(-BCK_STEP, dir, r);
    *move -= BCK_STEP;
    ++n_bck;
    ti = GET_TISSUE_V(r, sp, tissueType);
  }
  if (n_bck==MAX_STEPS)
    printf("tissue_penetrate: max # backward steps reached! (s/d outside?)\n");

  /* Step forward until in tissue or goes from inside to outside volume */
  do {
    in_vol = IN_VOLUME_V(r, sp);
    axpy3(FWD_STEP, dir, r);
    *move += FWD_STEP;
    ++n_fwd;
    ti = GET_TISSUE_V(r, sp, tissueType);
    left_vol = in_vol && !IN_VOLUME_V(r, sp);
  } while (!IN_TISSUE(ti) && !left_vol && n_fwd<MAX_STEPS);

  /* if (n_fwd==MAX_STEPS)
    printf("tissue_penetrate: max # fwd steps reached! (s/d outside box?)\n");
  
  if (left_vol) {
    printf("Try to launch photon at tissue surface failed: missed tissue!\n"\
	   "Exited system at voxel coords (%f,%f,%f) in tissue %d,\n"\
	   "after %d backward steps then %d forward steps\n",
	   r[0], r[1], r[2], ti, n_bck, n_fwd);
  }

  printf("n_bck=%d, n_fwd=%d\n", n_bck, n_fwd); */

  /* add penetration depth... */
  axpy3(depth, dir, r);
  /* *move += depth; only care about amount moved until hit surface */

  return (ti>0) ? ti : 0;
}

void beamlet_dump(const struct Config *sp, real *ov_dense, int ***ni,\
		   double *r, double depth_correct_fac)
     /* dump a single 'beamlet' (ie photon) into dense optode vector
      *
      * barnett 7/28/02
      */
{
  int i, j, k, n;
  double fx, fy, fz;

  assert(IN_VOLUME_V(r, sp));
  /* add linear interpolation vector for fluence at r... */
  i = (int)r[0];
  j = (int)r[1];
  k = (int)r[2];
  fx = r[0] - (double)i;     /* fractional part of r */
  fy = r[1] - (double)j;
  fz = r[2] - (double)k;
  if ((n=ni[i][j][k])>=0)     /* excludes NOT_A_DOF and NOT_IN_BOX */
    ov_dense[n] += depth_correct_fac*(1.0-fx)*(1.0-fy)*(1.0-fz);
  if ((n=ni[i+1][j][k])>=0)
    ov_dense[n] += depth_correct_fac*(fx)*(1.0-fy)*(1.0-fz);
  if ((n=ni[i][j+1][k])>=0)
    ov_dense[n] += depth_correct_fac*(1.0-fx)*(fy)*(1.0-fz);
  if ((n=ni[i+1][j+1][k])>=0)
    ov_dense[n] += depth_correct_fac*(fx)*(fy)*(1.0-fz);
  if ((n=ni[i][j][k+1])>=0)
    ov_dense[n] += depth_correct_fac*(1.0-fx)*(1.0-fy)*(fz);
  if ((n=ni[i+1][j][k+1])>=0)
    ov_dense[n] += depth_correct_fac*(fx)*(1.0-fy)*(fz);
  if ((n=ni[i][j+1][k+1])>=0)
    ov_dense[n] += depth_correct_fac*(1.0-fx)*(fy)*(fz);
  if ((n=ni[i+1][j+1][k+1])>=0)
    ov_dense[n] += depth_correct_fac*(fx)*(fy)*(fz);
}


/* -------------------------------- OPTODE VECTOR -------------------------- */

#define OPTODE_SAMPLE 0.2   /* 2d beam sample spacing in radius units */
#define MAX_PARALLEL 0.99   /* how parallel can vectors be for orthog calc */
#define GRAD_GAP 0.5        /* gap in voxels between 2 depths for grad meas */

int sparse_optode_vec(int *cov, real *ov, double *loc, double *dir,
		      double rad, const struct Config *system,
		      TISSUE ***tissueType, int ***ni,
		      int nn, double optode_depth, int grad_flag)
     /* generate sparse vector representing optode coupling to fluence field
      * Need not be fast. Returns the number of non-zero elements in vec.
      * Fills the already-allocated arrays cov and ov with sparse rep of vec.
      *
      * Currently uses circular beam dumped constant OPTODE_DEPTH below
      * surface, with factor of 1/musp to account for `true' depth of 1/musp.
      * grad_flag tells it to subtract 2 such beam-dumps at different
      * depths, giving a gradient measurement.
      *
      * barnett 7/26/02 grad added 7/28/02
      */
{
  int nsamp = 0, nhit = 0, c = 0, i;
  double u, v, du[3], dv[3], launch[3], r[3];
  real *ov_dense; 
  double move, tot_move = 0.0, depth_correct_fac;
  TISSUE ti;
  
  if (dir[0]==0.0 && dir[1]==0.0 && dir[2]==0.0) {
    printf("sparse_optode_vec: optode must have defined direction!\n");
    return 0;
  }
  /* compute 2 orthogonal vecs spanning transverse plane...
     |dir|=1 assumed. Operation must be repeatable given system. */
  du[0] = 0.0;    /* guess e_z */
  du[1] = 0.0;
  du[2] = 1.0;
  if (fabs(INNER3(du, dir)) > MAX_PARALLEL) {
    du[1] = 1.0;   /* since was too parallel, now guess e_y instead */
    du[2] = 0.0;
  }
  axpy3(-INNER3(du, dir), dir, du);
  nrm3(du);
  cross3(dir, du, dv);
  if ((INNER3(du,dv) + INNER3(du,dir) + INNER3(dir,dv)) > 1e-6)
    printf("sparse_optode_vec: lack of orthogonality: dir = %g %g %g, du = %g %g %g, dv = %g %g %g\n", dir[0], dir[1], dir[2], du[0], du[1], du[2], dv[0], dv[1], dv[2]);

  /* allocate and zero dense vec... */
  ov_dense = (real*)malloc(nn*sizeof(real));
  for (i=0;i<nn;++i)
    ov_dense[i] = 0.0;

  /* sample over square, rejecting on unit circle... */
  for (u=-1+OPTODE_SAMPLE/2; u<=1; u+=OPTODE_SAMPLE)
    for (v=-1+OPTODE_SAMPLE/2; v<=1; v+=OPTODE_SAMPLE)
      if ((u*u + v*v)<=1.0) {
	/* find r where this part of beam goes into tissue right depth... */
	copy3(loc, launch);
	axpy3(rad*u, du, launch);
	axpy3(rad*v, dv, launch);
	ti = tissue_penetrate(r, &move, launch, dir,
			      optode_depth, system, tissueType);
        if (IN_TISSUE(ti)) {
	  depth_correct_fac = 1/(optode_depth * musp(system, ti));
	  /* printf("ti=%d, musp=%f\n\n",ti, musp(system, ti)); */
	  tot_move += move;
	  if (grad_flag) {
	    /* gradient meas using dipole beam dump at 2 depths */
	    beamlet_dump(system, ov_dense, ni, r, -depth_correct_fac/GRAD_GAP);
	    /* move further in and add with opposite sign... */
	    axpy3(GRAD_GAP, dir, r);
	    beamlet_dump(system, ov_dense, ni, r, depth_correct_fac/GRAD_GAP); 
	  } else {
	    /* single beam dump */
	    beamlet_dump(system, ov_dense, ni, r, depth_correct_fac);
	  }
	  ++nhit;
	}
	++nsamp;
      }

  /* sparsify ov, also normalize by 1/nsamp ... */
  for (i=0;i<nn;++i)
    if (ov_dense[i]!=0.0) {
      cov[c] = i;
      ov[c] = ov_dense[i] / nsamp;
      ++c;
    }
  free(ov_dense);

  if (nhit==0)
    printf("no tissue-hits for this optode!\n");
  else
    printf("tissue-hit frac = %g, moved avg %g voxels, involves %d dofs\n",
	   nhit/(double)nsamp, tot_move/nhit, c);

  return c;
}


/* ---------------------- EXTRACT DETECTED SIGNALS ---------------------- */
int extract_sig(const struct Config* sp, real* F, int nn, \
		double** sig_of_s, double t_beg, double dt, \
		int* ncdv, int** cdv, real** dv)
 /* Adds to sig_of_s using fluence in vector F, and factor TIME_WEIGHT.
  * barnett 3/18/02 mod 5/3/02 to include interpolation into timegates,
  * with signal from F having uniform strength in time from t_beg to t_end.
  * Returns timegate number for current t.
  * Use of sparse optode vectors done 7/26/02.
  * Corrected to handle Tmin correctly, barnett 5/28/03.
  */
{
  int d, i_beg, i_end, i, i_lo, i_hi;
  double det_sig[MAXDET], t_end, t_in, t_i;

  /* rapid dot product of F with det vecs... */
  for (d=0; d<sp->nDets; ++d)
    det_sig[d] = sparse_dense_inner(ncdv[d], cdv[d], dv[d], nn, F);

  /* now interpolate det_sig into sig_of_s 2d array... messy */
  t_end = t_beg + dt; /* det_sig taken uniform from t_beg to t_end. */
  i_beg = (int)((t_beg - sp->Tmin)/sp->stepT);
  i_end = (int)((t_end - sp->Tmin)/sp->stepT);
  i_lo = max(0, i_beg);  /* clip to timegate array bounds */
  i_hi = min(sp->nTstep-1, i_end);

  /* loop over relevant timegates, adding in signal is needed... */
  for (i=i_lo; i<=i_hi; ++i) {
    t_i = sp->Tmin + i*sp->stepT;   /* start of current gate */
    t_in = sp->stepT - min(sp->stepT, max(0, t_beg-t_i)) -
      min(sp->stepT, max(0, t_i+sp->stepT-t_end));
    for (d=0; d<sp->nDets; ++d)
      sig_of_s[d][i] += TIME_WEIGHT * t_in * det_sig[d];
  }

  return i_beg; /* if <0 then indicates nothing has been written yet */
}


/* -------------------------- 2PT FILL (vector) -------------------------- */
void record_fluence_in_2pt(const struct Config *system, int t, real *F,
			   int ***ni, double *II, double *JJ)
 /* copies fluence vec F into current tgate=t part of 2pt time-fluence II
  *
  * currently does not average to go from node-based to voxel-based,
  * rather data is shifted by +1/2 in each axis, losing the data with
  * maximum index in each axis (ie i=nx, etc).
  *
  * no filling of JJ yet.
  *
  * 5/28/03 writes nothing if t<0. 
  */
{
  int i,j,k,n, VoxN;

  assert(t < system->nTstep);
	
  if (t>=0) {
    for (i=system->Ixmin; i<=system->Ixmax; ++i)
      for (j=system->Iymin; j<=system->Iymax; ++j)
	for (k=system->Izmin; k<=system->Izmax; ++k) {
	  
	  /* Use Matlab ordering so I can dump to disk 
	   * the entire block of memory later */
	  VoxN = ((j - system->Iymin) + system->nIystep * 
		  ((i - system->Ixmin) + system->nIxstep *
		   ((k - system->Izmin) + system->nIzstep * t)));
	  
	  assert(VoxN < (system->nIxstep * system->nIystep *
			 system->nIzstep * system->nTstep));
	  
	  if ((n = ni[i][j][k]) != -1)
	    II[VoxN] = F[n];
	  else
	    II[VoxN] = 0.0;
	}
  }
}


/* ------------------------------------- CATEGORIZE NODE ---------------- */
int categorize_node(TISSUE ***tt, int x, int y, int z, int nx, int ny, int nz)
 /* return:
  * NODE_AIR if surrounded by air,
  * NODE_BDRY if boundary,
  * NODE_TIS if surrounded by tissue.
  *
  * barnett 3/27/02, simplified 4/22/02
  */
{
  int n=0;    /* counter for non-air voxels */

  if (x<0 || x>nx || y<0 || y>ny || z<0 || z>nz) {
    /* will cause error if look up - outside box... */
    return NOT_IN_BOX;
  } else {
    if (x<nx && y<ny && z<nz)
      if (tt[x][y][z]!=0)
	++n;
    if (x>0 && y<ny && z<nz)
      if (tt[x-1][y][z]!=0)
	++n;
    if (x<nx && y>0 && z<nz)
      if (tt[x][y-1][z]!=0)
	++n;
    if (x>0 && y>0 && z<nz)
      if (tt[x-1][y-1][z]!=0)
	++n;
    if (x<nx && y<ny && z>0)
      if (tt[x][y][z-1]!=0)
	++n;
    if (x>0 && y<ny && z>0)
      if (tt[x-1][y][z-1]!=0)
	++n;
    if (x<nx && y>0 && z>0)
      if (tt[x][y-1][z-1]!=0)
	++n;
    if (x>0 && y>0 && z>0)
      if (tt[x-1][y-1][z-1]!=0)
	++n;

  /* basic node category... */
    if (n==8)
      return NODE_TIS;
    else if (n!=0)
      return NODE_BDRY;
    else
      return NODE_AIR;
  }
}


/* --------------------- ADI-setup cyclic routines ------------------- */

int touch_material_cyc(TISSUE ***tissueType, int d, int a, int b, int c, \
		       int na, int nb)
     /* returns true if the edge from node (a,b,c) to (a,b,c+1) touches
	any kind of non-air material (ie tissue index > 0).
	Uses na to safely stay within range a = 0...na, similarly nb.
	c must be 0...nc-1.
	a,b,c are cyclic permutation of x,y,z, given by d = 0 (or 3), 1 or 2.
     */
{
  /* the ordering of &&'s is important: the range checks on a,b prevent
     out-of-array requests for Tgetcyc. */
  return \
    (a<na && b<nb && Tgetcyc(tissueType,d,a  ,b  ,c) > 0) || \
    (a>0  && b<nb && Tgetcyc(tissueType,d,a-1,b  ,c) > 0) || \
    (a<na && b>0  && Tgetcyc(tissueType,d,a  ,b-1,c) > 0) || \
    (a>0  && b>0  && Tgetcyc(tissueType,d,a-1,b-1,c) > 0);
}

double get_edge_prop_cyc(TISSUE ***tissueType, double *p, int d, int a, \
			 int b, int c, int na, int nb)
     /* returns an optical property of an edge defined as in
	touch_material_cyc. p array gives property for each tissue index.
	In other words, d=0 is edge in z dir, d=1 x dir, d=2 y dir.
	The value is the average of the value in the 1-4 adjacent non-air
	voxels.
	If the edge is surrounded by air, prints error and returns zero.
	Array indexing and range checks as in touch_material_cyc.
     */
{
  int n_adj_voxels = 0, ti;
  double prop_tot = 0.0;

  if (a<na && b<nb && (ti=Tgetcyc(tissueType,d,a  ,b  ,c)) > 0) {
    ++n_adj_voxels;
    prop_tot += p[ti];
  }
  if (a>0  && b<nb && (ti=Tgetcyc(tissueType,d,a-1,b  ,c)) > 0) {
    ++n_adj_voxels;
    prop_tot += p[ti];
  }
  if (a<na && b>0  && (ti=Tgetcyc(tissueType,d,a  ,b-1,c)) > 0) {
    ++n_adj_voxels;
    prop_tot += p[ti];
  }
  if (a>0  && b>0  && (ti=Tgetcyc(tissueType,d,a-1,b-1,c)) > 0) {
    ++n_adj_voxels;
    prop_tot += p[ti];
  }

  if (n_adj_voxels == 0) {
    printf("get_edge_prop_cyc: edge is in air! dabc = %d %d %d %d\n", \
	   d,a,b,c);
    return 0;
  } else
    return prop_tot/n_adj_voxels;
}

void icyc(int d, int *a, int *b, int *c, int x, int y, int z)
     /* cycles abc <- xyz (for d=0)
	              yzx        1
		      zxy        2
        Works for type int.
     */
{
  if (d==0 || d==3) {
    /* case of d=3 allows convenient d <- (3-d) to reverse cycling. */
    *a = x;
    *b = y;
    *c = z;
  } else if (d==1) {
    *a = y;
    *b = z;
    *c = x;
  } else if (d==2) {
    *a = z;
    *b = x;
    *c = y;
  } else {
    fprintf(stderr,"invalid d in icyc.\n");
    exit(1);
  }
}

TISSUE Tgetcyc(TISSUE ***T, int d, int a, int b, int c)
     /* gets TISSUE value from matrix T, using indeces [x][y][z] cycled
	as in icyc above
     */
{
  int x,y,z;
  /* printf("Tgetcyc: abc = %d %d %d\n",a,b,c); */
  icyc(3-d,&x,&y,&z,a,b,c);
  return T[x][y][z];
}

real dgetcyc(real ***A, int d, int a, int b, int c)
     /* gets real value from matrix A, using indeces [x][y][z] cycled
	as in icyc above
     */
{
  int x,y,z;
  icyc(3-d,&x,&y,&z,a,b,c);
  return A[x][y][z];
}






/* --------------------- MULTIDIMENSIONAL ARRAYS ALLOCATION ----------- */

/* ----------------------------------------------------------------- */
double ***alloc_3d_array(int x, int y, int z)
     /* allocates 3d array of doubles. adapted from Jon. */
{
  int i,j;
  double ***p;
  
  if ((p = (double ***)malloc(x * sizeof(double **))) == NULL) {
    fprintf(stderr, "malloc() failed, can't allocate %d bytes\n",
	    x * sizeof(double **));
    return NULL;
  }
  
  for (i=0; i<x; i++) {
    if ((p[i] = (double **)malloc(y * sizeof(double *))) == NULL) {
      fprintf(stderr, "malloc() failed, can't allocate %d bytes\n",
	      y * sizeof(double *));
      return NULL;
    } else
      for (j=0; j<y; j++)
	if ((p[i][j] = (double *)malloc(z * sizeof(double))) == NULL) {
	  fprintf(stderr, "malloc() failed, can't allocate %d bytes\n",
		  z * sizeof(double));
	  return NULL;
	}
  }
  
  return p;
}


/* ----------------------------------------------------------------- */
int ***alloc_3d_int_array(int x, int y, int z)
     /* allocates 3d array of ints. */
{
  int i,j;
  int ***p;
  
  if ((p = (int ***)malloc(x * sizeof(int **))) == NULL) {
    fprintf(stderr, "malloc() failed, can't allocate %d bytes\n",
	    x * sizeof(int **));
    return NULL;
  }
  
  for (i=0; i<x; i++) {
    if ((p[i] = (int **)malloc(y * sizeof(int *))) == NULL) {
      fprintf(stderr, "malloc() failed, can't allocate %d bytes\n",
	      y * sizeof(int *));
      return NULL;
    } else
      for (j=0; j<y; j++)
	if ((p[i][j] = (int *)malloc(z * sizeof(int))) == NULL) {
	  fprintf(stderr, "malloc() failed, can't allocate %d bytes\n",
		  z * sizeof(int));
	  return NULL;
	}
  }
  
  return p;
}



/* ----------------------------------------- RESCALE -------------- */
void rescale_system(struct Config *system)
/* Rescale the Universe to get unit grid spacing.  Units are
   specified in mm in the input file.  system.dl has dimension
   of mm/voxel, which takes everything into voxel units.
   
   This is the only place where system struct can be modified, apart
   from if sources or detectors moved.
*/
{
  int i;

  system->v0 /= system->dl;
  system->k0 *= system->dl;

  system->Lmax  /= system->dl;
  system->Lmin  /= system->dl;
  system->stepL /= system->dl;

  for (i = 1; i <= system->Ntissue; i++) {
    system->tmus[i] *= system->dl;
    system->tmua[i] *= system->dl;
  }

  for (i = 0; i < system->nDets; i++) {
    system->detLoc[i][0] /= system->dl;
    system->detLoc[i][1] /= system->dl;
    system->detLoc[i][2] /= system->dl;
    system->detRad[i]    /= system->dl;
  }
  for (i = 0; i < system->nSrcs; i++) {
    system->srcLoc[i][0] /= system->dl;
    system->srcLoc[i][1] /= system->dl;
    system->srcLoc[i][2] /= system->dl;
    system->srcRad[i]    /= system->dl;
  }

  system->xstep /= system->dl;   /* these are never used (cubic voxels) */
  system->ystep /= system->dl;
  system->zstep /= system->dl;
}



/* ======================== FILE INPUT / OUTPUT ============================ */


/* -------------------------------------------------------------------- */
int save_sigbinary(const struct Config *system, double ***sig, char *filehead)
{
  int s, d, nout;
  FILE *fp;
  char sigfile[FILENMLEN];

  /* open pointer to signal output file... */
  snprintf(sigfile, FILENMLEN, "%s.sig.b", filehead);
  if ((fp = fopen(sigfile, "wb")) == NULL) {
    fprintf(stderr, "ERROR: Can't open signal file %s for writing\n",sigfile);
    return 1;
  }

  /* write ordering (slow->fast) as (d,s,t) so Matlab reads as (s,d,t)... */
  for (d=0; d<system->nDets; ++d)
    for (s=0; s<system->nSrcs; ++s) {
      nout = system->nTstep; /* write all time-gates as a block */
      if (fwrite(sig[s][d], sizeof(double), nout, fp) != nout) {
        perror("Error while writing sig.b file\n");
        fclose(fp);
        return 1;
      }
    }

  fclose(fp);
  return 0;
}

int save_sig(const struct Config *system, double ***sig, double courant,
	     char *filehead)
     /* write sig(s,d,t) in that order as human-readable data,
        also gnuplottable. */
{
  int s,d,t;
  char filenm[FILENMLEN];
  FILE *fp;

  snprintf(filenm, FILENMLEN, "%s.sig", filehead);

  if ((fp = fopen(filenm, "w")) == NULL) {
    perror("Can't open sig file for writing\n");
    return 1;
  }

  fprintf(fp, "# %g\n\n", courant);

  for (s=0; s<system->nSrcs; ++s) {
    for (d=0; d<system->nDets; ++d) {
      for (t=0; t<system->nTstep; ++t)
	fprintf(fp, "%g\n", sig[s][d][t]);
      fprintf(fp, "\n");
    }
    fprintf(fp, "\n");
  }

  fclose(fp);

  return 0;
}


/*********************************************************
             OPEN AND READ THE INPUT FILE

  Returns non-zero on error
    1 -> File error (can't open file, unexpected EOF, etc.)
    2 -> Invalid input parameter
    3 -> System error (out of memory, etc.)
  Modified barnett.
  4/18/02: removed zero-ing of struct, so can leave flags untouched.

*********************************************************/

int loadconfig(struct Config *p, char *filehead)
{
  char cfgfile[FILENMLEN];
  FILE *fp;
  int i, n;

  /* Set some constants */

  for (i = 0; i < MAXTISSUE; i++)
    {
      p->tmus[i] = MUS_AIR;
      p->tmua[i] = MUA_AIR;
      p->tg[i]   =   G_AIR;
      p->tn[i]   = IDX_AIR;
    }

  /* Open input file... */
  snprintf(cfgfile, FILENMLEN, "%s.cfg", filehead);
  if ((fp = fopen(cfgfile, "rt")) == NULL) {
    fprintf(stderr,
	    "input_file \"%s\" not found, assuming you meant \"%s\"\n",
	    cfgfile, filehead);
    
    if ((fp = fopen(filehead, "rt")) == NULL) {
      fprintf(stderr, "usage: tFDimg input_file (.cfg assumed)\n");
      fprintf(stderr, "input_file = %s does not exist.\n", cfgfile);
      return 1;
    } 
    else {
      /* chop off the extension */
      char *p = rindex(filehead, '.');
      
      if (p != NULL)
	*p = '\0';
    }
  }

  /* READ THE INPUT FILE */
  n = parseconfig(p, fp);

  /* CLOSE FILE AND RETURN */
  fclose(fp);
  return n;
}


/*********************************************************
        OPEN AND READ THE SEGMENTATION FILE

  Returns non-zero on error
    1 -> File error (can't open file, unexpected EOF, etc.)
    2 -> Invalid input parameter
    3 -> System error (out of memory, etc.)

*********************************************************/

int loadsegments(const struct Config *p, TISSUE ****tt)
{
  TISSUE ***tissueType;
  FILE *fp;
  int i, j, k;

  assert(p != NULL);
  assert(tt != NULL);

  if ((fp = fopen(p->segFile, "rb")) == NULL) 
    {
      fprintf(stderr,
              "ERROR: THe binary image file %s was not found!\n",
              p->segFile);
      return 1;
    }

  rewind(fp);			  /* Check actual vs expected file sizes */
  fseek(fp, 0, SEEK_END);

  if (ftell(fp) != p->nxstep * p->nystep * p->nzstep * sizeof(TISSUE)) 
    {
      fprintf(stderr, 
              "Size of segmentation file != nxstep*nystep*nzstep elements\n");
      return 1;
    } 
  else
    rewind(fp);

  if ((*tt = (TISSUE ***)malloc(p->nxstep * sizeof(TISSUE **))) == NULL) 
    {
      fprintf(stderr, "malloc() failed, can't allocate %d bytes\n",
              p->nxstep * sizeof(TISSUE **));

      return 3;
    }

  tissueType = *tt;

  for (i = 0; i < p->nxstep; i++) 
    {
      tissueType[i] = (TISSUE **)malloc(p->nystep * sizeof(TISSUE *));

      if (tissueType[i] == NULL) 
        {
          fprintf(stderr, "malloc() failed, can't allocate %d bytes\n",
                  p->nystep * sizeof(TISSUE *));
          return 3;
        } 
      else
        for (j = 0; j < p->nystep; j++) 
          {
            tissueType[i][j] = (TISSUE *)malloc(p->nzstep * sizeof(TISSUE));
            
            if (tissueType[i][j] == NULL) 
              {
                fprintf(stderr,
                        "malloc() failed, can't allocate %d bytes\n",
                        p->nzstep * sizeof(TISSUE));
                return 3;
              }
          }
    }

  /* Matlab stores data column [Y,X,Z], I use data row
   * [X,Y,Z] first.  Flip the data around as I read it in.
   */

  for (k = 0; k < p->nzstep; k++)
    {
      for (i = 0; i < p->nxstep; i++) 
        {
          for (j = 0; j < p->nystep; j++) 
            {
              TISSUE s;

              if (fread(&s, sizeof(TISSUE), 1, fp) != 1) 
                { 
                  fprintf(stderr,
                          "Unexpected end of segmentation file %s", 
                          p->segFile);
                  return 1;
                }  
              else 
                tissueType[i][j][k] = s;
            }
        }

      printf(".");
      fflush(stdout);
    }

  printf("\n");

  fclose(fp);

  return 0;
}


/*********************************************************
              WRITE FLUENCE DATA TO DISK  

  Returns non-zero on error
    1 -> File error (can't open file, unexpected EOF, etc.)
    2 -> Invalid input parameter
    3 -> System error (out of memory, etc.)
*********************************************************/

/* preserved double precision, since that is part of format. barnett 11/17/04
 */

int save2pt(const char *basen, const double *udata, 
            const double *vdata, int ndata, const char *suffix)
{
  char filenm[FILENMLEN];
  FILE *fp;
  int nu, nv;

  snprintf(filenm, FILENMLEN, "%s.%s", basen, suffix);

  if ((fp = fopen(filenm, "wb")) == NULL) 
    {
      perror("Can't open 2pt file for writing\n");
      return 1;
    }

  assert(udata != NULL);

  nu = fwrite(udata, sizeof(double), ndata, fp);

  if (nu != ndata) 
    {
      fclose(fp);
      perror("Error while writing 2pt file\n");
      return 1;
    }

  /* vdata == NULL is allowed for CW data */

  if (vdata != NULL) 
    if ((nv = fwrite(vdata, sizeof(double), ndata, fp)) != ndata) 
      {
        perror("Error while writing 2pt file\n");
        fclose(fp);
        return 1;
      }

  fclose(fp);

  return 0;
}


/* end of file */
