/* Generic physical constants */

#define C_VACUUM  2.99792458e11 
#define PI        3.14159265358979323846

#define MUS_AIR   -999.0
#define MUA_AIR   -999.0
#define IDX_AIR   -999.0
#define   G_AIR   -999.0

#define FILENMLEN   1024
#define MAXTISSUE    127 /* Limited by size of TISSUE below */
#define MAXDET       255

/* Useful macros */

#ifndef SQR
#define SQR(x) ((x)*(x))
#endif

#ifndef max
#define max(a,b)	((a) > (b) ? (a) : (b))
#endif

#ifndef min
#define min(a,b)	((a) < (b) ? (a) : (b))
#endif

/* Given a position (in real-valued voxel units),
 * return 1 if the position is inside the defined system
 * volume, otherwise return 0.
 * NOTE: All Jon's nested ? : 's replaced with multiple-&&'s, & debugged.
 */
#define IN_VOLUME(x, y, z, sp) \
    ((int)(x) >= 0 && (int)(x) < (sp)->nxstep && \
     (int)(y) >= 0 && (int)(y) < (sp)->nystep && \
     (int)(z) >= 0 && (int)(z) < (sp)->nzstep)

#define IN_VOLUME_V(r, sp) \
    ((int)(r[0]) >= 0 && (int)(r[0]) < (sp)->nxstep && \
     (int)(r[1]) >= 0 && (int)(r[1]) < (sp)->nystep && \
     (int)(r[2]) >= 0 && (int)(r[2]) < (sp)->nzstep)

/* Same, but for image volume */
#define IN_ROI(x, y, z, sp) \
    ((int)(x) >= (sp)->Ixmin && (int)(x) <= (sp)->Ixmax && \
     (int)(y) >= (sp)->Iymin && (int)(y) <= (sp)->Iymax && \
     (int)(z) >= (sp)->Izmin && (int)(z) <= (sp)->Izmax)

/* Return 1 if the tissue index > 0, which corresponds to propagation
 *  through "tissue" (as opposed to "air" or outside system)
 */
#define IN_TISSUE(ttidx) (ttidx > 0)

#define INDEX_MATCHED(x, y, z, sp, tt, ti) \
    ((sp)->tn[(int)GET_TISSUE(x,y,z,sp,tt)] == (sp)->tn[(ti)])

/* Given a position, return the corresponding tissue type.
 * Note sp is _pointer_ to system struct.
 *
 *  Note change from Jon's returning 0:  Alex 4/18/02
 *  Returns -1 for positions outside the defined volume
 *  (zero would not distinguish from air).
 */
#define GET_TISSUE(x,y,z,sp,tt) \
(((int)(x) >= 0 && (int)(x) < (sp)->nxstep && \
(int)(y) >= 0 && (int)(y) < (sp)->nystep && \
(int)(z) >= 0 && (int)(z) < (sp)->nzstep) ? \
tt[(int)(x)][(int)(y)][(int)(z)] : -1)

/* Same, except expects location as 3-vector... */
#define GET_TISSUE_V(r,sp,tt) \
(((int)(r[0]) >= 0 && (int)(r[0]) < (sp)->nxstep && \
(int)(r[1]) >= 0 && (int)(r[1]) < (sp)->nystep && \
(int)(r[2]) >= 0 && (int)(r[2]) < (sp)->nzstep) ? \
tt[(int)(r[0])][(int)(r[1])][(int)(r[2])] : -1)


/* Constants, faster than forever re-computing the square roots */
#ifdef M_SQRT2
# define SQRT2 M_SQRT2
#else
# define SQRT2 1.41421356237309504880
#endif

#ifdef M_SQRT3
# define SQRT3 M_SQRT3
#else
# define SQRT3 1.73205080756887729352
#endif


/* Type of tissueType array.  Change to "short" or "int" if more tissue
 *  types are needed, but that will break compatibility with existing
 *  segmentation files.
 *
 *  NOTE: TISSUE must be a signed type since "-1" is used to indicate
 *  undefined values.
 */
typedef char TISSUE;


/* Store the entire system configuration.  These values should be
 * treated as read-only once the system configuration routines
 * are finished.
 */
struct Config
{
  int NT;			  /* Total number of photons */
  long seed;			  /* Random number seed */
  double v0;			  /* Speed of light (vacuum) */
  double frequency;		  /* RF Frequency */
  double k0;			  /* 2pi/lambda_rf */
  double dl;			  /* conversion from mm to voxels */

  /* ************************************************************ */

#if 0				  /* OBSOLETE */
  double xi, yi, zi;		  /* Initial photon position */
  double cxi, cyi, czi;		  /* Initial photon direction */
#endif

  /* ************************************************************ */

  double xstep, ystep, zstep;	        /* Voxel dimensions */
  int nxstep, Ixmin, Ixmax, nIxstep;	/* Image (ROI) dimensions */
  int nystep, Iymin, Iymax, nIystep;
  int nzstep, Izmin, Izmax, nIzstep;

  double Tmin, Tmax, stepT;	  /* Time stepping for recording */
  int nTstep;
  double Lmin, Lmax, stepL;	  /* min/max photon lenght, computed
				   * from min/max propagation times */

  /* ************************************************************ */

  char segFile[FILENMLEN];	  /* Filename for segmentation file */
  int Ntissue;			  /* Tissue properties */

  double tmus[MAXTISSUE];	  /* Optical properties, by type */
  double tmua[MAXTISSUE];
  double tg[MAXTISSUE];
  double tn[MAXTISSUE];

  /* ************************************************************ */

  int nDets;			  /* # of Detectors */

#if 0
  double detRad, detNA;           /* Obsolete */
#endif

  double detLoc[MAXDET][3];       /* Position */
  double detDir[MAXDET][3];       /* Orientation (unit vector) */
  double detRad[MAXDET];          /* Radius */
  double  detNA[MAXDET];          /* NA */

  /* ************************************************************ */

  /* tMCimg only supports one source (the first one defined) right
   *   now.  Arrays are defined for future expansion and for code
   *   sharing with other programs.
   */

  int nSrcs;

#if 0
  double srcRad, srcNA;
#endif

  double srcLoc[MAXDET][3];
  double srcDir[MAXDET][3];
  double srcRad[MAXDET];
  double  srcNA[MAXDET];

  /* ************************************************************ */

  /* Boolean flags that control program behavior */

  struct 
    {
      char gen_twopt;          /* Generate .2pt file? */
      char source_move;        /* Advance sources to boundary? */
      char move_before_NA;     /* Move each photon to surface b4 randomizing
				  direction appropriate to src NA ? */
      char det_move;           /* Advance detectors to boundary */
      char exact_exit_time;    /* Store exit time or exit gate */
      char mirror;             /* Mirrored walls? */
    } 
  flags;
};

