/*********************************************************************
 * tMCimg
 *
 *  mods from tMCimg6: barnett 3/1/02 added cmd line options -nst
 *                     fixed voxel<->mm dimensional problems.
 *      HIS formats    .his variants are .his.t (det #, exact times) and
 *                     .his.a (all photon locations, exact times) (nDet=0)
 *  4/4/02 Jon's mirrored-air-bdry mod cmd line, tissuetype update bugfix.
 *  
 *  11Apr02 Merged JS and AB versions, changed filename from
 *            tMCimg6.c to tMCimg6b.c.  Moved flags into the config
 *            structure.  Changed parser to allow different radii, NA,
 *            direction, and position for every source and detector.
 *            tMCimg still only uses the first source and does not yet
 *            support detector NA though.
 *  6/17/02 multiple sources, bugs fixed in finite radius, IN_VOLUME.
 *
 *********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include "tMCimg.h"

#ifdef _WIN32
/* For some reason, windows adds an extra underscore */
#define snprintf _snprintf
#else
# include <unistd.h>
#endif

# if (defined __linux__) && (defined __i386__)
/* Optional FPU controls generate SIGFPE on certain math errors.
   In particular divide-by-zero will cause core dumps instead of
   silently proceeding. Define INHIBIT_FPU_CONTROL to restore default
   behavior. */
/* #define INHIBIT_FPU_CONTROL */
#endif

#define USAGE printf("USAGE:: tMCimg [-opts] input_file (.cfg assumed)\n"\
"\n\ntMCimg version 1.0\n\n"\
"opts: if string contains...\n"\
"\t? - display the option help (this)\n"\
"\tn - inhibit .2pt file calc and dump\n"\
"\ts - don't move source to the boundary\n"\
"\ta - move photon launch to boundary _after_ applying diffuse src NA\n"\
"\tt - write exact exit time to .his.t instead of tgate # to .his\n"\
"\tm - make tissue-air boundaries perfectly mirrored\n\n")

/*********************************************************************
 Function prototypes
 *********************************************************************/

extern double ran(int *idum);
extern int parseconfig(struct Config *, FILE *);

int  loadconfig(struct Config *p, int argc, char *argv[]);
int  loadsegments(const struct Config *p, TISSUE ****tt);
void fpusetup(void);
void run_simulation(const struct Config *, const TISSUE ***,
                    double *, double *, FILE *, int);
void rotate_phase(double *, double *, const struct Config *, const float *);
int  save2pt(const char *, const double *, const double *, int);
void savehis(const struct Config *, FILE *,
             double, double, double, double, float *);
void scatter(double, double *, double *, double *, int *);
int move_src_just_inside(double *x, double *y, double *z,\
			 double cx, double cy, double cz,\
			 const struct Config *system,\
			 const TISSUE ***tissueType);
void rotate(double *x, double *y, double *z, double th0, double ph0);

/*********************************************************************
 BEGIN PROGRAM CODE
 *********************************************************************/

int main(int argc, char *argv[])
{
  struct Config system;		  /* System configuration, constructed
				   *   from the .cfg file */
  double *II, *JJ;		  /* FOR STORING THE 2-PT FLUENCE */
  TISSUE ***tissueType;		  /* STORE THE IMAGE FILE */
  FILE *fp;			  /* FILE POINTERS FOR SAVING THE DATA */
  char filenm[FILENMLEN];	  /* BASE FILE NAME FOR DATA FILE */
  char filenm_s[FILENMLEN];	  /* FILE NAME FOR DATA FILE */
  int i, j, n, s, save_2pt_err = 0;

  /* Initialize the configuration structure's memory */

  memset(&system, 0, sizeof(struct Config));

  system.flags.gen_twopt       = 1;
  system.flags.source_move     = 1;
  system.flags.move_before_NA  = 1;
  system.flags.exact_exit_time = 0;
  system.flags.mirror          = 0;

  /* GET THE COMMAND LINE ARGUMENT(S) */

  if (argc == 1)
    {
      /* help for options... */
      USAGE;
      printf("Included: ");
#if (RF_SUPPORT)
      printf("RF support, ");
#endif
#if (REFLECTIONS)
      printf("boundary reflections, ");
#endif
      printf("segmentation");
      printf("\n\n");

      return 1;
    }
  else
    for (i = 1; i < argc; i++)
      {
	if (argv[i][0] == '-')	  /* introduces option flags */
	  {
	    /* search through command line option words to find
	     * meaningful chars... */

	    for (j = 1; argv[i][j] != '\0'; j++)
	      switch (argv[i][j])
		{
		  case 'm':
		  case 'M':
		    system.flags.mirror = 1;
		    printf("Tissue-air boundaries will be mirrored.\n");
		    break;
		  case 'n':
		  case 'N':
		    system.flags.gen_twopt = 0;
		    printf("No 2pt will be generated.\n");
		    break;
		  case 's':
		  case 'S':
		    system.flags.source_move = 0;
		    printf("Source photons will not be moved: may crash.\n");
		    break;
		  case 'a':
		  case 'A':
		    system.flags.move_before_NA = 0;
		    printf("Src photon moving will be applied after NA diffusion\n");
		    break;
		  case 't':
		  case 'T':
		    system.flags.exact_exit_time = 1;
		    printf("History file will record exact exit times.\n");
		    break;
                  case 'h':
                  case 'H':
                  case '?':
                    USAGE;
                    return 0;
		  default:
		    printf("Unknown option -%c ignored\n", argv[i][j]);
		}
	  }
	else
	  {
	    /* Assume that since it isn't a command-line flag that it's
	     *  the name of the configuration file.  Diddle the argument
	     *  vector to fool the parser into doing the right thing.
	     */

	    argv[1] = argv[i];
	    argv[2] = NULL;	  /* may or may not be necessary */

	    argc = 2;

	    break;
	  }
      }

  /* PARSE THE INPUT FILE */

  if ((n = loadconfig(&system, argc, argv)) != 0)
    exit(n);

  /* Rescale the Universe to get unit grid spacing.  Units are
   * specified in mm in the input file.  system.dl has dimension of
   * mm/voxel, which takes everything into voxel units 
   */

  /* This is the only function where system.* should be modified. */

  assert(system.dl > 0);

  system.v0    /= system.dl;
  system.k0    *= system.dl;

  system.Lmax  /= system.dl;
  system.Lmin  /= system.dl;
  system.stepL /= system.dl;

  for (i = 1; i <= system.Ntissue; i++) 
    {
      system.tmus[i] *= system.dl;
      system.tmua[i] *= system.dl;
    }

  for (i = 0; i < system.nDets; i++) 
    {
      system.detLoc[i][0] /= system.dl;
      system.detLoc[i][1] /= system.dl;
      system.detLoc[i][2] /= system.dl;

      system.detRad[i]    /= system.dl;
    }

  for (i = 0; i < system.nSrcs; i++) 
    {
      system.srcLoc[i][0] /= system.dl;
      system.srcLoc[i][1] /= system.dl;
      system.srcLoc[i][2] /= system.dl;

      system.srcRad[i]    /= system.dl;
    }

  system.xstep  /= system.dl;   /* Voxels now have unit dimensions */
  system.ystep  /= system.dl;   /* currently these are never used */
  system.zstep  /= system.dl;

  assert(system.xstep == 1);
  assert(system.ystep == 1);
  assert(system.zstep == 1);

  /* READ IN THE SEGMENTED DATA FILE */

  if ((n = loadsegments(&system, &tissueType)) != 0)
    exit(n);

  /* ALLOCATE SPACE FOR PHOTON FLUENCE */

  if (system.flags.gen_twopt) 
    {
      II = (double *)malloc(system.nIxstep * system.nIystep *
                            system.nIzstep * system.nTstep  * sizeof(double));

      if (II == NULL) 
        {
          fprintf(stderr, 
                  "ERROR: malloc() failed, can't allocate %lu bytes\n",
                  (unsigned long)system.nIzstep * system.nIxstep *
                  system.nIystep * system.nTstep  * sizeof(double));

          exit(1);
        }

#if (RF_SUPPORT)
      JJ = (double *)malloc(system.nIzstep * system.nIxstep *
                            system.nIystep * system.nTstep  * sizeof(double));

      if (JJ == NULL) 
        {
          fprintf(stderr, 
                  "ERROR: malloc() failed, can't allocate %lu bytes\n",
                  (unsigned long)system.nIzstep * system.nIxstep *
                  system.nIystep * system.nTstep  * sizeof(double));

          exit(1);
        }
#else
      JJ = NULL;
#endif

      /* array size... */
      n = system.nIxstep * system.nIystep * system.nIzstep * system.nTstep;

    }
  else
    {
      II = NULL;
      JJ = NULL;
    }

  /* Set up the floating point co-processor */

  fpusetup();



  /* Choose base his filename... */
  if (system.nDets == 0)    {
    printf("No detectors: writing all exiting photons to %s.his.a*\n",
	   argv[1]);
    snprintf(filenm, FILENMLEN, "%s.his.a", argv[1]);
  } else if (system.flags.exact_exit_time) {
    printf("Writing exact exit times to %s.his.t*\n",argv[1]);
    snprintf(filenm, FILENMLEN, "%s.his.t", argv[1]);
  } else {
    printf("Writing exiting tgate number to %s.his*\n",argv[1]);
    snprintf(filenm, FILENMLEN, "%s.his", argv[1]);
  }
  if (system.nSrcs > 1) {
    printf("%d sources, so multiple his and 2pt files with suffix\n"
	   "\tnumbered 0-%d will be written.\n", system.nSrcs,
	   system.nSrcs-1);
  }

  /* --------------------------- main loop over sources ---------------- */
  for (s=0; s<system.nSrcs; ++s) {
    
    /* zero the fluence... */
    if (system.flags.gen_twopt) 
      for (i = 0; i < n; i++) {
	II[i] = 0.0;
#if (RF_SUPPORT)
	JJ[i] = 0.0;
#endif
      }
    
    /* open history file... */
    if (system.nSrcs > 1)
      snprintf(filenm_s, FILENMLEN, "%s%d", filenm, s);
    else
      snprintf(filenm_s, FILENMLEN, "%s", filenm);
    if ((fp = fopen(filenm_s, "wb")) == NULL) 
      {
	fprintf(stderr, "ERROR: Can't open history file %s for writing\n", \
		filenm_s);
	exit(1);
      }
    
    /* RUN THE MONTE CARLO SIMULATION, passing file pointer for his output */
    run_simulation(&system, (const TISSUE ***)tissueType, II, JJ, fp, s);
    
    /* CLOSE HISTORY FILE */
    fclose(fp);
    
    /* Save fluence to disk... */
    if (system.flags.gen_twopt) 
      {
	/* Print some useful information if you're manipulating tables
	 *  by hand from inside matlab */
      printf("\twriting 2pt file of size %d x %d x %d x %d\n",
             system.nIxstep, system.nIystep, system.nIzstep, system.nTstep);
    
      if (system.nSrcs > 1)
	snprintf(filenm_s, FILENMLEN, "%s.2pt.%d", argv[1], s);
      else
	snprintf(filenm_s, FILENMLEN, "%s.2pt", argv[1]);
      save_2pt_err = save_2pt_err || save2pt(filenm_s, II, JJ,
					     system.nIxstep * system.nIystep *
					     system.nIzstep * system.nTstep);
      }

  } /* ------------------------- end main loop over sources --------------- */

  return save_2pt_err;
}

/*********************************************************
  OPEN AND READ THE INPUT FILE

  Returns non-zero on error
    1 -> File error (can't open file, unexpected EOF, etc.)
    2 -> Invalid input parameter
    3 -> System error (out of memory, etc.)

*********************************************************/

int loadconfig(struct Config *p, int argc, char *argv[])
{
  char filenm[FILENMLEN];
  FILE *fp;
  int i, n;

  /* Initialize the config structure - Jon: why isn't this done in parseconfig? */

  p->Ntissue = 0;

  for (i = 0; i < MAXTISSUE; i++)
    {
      p->tmus[i] = MUS_AIR;
      p->tmua[i] = MUA_AIR;
      p->tg[i]   =   G_AIR;
      p->tn[i]   = IDX_AIR;
    }

  p->nDets = 0;

  for (i = 0; i < MAXDET; i++)
    {
      p->detLoc[i][0] = -1;
      p->detLoc[i][1] = -1;
      p->detLoc[i][2] = -1;
       
      p->detDir[i][0] = -1;
      p->detDir[i][1] = -1;
      p->detDir[i][2] = -1;
       
      p->detRad[i]    = -1;
      p->detNA[i]     = -1;
    }

  p->nSrcs = 0;

  for (i = 0; i < MAXDET; i++)
    {
      p->srcLoc[i][0] = -1;
      p->srcLoc[i][1] = -1;
      p->srcLoc[i][2] = -1;
       
      p->srcDir[i][0] = -1;
      p->srcDir[i][1] = -1;
      p->srcDir[i][2] = -1;
       
      p->srcRad[i]    = -1;
      p->srcNA[i]     = -1;
    }

  /* Open input file */

  snprintf(filenm, FILENMLEN, "%s.cfg", argv[1]);

  if ((fp = fopen(filenm, "rt")) == NULL) 
    {
      fprintf(stderr, 
              "input_file \"%s\" not found, assuming you meant \"%s\"\n",
              filenm, argv[1]);

      if ((fp = fopen(argv[1], "rt")) == NULL) 
        {
          fprintf(stderr, "usage: tMCimg input_file (.cfg assumed)\n");
          fprintf(stderr, "input_file = %s does not exist.\n", filenm);
          return 1;
        } 
      else 
        {
          /* chop off the extension */
          char *p = rindex(argv[1], '.');

          if (p != NULL)
            *p = '\0';
        }
    }

  /* READ THE INPUT FILE */

  n = parseconfig(p, fp);

  /* CLOSE FILE AND RETURN */

  fclose(fp);
  return n;
}

/*********************************************************
        OPEN AND READ THE SEGMENTATION FILE

  Returns non-zero on error
    1 -> File error (can't open file, unexpected EOF, etc.)
    2 -> Invalid input parameter
    3 -> System error (out of memory, etc.)

*********************************************************/

int loadsegments(const struct Config *p, TISSUE ****tt)
{
  TISSUE ***tissueType, t;
  FILE *fp;
  int i, j, k;

  assert(p != NULL);
  assert(tt != NULL);

  if ((fp = fopen(p->segFile, "rb")) == NULL) 
    {
      fprintf(stderr, "ERROR: THe binary image file %s was not found!\n", 
              p->segFile);
      return 1;
    }

  rewind(fp);			  /* Check actual vs expected file sizes */
  fseek(fp, 0, SEEK_END);

  if (ftell(fp) != p->nxstep * p->nystep * p->nzstep * sizeof(TISSUE)) 
    {
      fprintf(stderr, 
              "Size of segmentation file != nxstep*nystep*nzstep elements\n");
      return 1;
    } 
  else
    rewind(fp);

  if ((*tt = (TISSUE ***)malloc(p->nxstep * sizeof(TISSUE **))) == NULL) 
    {
      fprintf(stderr, "malloc() failed, can't allocate %d bytes\n",
              p->nxstep * sizeof(TISSUE **));

      return 3;
    }

  tissueType = *tt;

  for (i = 0; i < p->nxstep; i++) 
    {
      tissueType[i] = (TISSUE **)malloc(p->nystep * sizeof(TISSUE *));

      if (tissueType[i] == NULL) 
        {
          fprintf(stderr, "malloc() failed, can't allocate %d bytes\n",
                  p->nystep * sizeof(TISSUE *));
          return 3;
        } 
      else
        for (j = 0; j < p->nystep; j++) 
          {
            tissueType[i][j] = (TISSUE *)malloc(p->nzstep * sizeof(TISSUE));
            
            if (tissueType[i][j] == NULL) 
              {
                fprintf(stderr,
                        "malloc() failed, can't allocate %d bytes\n",
                        p->nzstep * sizeof(TISSUE));
                return 3;
              }
          }
    }

  /* Matlab stores data column [Y,X,Z], I use data row
   * [X,Y,Z] first.  Flip the data around as I read it in.
   */

  for (k = 0; k < p->nzstep; k++)
    {
      for (i = 0; i < p->nxstep; i++) 
        {
          for (j = 0; j < p->nystep; j++) 
            {
              if (fread(&t, sizeof(TISSUE), 1, fp) != 1) 
                { 
                  fprintf(stderr, "Unexpected end of segmentation file %s", 
                          p->segFile);
                  return 1;
                }  
              else 
                tissueType[i][j][k] = t;
            }
        }

      /* Print "." as every plane is read to provide some feedback */

      printf(".");
      fflush(stdout);
    }

  printf("\n");

  fclose(fp);

  return 0;
}

/* If used, causes certain invalid mathematical operations to generate
 *  a SIGFPE and dump core rather than silently proceeding.  For 
 *  normal operation, this should never happen.  It is most useful,
 *  therefore, in tracking down subtle numerical bugs.
 *
 * If DO_FPU_CONTROL is not defined, this simply returns without
 *  changing the state of the FPU
 */

#if (defined __linux__) && (defined __i386__)
# ifndef INHIBIT_FPU_CONTROL
# include <fpu_control.h>
/* I want to catch certain math errors that normally fail silently */
/* This file defines _FPU_CONTROL_H for me, which is used below */
# define DO_FPU_CONTROL 1
# endif
#endif

void fpusetup(void)
{
#if (DO_FPU_CONTROL)
  fpu_control_t fpu;

  _FPU_GETCW(fpu);

  /*
   * If not masked off (as is the default), these raise an interrupt
   * which gets turned into a SIGFPE by the C runtime libraries.
   *
   * _FPU_MASK_IM: Invalid operation mask
   * _FPU_MASK_DM: Denormalized operand mask
   * _FPU_MASK_ZM: Zero-divide mask
   * _FPU_MASK_OM: Overflow mask
   * _FPU_MASK_UM: Underflow mask
   * _FPU_MASK_PM: Precision (inexact result) mask
   */

  /* Take off some of the masks */
  fpu &= ~(_FPU_MASK_IM | _FPU_MASK_ZM |
           _FPU_MASK_OM | _FPU_MASK_UM | _FPU_MASK_DM);

  fpu &= ~(_FPU_EXTENDED);
  fpu |= _FPU_DOUBLE;

  _FPU_SETCW(fpu);
#endif

  return;
}







/* =========================================== SIMULATION =================
 ==========================================================================
*/

void run_simulation(const struct Config *system, const TISSUE ***tissueType,
                    double *II, double *JJ, FILE *fp, int s)

     /* s is source number to launch N photons from
	Added barnett 7/16/02
     */
{
  double xnew, ynew, znew;
  int l, i, idum;
  int N;			  /* NUMBER OF PHOTONS RUN SO FAR */
  int VoxN;			  /* INDEX TO VOXEL CONTAINING PHOTON */
  double this_mus;
  int    tissueIndex, tissueIndex_new;
  double sn1, sn2, sn3;
  double ctheta, ctheta_out, ctheta_c2;
  double R_fresnel, R_fresnel_1, R_fresnel_2;
  double x, y, z;		  /* CURRENT PHOTON POSITION */
  double xold, yold, zold;	  /* LAST POSITION OF PHOTON */
  double c1, c2, c3;		  /* DIRECTION COSINES */
  double ic1, ic2, ic3;           /* INITIAL (pre change due to NA)
				     DIRECTION COSINES */
  double Lresid, Ltot, Lnext, step;
  double Ttot;
  int tindex;
  double P2pt;			  /* PHOTON WEIGHT */
#if (RF_SUPPORT)
  double Q2pt;
#else
  double mu_l;
#endif
  float lenTiss[MAXTISSUE];	  /* LENGTH SPENT IN EACH TISSUE */
  double rnm;			  /* RANDOM NUMBER */

  /* NA and finite radius workspace... */
  double th0, dth;
  double ph0, dph;

  /* NUMBER PHOTONS EXECUTED SO FAR */
  N = 0;

  if (system->seed < 0)
    idum = (int)system->seed;
  else
    idum = -(int)system->seed;

  printf("run_simulation: src %d...\n", s);

  /*******************************************************************
   * START MIGRATING THE PHOTON GENERATING PHOTONS UNTIL NUMBER OF
   * PHOTONS EXECUTED (N) IS EQUAL TO THE NUMBER TO BE GENERATED (NT)
   *******************************************************************/

  while (N < system->NT) {
    ++N;

    if (system->NT >= 10 && (N % (system->NT / 10)) == 0) 
      {
        printf("\tcompleted %d of %d photons\n", N, system->NT);
        fflush(stdout);
      }

    /* SET THE PHOTON WEIGHT TO 1 AND INITIALIZE PHOTON LENGTH PARAMETERS */

    P2pt   = 1.0;
#if (RF_SUPPORT)
    Q2pt   = 0.0;
#endif

    Ltot   = 0.0;
    Ttot   = 0.0;
    Lnext  = 0.0; /* Was 1.0, but I think it should be 0.0 -JS */
    Lresid = 0.0;

    /* INITIALIZE THE LENGTH IN EACH TISSUE TYPE */

    for (i = 0; i <= system->Ntissue; i++)
      lenTiss[i] = 0.0;



    /* INITIAL DIRECTION OF PHOTON ...................................... */

    c1 = system->srcDir[s][0];
    c2 = system->srcDir[s][1];
    c3 = system->srcDir[s][2];
    ic1 = c1;
    ic2 = c2;
    ic3 = c3;

    /* report if direction vector wrong length... */
    assert((c1 >= -1) && (c1 <= 1));
    assert((c2 >= -1) && (c2 <= 1));
    assert((c3 >= -1) && (c3 <= 1));
    assert(SQR(c1)+SQR(c2)+SQR(c3) > 0.9999);
    assert(SQR(c1)+SQR(c2)+SQR(c3) < 1.0001);

    
    /* The original input direction can be defined as
       the vector (0,0,1) times a pair of rotation matrices
       characterized by these two angles.
       Generate theta and phi for this rotation... */
    th0 = acos(c3);
    ph0 = atan2(c2, c1);

    /* Allow for finite source NA */
    if (system->srcNA[s] > 0) {

      /* Change in the propagation direction, from z axis... */
      dth = system->srcNA[s] * ran(&idum);
      dph = 2*PI*ran(&idum);
      
      /* Convert perturbation into direction cosines then rotate it... */
      c1 = sin(dth) * cos(dph);
      c2 = sin(dth) * sin(dph);
      c3 = cos(dth);
      rotate(&c1, &c2, &c3, th0, ph0);

      /* check rotated length... */
      assert((c1 >= -1) && (c1 <= 1));
      assert((c2 >= -1) && (c2 <= 1));
      assert((c3 >= -1) && (c3 <= 1));
      assert(SQR(c1)+SQR(c2)+SQR(c3) > 0.9999);
      assert(SQR(c1)+SQR(c2)+SQR(c3) < 1.0001);
    }


    /* INITIAL PHOTON POSITION ............................................ */

    x = system->srcLoc[s][0];
    y = system->srcLoc[s][1];
    z = system->srcLoc[s][2];

    /* printf("initial src = (%f %f %f), tissue %d\n",x,y,z,\
       GET_TISSUE(x, y, z, system, tissueType)); */

    /* Allow for finite source radius: move photon to side of original dir */
    if (system->srcRad[s] > 0)
      {
        double du, dv, dw;

        do {
          /* Generate on a square then reject based on a circle */
          du = 2 * ran(&idum) - 1;
          dv = 2 * ran(&idum) - 1;
        } while (du * du + dv * dv > 1.0);

	/* displacement in rotated coord sys is (du, dv, dw), rotate back... */
        du = du * (system->srcRad[s]);
        dv = dv * (system->srcRad[s]);
	dw = 0.0;
	rotate(&du, &dv, &dw, th0, ph0);

	x += du;
	y += dv;
	z += dw;

      }

    if (system->flags.source_move) {
      /* move along direction vector ... either pre-or post-NA diffusion */
      
      if (system->flags.move_before_NA) {
	/* move photon so just inside tissue, using original direction... */
	if (move_src_just_inside(&x, &y, &z, ic1, ic2, ic3, system,
				 tissueType))
	  continue;
      } else {
	/* move using new, NA-diffused, direction, representing a true fiber
	   end held for instance a certain distance away from surface... */
	if (move_src_just_inside(&x, &y, &z, c1, c2, c3, system,
				 tissueType))
	  continue;
      }
    }
	  

    /* printf("src after finite rad + move = (%f %f %f), tissue %d\n",x,y,z,\
       GET_TISSUE(x, y, z, system, tissueType)); */

    xold = x;
    yold = y;
    zold = z;

    tissueIndex = GET_TISSUE(x, y, z, system, tissueType);

    /* make sure photon doesn't start in air... it shouldn't by now! */
    assert(tissueIndex != 0);

    /* Loop over scattering events:
     * UNTIL TIME EXCEEDS GATE OR PHOTON ESCAPES (leaves whole volume
     * or finds itself in air) 
     */

    while (Ttot < system->Tmax && IN_VOLUME(x, y, z, system) 
           && IN_TISSUE(tissueIndex)) {

      /* CALCULATE NEW SCATTERING LENGTH */

      rnm = ran(&idum);

      /* PROPAGATE THE PHOTON */

      /* Lresid is an exponentially distributed random number with a
         mean of 1.0.  It is also is the dimensionless quantity mu*x */

      Lresid = -log(rnm);

      while ((Ttot < system->Tmax) && (Lresid > 0.0) &&
             IN_VOLUME(x, y, z, system) && IN_TISSUE(tissueIndex)) {

        this_mus = system->tmus[tissueIndex]; /* Compute length of this step */

        if (this_mus > Lresid) 
          {
            step = Lresid / this_mus;
            Lresid = 0;
          } 
        else 
          {
            step = 1.0;
            Lresid -= this_mus;
          }

        /* Score the pixel in this voxel as usual, then decide
         * what the destination will be.  Keep the position in
         * temporary variables until I know where the final
         * voxel for this iteration is */

        if (system->flags.gen_twopt 
            && (Ltot >= Lnext) && (Ttot > system->Tmin)) 
          {
            /* count photon in 2pt array... */
            if (IN_ROI(x, y, z, system)) 
              {
                tindex = (int)((Ttot - system->Tmin) / system->stepT);

                /* Use Matlab ordering so I can dump to disk 
                 * the entire block of memory later 
                 */
                VoxN = (((int)y - system->Iymin) + system->nIystep * 
                        (((int)x - system->Ixmin) + system->nIxstep *
                         (((int)z - system->Izmin) + system->nIzstep * 
                          (tindex))));

                assert(tindex < system->nTstep);
                assert(VoxN < (system->nIxstep * system->nIystep *
                               system->nIzstep * system->nTstep));

                II[VoxN] += P2pt;
#if (RF_SUPPORT)
                JJ[VoxN] += Q2pt;
#endif
              }

            Lnext += 1.0;      /* Don't double-score a photon in each voxel */
          }

        /* Start updating the position */

        Ltot += step;

        if (IN_TISSUE(tissueIndex))
          Ttot += step / (system->v0 / system->tn[tissueIndex]);
        else
          Ttot += step / (system->v0 / 1.0);

        lenTiss[tissueIndex] += (float)step;

        if (system->flags.gen_twopt) 
          {
            /* Compute new amplitude [and phase] from updated lenTiss */
#if (RF_SUPPORT)
            rotate_phase(&P2pt, &Q2pt, system, lenTiss);
#else
            mu_l = 0.0;

            for (l = 1; l <= system->Ntissue; l++)
              mu_l += system->tmua[l]*lenTiss[l];

            P2pt = exp(-mu_l);
#endif
          }

        xnew = x + c1 * step;
        ynew = y + c2 * step;
        znew = z + c3 * step;

        if (! IN_VOLUME(xnew, ynew, znew, system)) 
          {
            /* Stepped out of system, photon is officially gone */

            x = xnew;
            y = ynew;
            z = znew;
          } 
        else if ((GET_TISSUE(xnew, ynew, znew,
                              system, tissueType) == tissueIndex) ||
                 INDEX_MATCHED(xnew, ynew, znew,
                               system, tissueType, tissueIndex)) 
          {
            /* Trivial case, either same tissue or at least the
             * same index of refraction before and after
             * stepping, so no possibility for reflection */

            x = xnew;
            y = ynew;
            z = znew;
          } 
        else 
          {
            double nold, nnew;
            int    norm;

            /* Tissue changed, need to worry about reflections */

            assert(IN_VOLUME(xnew, ynew, znew, system));

            tissueIndex_new = GET_TISSUE(xnew, ynew, znew, 
                                          system, tissueType);

#if (REFLECTIONS)
            nold = system->tn[tissueIndex];
            nnew = system->tn[tissueIndex_new];

            /* Change place-holder values into 1.0 */

            if (nold == IDX_AIR)
              nold = 1.0;
            if (nnew == IDX_AIR)
              nnew = 1.0;

            /* 2.75 is physically reasonable, but arbitrary */
            assert((nold >= 1.0) && (nold < 2.75));
            assert((nnew >= 1.0) && (nnew < 2.75));

            /* CALCULATE SURFACE NORMAL VECTOR OUT OF TISSUE */

            /* There are really only 3 possible cases using
             * our linearized approach
             *
             * 1) Along a single coordinate axis (cube face)
             * 2) Pair of coordinate axis (cube edge)
             * 3) Along 3 coordinate axes (cube corner)
             * 
             * Take the interface to be orthogonal to on
             * of these three instances 
             */

            sn1 = (int)xnew - (int)x;
            sn2 = (int)ynew - (int)y;
            sn3 = (int)znew - (int)z;

            /* inew-i = 0 or +/- 1, so the square is always 0 or 1 */
            norm = SQR(sn1) + SQR(sn2) + SQR(sn3);

            /* Because its computed from integers, the normalization
               factors can be specified excatly in advance - saves time */

            switch (norm)
              {
              case 1:
                break;                /* Already normalized */
              case 2:
                sn1 /= SQRT2;
                sn2 /= SQRT2;
                sn3 /= SQRT2;
                break;
              case 3:
                sn1 /= SQRT3;
                sn2 /= SQRT3;
                sn3 /= SQRT3;
                break;
              default:
                assert((norm != 1) && (norm != 2) && (norm != 3));
              }

            /* cosine of angle of incidence */
            ctheta = sn1 * c1 + sn2 * c2 + sn3 * c3;

            /* squared cosine of critical angle (incident) */
            ctheta_c2 = 1 - SQR(nnew) / SQR(nold);

            if (SQR(ctheta) < ctheta_c2) 
              {
                R_fresnel  = 1;	  /* Total internal reflection */
                ctheta_out = -999;
              } 
            else if (system->flags.mirror && 
                     (GET_TISSUE(xnew, ynew, znew, system, tissueType) == 0))
              {
                R_fresnel  = 1;	  /* Reflecting boundaries to air */
                ctheta_out = -999;
              } 
            else 
              {
                /* cosine of angle transmitted at */
                ctheta_out = sqrt(1 - (SQR(nold)/SQR(nnew))*(1 - SQR(ctheta)));

                R_fresnel_1 = ((nold * ctheta_out - nnew * ctheta) /
                               (nold * ctheta_out + nnew * ctheta));
                R_fresnel_2 = ((nold * ctheta - nnew * ctheta_out) /
                               (nold * ctheta + nnew * ctheta_out));

                R_fresnel = (R_fresnel_1 * R_fresnel_1 +
                             R_fresnel_2 * R_fresnel_2) / 2;
              }

            /* Reflect if ran() < R_fresnel */

            rnm = ran(&idum);

            if (rnm <= R_fresnel) 
              {
                /* Reflection occurred, return back to previous
                 * position.  Still score the iteration as a step in
                 * this voxel and after the reflection we'll pick up
                 * the second step.  Flip the direction vector
                 * before we leave */

                c1 -= 2 * ctheta * sn1;
                c2 -= 2 * ctheta * sn2;
                c3 -= 2 * ctheta * sn3;

                assert((c1 >= -1) && (c1 <= 1));
                assert((c2 >= -1) && (c2 <= 1));
                assert((c3 >= -1) && (c3 <= 1));
                assert(SQR(c1)+SQR(c2)+SQR(c3) > 0.9999);
                assert(SQR(c1)+SQR(c2)+SQR(c3) < 1.0001);
              } 
            else 
              {
                /* Transmission occured, move to new position */

                x = xnew;
                y = ynew;
                z = znew;

                /* Refraction */
                assert(ctheta_out != -999);

                /* Subtract cos(theta), leaving sin(theta) terms */
                c1 -= ctheta * sn1;
                c2 -= ctheta * sn2;
                c3 -= ctheta * sn3;

                /* Rescale sin(theta), preserving sin(phi),cos(phi)
                   to get sin(theta') [E_{||} terms] */

                c1 *= nold/nnew;
                c2 *= nold/nnew;
                c3 *= nold/nnew;

                /* Add back in cos(theta) [E_\perp terms] */

                c1 += ctheta_out * sn1;
                c2 += ctheta_out * sn2;
                c3 += ctheta_out * sn3;

                assert((c1 >= -1) && (c1 <= 1));
                assert((c2 >= -1) && (c2 <= 1));
                assert((c3 >= -1) && (c3 <= 1));
                assert(SQR(c1)+SQR(c2)+SQR(c3) > 0.9999);
                assert(SQR(c1)+SQR(c2)+SQR(c3) < 1.0001);
              }
#else  /* not REFLECTIONS */
            x = xnew;
            y = ynew;
            z = znew;
#endif
          }

        /* Update tissue type after making this step */
        tissueIndex = GET_TISSUE(x, y, z, system, tissueType);

        if ((Lresid == 0) && IN_VOLUME(x, y, z, system)
            && IN_TISSUE(tissueIndex)) 
          {
            /* CALCULATE THE NEW SCATTERING ANGLE USING HENYEY-GREENSTEIN */

            scatter(system->tg[tissueIndex], &c1, &c2, &c3, &idum);
          }
      }			  /* PROPAGATE PHOTON */
    }			  /* LOOP UNTIL END OF SINGLE PHOTON */

    /* SCORE EXITING PHOTON */

    tissueIndex = GET_TISSUE(x, y, z, system, tissueType);

    if (!IN_TISSUE(tissueIndex) && (Ttot < system->Tmax)) 
      {
        if (system->flags.gen_twopt) 
          {
            if (IN_ROI(x, y, z, system))
              {
                tindex = (int)((Ttot - system->Tmin) / system->stepT);

                /* Use Matlab ordering so I can dump to disk 
                 * the entire block of memory later 
                 */
                VoxN = (((int)y - system->Iymin) + system->nIystep * 
                        (((int)x - system->Ixmin) + system->nIxstep *
                         (((int)z - system->Izmin) + system->nIzstep * 
                          (tindex))));

                assert(tindex < system->nTstep);
                assert(VoxN < (system->nIxstep * system->nIystep *
                               system->nIzstep * system->nTstep));

                II[VoxN] -= P2pt;
#if (RF_SUPPORT)
                JJ[VoxN] -= Q2pt;
#endif
              }
          }

        if (IN_VOLUME(x, y, z, system))
          {
            /* BUG: Photons come out some distance into the "air"
             * (exponentially distributed according to the previous
             * tissue; this is probably a bug everywhere in the
             * program).  When the grid spacing is large, this can
             * leave photons a sizable distance from the interface,
             * often larger than the detector radius itself, which
             * means these photons never get counted even though they
             * "hit" the detector along the way!  */

            /* SAVE TO HISTORY FILE */

            savehis(system, fp, x, y, z, Ttot, lenTiss);
          }
      }
  }			  /* LOOP UNTIL ALL PHOTONS EXHAUSTED */
}

/**************************************************************
  HANDLE PHASE ROTATION BETWEEN REAL AND IMAGINARY PART
**************************************************************/

/* Originally, this handled just the increment due to the current
 *  step.  That proved sufficiently non-robust against round-off
 *  errors to warrent recomputing the photon amplitude and phase
 *  every step of the simulation.  Since I was already computing
 *  an exp(), a sin(), and a cos(), this is not significantly
 *  slower than the old way, which is also nice.
 */

void rotate_phase(double *u, double *v, 
                  const struct Config *p, const float *lenTiss)
{
  double len, phi, mag;
  int i;

  len = 0.0;
  phi = 0.0;

  for (i = 1; i <= p->Ntissue; i++)
    {
      len += lenTiss[i]*p->tmua[i];          /* real part */
      phi += lenTiss[i]*(p->tn[i]*p->k0);    /* imaginary part */
    }
  
  mag = exp(-len);

  *u = mag * cos(phi);
  *v = mag * sin(phi);

  assert(SQR(*u) + SQR(*v) > 0);

  return;
}

/**************************************************************
 CALCULATE THE NEW SCATTERING ANGLE USING HENYEY-GREENSTEIN
**************************************************************/

void scatter(double gg, double *c1, double *c2, double *c3, int *idum)
{
  /* Make sure the anisotropy is physically meaningful */

  assert((gg >= -0.5) && (gg <= 1.0));

  /* Sanity-check the input direction cosines */

  assert((*c1 >= -1) && (*c1 <= 1));
  assert((*c2 >= -1) && (*c2 <= 1));
  assert((*c3 >= -1) && (*c3 <= 1));
  assert(SQR(*c1)+SQR(*c2)+SQR(*c3) > 0.9999);
  assert(SQR(*c1)+SQR(*c2)+SQR(*c3) < 1.0001);

  /* If gg is identically zero, the calculated distribution has a
   * divide-by-zero, but gg==0 is just the isotropic case, which is
   * trivial to handle as a special case */

  if (gg == 0.0) 
    {
      double nrm, C1, C2, C3;

      do {
        C1 = 2*ran(idum) - 1;
        C2 = 2*ran(idum) - 1;
        C3 = 2*ran(idum) - 1;
      } while (SQR(C1) + SQR(C2) + SQR(C3) > 1.0);

      /* (c1,c2,c3) must be a unit vector, rescale vector length */
      nrm = sqrt(SQR(C1) + SQR(C2) + SQR(C3));
      assert(nrm > 0.0);

      *c1 = C1 / nrm;
      *c2 = C2 / nrm;
      *c3 = C3 / nrm;
    } 
  else 
    {                             /* gg != 0 */
      double c1o, c2o, c3o;	  /* OLD DIRECTION COSINES */
      double phi, cphi, sphi, theta, stheta, ctheta;
      double rnm, nrm, foo;

      c1o = *c1;
      c2o = *c2;
      c3o = *c3;

      phi  = 2 * PI * ran(idum);  /* Scatter randomly in transverse plane */
      cphi = cos(phi);
      sphi = sin(phi);

      rnm = ran(idum);            /* Scatter with H-G dist in prop. dir */

      foo = (1.0 - SQR(gg)) / (1.0 - gg + 2.0 * gg * rnm);
      foo = SQR(foo);
      ctheta = (1.0 + SQR(gg) - foo) / (2.0 * gg);

      assert((ctheta >= -1) && (ctheta <= 1));

      theta  = acos(ctheta);
      stheta = sin(theta);

      if ((*c3 > -1) && (*c3 < 1)) 
        {
          double bar = sqrt(1 - SQR(c3o));
          assert((bar > 0) && (bar <= 1.0));

#if 0
          *c1 =  stheta * (c1o * c3o * cphi - c2o * sphi) / bar
               + ctheta * c1o;
          *c2 =  stheta * (c2o * c3o * cphi + c1o * sphi) / bar
               + ctheta * c2o;
          *c3 = -stheta * cphi * bar
               + ctheta * c3o;
#else
          /* Bar on the order 1e-6 causes real problems.  Multiply
           * through by bar and renormalize to 1.0 at the end */

          *c1 =  stheta * (c1o * c3o * cphi - c2o * sphi)
               + ctheta * c1o * bar;
          *c2 =  stheta * (c2o * c3o * cphi + c1o * sphi)
               + ctheta * c2o * bar;
          *c3 = -stheta * cphi * SQR(bar)
               + ctheta * c3o * bar;

          nrm = sqrt(SQR(*c1) + SQR(*c2) + SQR(*c3));

          *c1 /= nrm;
          *c2 /= nrm;
          *c3 /= nrm;
#endif
        } 
      else 
        {
          /* Vertical direction causes problems, treat as a special case */

          *c1 = stheta * cphi;
          *c2 = stheta * sphi;
          *c3 = ctheta * c3o;
        } 
    }

  /* Sanity-check the output direction cosines */

  assert((*c1 >= -1) && (*c1 <= 1));
  assert((*c2 >= -1) && (*c2 <= 1));
  assert((*c3 >= -1) && (*c3 <= 1));
  assert(SQR(*c1)+SQR(*c2)+SQR(*c3) > 0.9999);
  assert(SQR(*c1)+SQR(*c2)+SQR(*c3) < 1.0001);

  return;
}

/*********************************************************
              WRITE FLUENCE DATA TO DISK  

  Returns non-zero on error
    1 -> File error (can't open file, unexpected EOF, etc.)
    2 -> Invalid input parameter
    3 -> System error (out of memory, etc.)
*********************************************************/

int save2pt(const char *filenm, const double *udata, 
            const double *vdata, int ndata)
{
  FILE *fp;
  int nu, nv;

  if ((fp = fopen(filenm, "wb")) == NULL) 
    {
      perror("Can't open 2pt file for writing\n");
      return 1;
    }

  assert(udata != NULL);

  nu = fwrite(udata, sizeof(double), ndata, fp);

  if (nu != ndata) 
    {
      fclose(fp);
      perror("Error while writing 2pt file\n");
      return 1;
    }

  /* vdata == NULL is allowed for CW data */

  if (vdata != NULL) 
    if ((nv = fwrite(vdata, sizeof(double), ndata, fp)) != ndata) 
      {
        perror("Error while writing 2pt file\n");
        fclose(fp);
        return 1;
      }

  fclose(fp);

  return 0;
}

/********************************************************************
 SAVE PHOTON TO HISTORY FILE : this converts length back to mm units.
*********************************************************************/

void savehis(const struct Config *system, FILE *fp,
             double x, double y, double z, double T, float *lenTiss)
{
  float buff[4];        /* temporary storage */
  int i, nf, tindex;

  /* Current time gate */

  tindex = (int)((T - system->Tmin) / system->stepT);

  /* IF NO DETECTORS DEFINED THEN SAVE EXIT POSITION 
   * ELSE LOOP THROUGH NUMBER OF DETECTORS */

  for (i = 1; i <= system->Ntissue; i++)
    lenTiss[i] *= system->dl;

  if (system->nDets == 0) 
    {
      /* Convert back to normal dimensions [mm] before saving */

      buff[0] = (float)(x * system->dl);
      buff[1] = (float)(y * system->dl);
      buff[2] = (float)(z * system->dl);
      buff[3] = (float)(T);

      if ((nf = fwrite(&buff, sizeof(float), 4, fp)) != 4) 
        {
          perror("Error writing to history file");
          exit(1);
        }

      if ((nf = fwrite(&lenTiss[1], sizeof(float), 
                       system->Ntissue, fp)) != system->Ntissue) 
        {
          perror("Error writing to history file");
          exit(1);
        }
    } 
  else                        /* system->nDets > 0 */
    {
      for (i = 0; i < system->nDets; i++) 
        {
          double dR = (SQR(x - system->detLoc[i][0]) +
                       SQR(y - system->detLoc[i][1]) +
                       SQR(z - system->detLoc[i][2]));

          /* TODO: Accept/reject based on detector NA...
           *
           * Better (for no photon loss), output a multiplicative
           * factor which gives fraction of photon that gets detected,
           * based on angle 
           */

          if (dR <= SQR(system->detRad[i]))
            {
              buff[0] = (float)i;      /* Detector #  */

              /* use either tgate # or exact photon exit time... */
              if (system->flags.exact_exit_time)
                buff[1] = (float)T;
              else
                buff[1] = (float)tindex;

              /* WRITE TO THE HISTORY FILE */

              if ((nf = fwrite(&buff, sizeof(float), 2, fp)) != 2) 
                {
                  perror("Error writing to history file");
                  exit(1);
                }

              if ((nf = fwrite(&lenTiss[1], sizeof(float), 
                               system->Ntissue, fp)) != system->Ntissue) 
                {
                  perror("Error writing to history file");
                  exit(1);
                }
            }
        }
    }

  return;
}


/* ----------------------------------------------------------------------- */

/* most efficient to have BCK_STEP be geometric mean of typ distance moved
   and FWD_STEP .... */
#define BCK_STEP 0.1
#define FWD_STEP 0.01

int move_src_just_inside(double *x, double *y, double *z,\
			  double cx, double cy, double cz,\
			  const struct Config *system,\
			  const TISSUE ***tissueType)
     /* Moves coords x,y,z of point to be just inside tissue (tissueType!=0)
	along the direction (cx,cy,cz).
	"Just inside" means maximum 1e-3 voxels, negligible for MC code.

	No timing is taken into account for this propagation distance.

	Returns true if failed to hit tissue before exiting system.

	Barnett 7/16/02
     */
{
  TISSUE ti;
  int n_bck = 0, n_fwd = 0, in_vol, houston_we_have_a_problem = 0;
  int bad_photon = 0;
  
  ti = GET_TISSUE(*x, *y, *z, system, tissueType);


  while (IN_TISSUE(ti)) {

    /* Step back to outside tissue, even if means stepping outside system */
    *x -= BCK_STEP * cx;
    *y -= BCK_STEP * cy;
    *z -= BCK_STEP * cz;
    ti = GET_TISSUE(*x, *y, *z, system, tissueType);
    ++n_bck;
  }

  while (!IN_TISSUE(ti) && !houston_we_have_a_problem) {

    in_vol = IN_VOLUME(*x, *y, *z, system);
    /* Step back to outside tissue, even if means stepping outside system */
    *x += FWD_STEP * cx;
    *y += FWD_STEP * cy;
    *z += FWD_STEP * cz;
    ti = GET_TISSUE(*x, *y, *z, system, tissueType);
    /* printf("ti= %d\n",ti); */
    /* raise alarm if exited system having just been inside it... */
    houston_we_have_a_problem = in_vol && !IN_VOLUME(*x, *y, *z, system);
    ++n_fwd;
  }
  
  if (houston_we_have_a_problem) {
    printf("Attempt to launch photon at tissue surface failed: missed tissue!\n"\
	   "Exited system at voxel coords (%f,%f,%f) in tissue %d,\n"\
	   "after %d backward steps then %d forward steps\n",
	   *x, *y, *z, ti, n_bck, n_fwd);
    bad_photon = 1;
  }
  return bad_photon;
}


/* ----------------------------------------------------------------------- */

void rotate(double *x, double *y, double *z, double th0, double ph0)
     
     /* Rotate a vector into direction-based
      * orientation using the same rotation matrices we would have
      * used for the original (a rotation by th about the y-axis,
      * followed by rotation by ph about the z-axis) 
      */
{
  double c1, c2, c3;
  
  c1 =  cos(th0) * (*x) + 0 * (*y) + sin(th0) * (*z);
  c2 =         0 * (*x) + 1 * (*y) +        0 * (*z);
  c3 = -sin(th0) * (*x) + 0 * (*y) + cos(th0) * (*z);
      
  *x =  cos(ph0) * c1 + sin(ph0) * c2 + 0 * c3;
  *y = -sin(ph0) * c1 + cos(ph0) * c2 + 0 * c3;
  *z =         0 * c1 +        0 * c2 + 1 * c3;
}
