/*
 *	program:	tmcv  -   views tMCimg files: .2pt etc.
 *
 *      usage:     tmcv filehead
 *
 *	Alex Barnett  2/7/02
 *                    2/25/02 reads .cfg file for info, using Jon's struct.
 *                    3/1/02 donesn't have to have a .2pt, S-D display.
 *                    3/11/02 multiple sources.
 *                    3/18/02 detector directions, logscale, suffix.
 *                    3/22/02 not just region of interest, tissuetype slice.
 *                    5/23/02 bug fix re overrunning f array in quadstrip.
 */

/* graphics */
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

/* mine */
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <math.h>

/* Jon */
#include <string.h>
#include <assert.h>
#include "tMCimg.h"


/* 2^31 - 1 written as a float */
#define MAXINT		2147483647.0
#define PI		3.14159265358979323846
#define max(a,b)	((a) > (b) ? (a) : (b))
#define min(a,b)	((a) < (b) ? (a) : (b))

/* DATA globals ----- data forced global by GLUT displayfunc --------- */

/* number of voxels */
int nv,nx,ny,nz,nTv,nTx,nTy,nTz,nt;

/* ROI offset in voxels */
int ox,oy,oz;

/* voxel size in mm */
double vs;

/* Jon's types - global data */
struct Config sys;
TISSUE ***tistyp;

/* fluence data */
double *II;


/* GRAPHICS globals ----------------------------------- */

/* mouse params for L+M adjust */
double v,w;

/* fluence view plane number ( goes from 0...nx-1, etc,)
 * timestep number 0...nt-1 */
int ip, tp, it;

/* window size, position */
int width, height, xpos, ypos;

/* old mouse position */
int omx, omy;

/* animation state... */
int freely_spinning,dragging,scaling,paraming,mouse_still_moving;

int helptoggle;
int displaymode;
int smoothmode;
int axesmode;
int planemode;
TISSUE tissueshow;
int dettoggle;
int transmode;

/* eye and angle info */
float ex, ey, ez;
float dist;
float th, ph, rh, xof, yof;
/* 2 * tangent of half-width of view in Y-direction... */
float iris;
float zscale;
float thdot, phdot;


float rc(float);
float gc(float);
float bc(float);


/* help information */
char *helptext[] = {

  " tcmv viewer",
  "",
  " left+middle adjust params v,w",
  " v = ...",
  " w = ...",
  "",
  " left - spin",
  " middle - distance, z-scale",
  " <space> cycle slice display modes",
  " l - transform normal/log/etc",
  " m - normalize zscale sensibly",
  " s - cycle slice / off",
  " n - next time step",
  " z,x - move fluence slice +,-",
  " <Enter> - step time",
  " t - cycle tissue show / slice / off",
  " [,] - move tissue slice +,-",
  " d - toggle detectors&source",
  " q - quit",
  "",
  "  Press c to remove help...",
  NULL };




double endian_8byte(double *in)
/* returns flipped 8 byte double big<->little endian, without destroying orig.
 * Incoming arg is a pointer, returns value. 99/10/15.
 * Bug re boundary alignment fixed 2/7/02
 */
{
        char *inc,*outc;
	double *a;

        inc = (char *)in;
	outc = (char *)a;
        outc[0] = inc[7];
        outc[1] = inc[6];
        outc[2] = inc[5];
        outc[3] = inc[4];
        outc[4] = inc[3];
        outc[5] = inc[2];
        outc[6] = inc[1];
        outc[7] = inc[0];

        return *a;
}





/* ------------------------initialize view-------------------------------- */


void setviewcoords()
{
  /* modes */
  displaymode = 0;
  smoothmode = 1;
  axesmode = 3;
  helptoggle = 0;
  planemode = 2;
  tissueshow = sys.Ntissue + 2; /* show y slice of tissue */
  dettoggle = 1;
  transmode = 0;

  /* eye */
  dist = 300.0;
  th = 0.4;
  ph = 0.7;
  rh = 0;
  xof = 0;
  yof = 0;
  thdot = 0.0;
  phdot = 0.0;
  iris = 0.6;
  ex = xof + dist*sin(th)*cos(ph);
  ey = yof + -dist*cos(th)*cos(ph);
  ez = -dist*sin(ph);

  /* colour */
  zscale = 0.1;

  /* fluence plane view index (0...min(nx,ny,nz)-1) */
  ip = nz/2;
  /* tissue plane view index (0...min(nx,ny,nz)-1) */
  tp = ny/2;
  /* timestep to view (0...nt-1) */
  it = 0;

}


void increment(void)
{
  /* increment view params which can freely drift.
   * In our case, eye angles (and keep them modulo 2*PI)... */
  th += thdot;
  th = (th > 0.0) ? ((th < 2*PI) ? th : th-2*PI) : th+2*PI;
  ph += phdot;
  ph = (ph > 0.0) ? ((ph < 2*PI) ? ph : ph-2*PI) : ph+2*PI;
}


void animate(void)
{
  if (freely_spinning)
    /* only increment view if mouse_motion isn't already doing it (drag) */
    increment();
  else
    /* don't retrigger animation loop if flag unset... */
    glutIdleFunc(NULL);
  
  /* if no motion, stop 'freely spinning' */
  if (thdot == 0.0 && phdot == 0.0)
    freely_spinning = 0;
  
  glutPostRedisplay();
}



/* --------------------------------- axes --------------------------------*/

#define N_AXES_MODES 4

void axes()
     /* plot 3d axes */
{
  /* box: */
  if (axesmode>0) {
    glColor3f(0.3,0.3,1.0); /* blue */
    glPushMatrix();
    glScalef((nx)*sys.dl/2.0,(ny)*sys.dl/2.0,(nz)*sys.dl/2.0);
    glBegin(GL_LINE_LOOP);
    glVertex3f(1.0,1.0,1.0);
    glVertex3f(1.0,-1.0,1.0);
    glVertex3f(-1.0,-1.0,1.0);
    glVertex3f(-1.0,1.0,1.0);
    glEnd();
    glBegin(GL_LINE_LOOP);
    glVertex3f(1.0,1.0,-1.0);
    glVertex3f(1.0,-1.0,-1.0);
    glVertex3f(-1.0,-1.0,-1.0);
    glVertex3f(-1.0,1.0,-1.0);
    glEnd();
    glBegin(GL_LINES);
    glVertex3f(1.0,1.0,1.0);
    glVertex3f(1.0,1.0,-1.0);
    glEnd();
    glBegin(GL_LINES);
    glVertex3f(1.0,-1.0,1.0);
    glVertex3f(1.0,-1.0,-1.0);
    glEnd();
    glBegin(GL_LINES);
    glVertex3f(-1.0,-1.0,1.0);
    glVertex3f(-1.0,-1.0,-1.0);
    glEnd();
    glBegin(GL_LINES);
    glVertex3f(-1.0,1.0,1.0);
    glVertex3f(-1.0,1.0,-1.0);
    glEnd();
    glPopMatrix();
  }

    glPushMatrix();
    /* axes size in mm... */
    glScalef(50.0,50.0,50.0);

  if (axesmode>1) {
    /* axes */
    glColor3f(0.3,0.8,1.0); /* cyan */
    glBegin(GL_LINE_STRIP);
    glVertex3f(1.0,0.0,0.0);
    glVertex3f(0.0,0.0,0.0);
    glVertex3f(0.0,1.0,0.0);
    glEnd();
    glBegin(GL_LINES);
    glVertex3f(0.0,0.0,0.0);
    glVertex3f(0.0,0.0,1.0);
    glEnd();
  }

  if (axesmode>2) {
    /* xyz */
    glColor3f(0.3,1.0,1.0); /* cyan */
    glRasterPos3f(1.1,0.0,0.0);
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'x');
    glRasterPos3f(0.0,1.1,0.0);
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'y');
    glRasterPos3f(0.0,0.0,1.1);
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'z');
  }

    glPopMatrix();
}


/* ---------------------------------------------------------- */
void texto(GLshort x, GLshort y, char *text)
{
  char *p;

  glRasterPos2s(x,y);
  for (p = text; *p; p++)
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, *p);
}

void fancytexto(GLshort x, GLshort y, char *text)
{
  char *p;

  glRasterPos2s(x,y);
  for (p = text; *p; p++)
    glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, *p);
}


/* -----------------------------------------------------------------------*/
void decorate()
/* 2D text plotting + pixel image */
{
  char str[100],slice[]="XYZ-",tiss[3];
  int i,Imin;

  /* 2d projection */
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0.0, width-0.5, 0.0, height-0.5);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  if (helptoggle) {
    /* display help */
    glColor3f(1.0,0.8,0.2);
    for (i = 0; helptext[i]; i++)
      fancytexto(16, height - 50 - i * 24, helptext[i]);
  }

  /* output params on screen */

  glColor3f(1.0,0,1.0);
  /* print correct offset for ip... */
  if (planemode==0)
    Imin = sys.Ixmin;
  else
    Imin = (planemode==1) ? sys.Iymin : sys.Izmin;
  /* use 1-indexing, like Matlab (but not C)... */
  sprintf(str," ip=%d  it=%d",ip+Imin+1,it);
  texto(0, height-20, str);

  if (tissueshow==sys.Ntissue+1)
    sprintf(tiss,"x ");
  else if (tissueshow==sys.Ntissue+2)
    sprintf(tiss,"y ");
  else if (tissueshow==sys.Ntissue+3)
    sprintf(tiss,"z ");
  else if (tissueshow==sys.Ntissue+4)
    sprintf(tiss,"- ");
  else
    sprintf(tiss,"%d ",tissueshow);

  glColor3f(0,1.0,0);
  sprintf(str," disp:%d slice:%c tiss:%s  zsc=%.6g",\
	  displaymode,slice[planemode],tiss,zscale);
  texto(0, 8, str);
}


/* ----------------------------- Show Tissue -------------------------- */

/* tissue plane's border box colour... */
#define TPLANE_BORDER_R 0.0
#define TPLANE_BORDER_G 1.0
#define TPLANE_BORDER_B 0.0

void showtissue(TISSUE t)
{
  int i,j,k,x,y,z;
  double a;

  if (t <= sys.Ntissue) {
    /* use crude volume-rendering via points */
    /* glDisable(GL_POINT_SMOOTH); */
    glBegin(GL_POINTS);
    glColor3f(1.0,1.0,0.0);
    for (i=0;i<nTx;++i)
      for (j=0;j<nTy;++j)
	for (k=0;k<nTz;++k) {
	  if (tistyp[i][j][k]==t)
	    glVertex3f((float)i,(float)j,(float)k);
	}
    glEnd();
    /* glEnable(GL_POINT_SMOOTH); */
  } else if (t == sys.Ntissue+1) {
    /* use plane slices, with colour scale set by zscale */
    /* -------------------- normal to x ---------- */
    tp = max(min(tp,nTx-1),0);
    x = tp;
    for (y=0; y<nTy-1; ++y) {
      glBegin(GL_QUAD_STRIP);
      for (z=0; z<nTz; ++z) {
	a = zscale * (double)tistyp[x][y][z];
	glColor3f(rc(a),gc(a),bc(a));
	glVertex3f((float)x,(float)y,(float)z);
	a = zscale * (double)tistyp[x][y+1][z];
	glColor3f(rc(a),gc(a),bc(a));
	glVertex3f((float)x,(float)(y+1),(float)z);
      }
      glEnd();
      glBegin(GL_LINE_LOOP);
      glColor3f(TPLANE_BORDER_R,TPLANE_BORDER_G,TPLANE_BORDER_B);
      glVertex3f((float)x,0.0,0.0);
      glVertex3f((float)x,0.0,(float)(nTz-1));
      glVertex3f((float)x,(float)(nTy-1),(float)(nTz-1));
      glVertex3f((float)x,(float)(nTy-1),0.0);
      glEnd();
    }
  } else if (t == sys.Ntissue+2) {
    /* -------------------- normal to y ---------- */
    tp = max(min(tp,nTy-1),0);
    y = tp;
    for (z=0; z<nTz-1; ++z) {
      glBegin(GL_QUAD_STRIP);
      for (x=0; x<nTx; ++x) {
	a = zscale * (double)tistyp[x][y][z];
	glColor3f(rc(a),gc(a),bc(a));
	glVertex3f((float)x,(float)y,(float)z);
	a = zscale * (double)tistyp[x][y][z+1];
	glColor3f(rc(a),gc(a),bc(a));
	glVertex3f((float)x,(float)y,(float)(z+1));
      }
      glEnd();
      glBegin(GL_LINE_LOOP);
      glColor3f(TPLANE_BORDER_R,TPLANE_BORDER_G,TPLANE_BORDER_B);
      glVertex3f((float)(nTx-1),(float)y,0.0);
      glVertex3f((float)(nTx-1),(float)y,(float)(nTz-1));
      glVertex3f(0.0,(float)y,(float)(nTz-1));
      glVertex3f(0.0,(float)y,0.0);
      glEnd();
    }
  } else if (t == sys.Ntissue+3) {
    /* -------------------- normal to z ---------- */
    tp = max(min(tp,nTz-1),0);
    z = tp;
    for (x=0; x<nTx-1; ++x) {
      glBegin(GL_QUAD_STRIP);
      for (y=0; y<nTy; ++y) {
	a = zscale * (double)tistyp[x][y][z];
	glColor3f(rc(a),gc(a),bc(a));
	glVertex3f((float)x,(float)y,(float)z);
	a = zscale * (double)tistyp[x+1][y][z];
	glColor3f(rc(a),gc(a),bc(a));
	glVertex3f((float)(x+1),(float)y,(float)z);
      }
      glEnd();
      glBegin(GL_LINE_LOOP);
      glColor3f(TPLANE_BORDER_R,TPLANE_BORDER_G,TPLANE_BORDER_B);
      glVertex3f(0.0,0.0,(float)z);
      glVertex3f(0.0,(float)(nTy-1),(float)z);
      glVertex3f((float)(nTx-1),(float)(nTy-1),(float)z);
      glVertex3f((float)(nTx-1),0.0,(float)z);
      glEnd();
    }
    
  }
}

/* ----------------------------- Show S & D -------------------------- */

#define SOURCE_LEN 10

void showSD()
{
  int d,s;
  char text[10], *p;

  /* sources */
  glColor3f(1.0,1.0,1.0);
  for (s=0;s<sys.nSrcs;++s) {
    glBegin(GL_LINES);
    glVertex3f(sys.srcLoc[s][0],sys.srcLoc[s][1],sys.srcLoc[s][2]);
    glVertex3f(sys.srcLoc[s][0]+SOURCE_LEN*sys.srcDir[s][0],\
	       sys.srcLoc[s][1]+SOURCE_LEN*sys.srcDir[s][1],\
	       sys.srcLoc[s][2]+SOURCE_LEN*sys.srcDir[s][2]);
    glEnd();
    glRasterPos3f(sys.srcLoc[s][0],sys.srcLoc[s][1],sys.srcLoc[s][2]);
    sprintf(text,"%d",s+1); /* numbering of src reflects matlab */
    for (p = text; *p; p++)
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, *p);
  }
  
  /* detectors */
  glColor3f(1.0,0.0,1.0); 
  for (d=0;d<sys.nDets;++d) {
    glBegin(GL_LINES);
    glVertex3f(sys.detLoc[d][0],sys.detLoc[d][1],sys.detLoc[d][2]);
    glVertex3f(sys.detLoc[d][0]+SOURCE_LEN*sys.detDir[d][0],\
	       sys.detLoc[d][1]+SOURCE_LEN*sys.detDir[d][1],\
	       sys.detLoc[d][2]+SOURCE_LEN*sys.detDir[d][2]);
    glEnd();
    glRasterPos3f(sys.detLoc[d][0],sys.detLoc[d][1],sys.detLoc[d][2]);
    sprintf(text,"%d",d+1); /* numbering of det reflects matlab */
    for (p = text; *p; p++)
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, *p);
  }
}


/* =============================== DISPLAY ============================= */

   /* colour scale... not added wrap-around of out-of-[0,1] values */
float rc(float a)
{
  float b = a; /* - floorf(a); */
  return min(1.0,max(0.0,5*b-1.0));
}
float gc(float a)
{
  float b = a; /* - floorf(a); */
  return min(1.0,max(0.0,5*b-3.0));
}
float bc(float a)
{
  float b = a; /* - floorf(a);*/
   if (b<0.2)
    return 5*b;
  else if (b<0.8)
    return min(1.0,max(0.0,3.0-5*b));
  else
    return 5*b - 4.0;
}


/* --------- transformation ----------- */
#define LOG_BOTTOM 1e-30

double trans(double f, int *flag)
/* returns transformed f, and sets flag if log bottomed-out.
 * log changed to have const 12 orders of magnitude span, 3/23/02
 */
{
  if (transmode==0) {
    *flag = 0;
    return f;
  } else if (transmode==1) {
    *flag = fabs(f) < LOG_BOTTOM;
    return 1.0 + log10( max(LOG_BOTTOM, fabs(f)) ) / 12;
  }
}

/* wire mesh colour... */
#define WIRE_R 0.7
#define WIRE_G 1.0
#define WIRE_B 0.4

void set_col(int flag)
     /* sends appropriate wire-mesh vertex colour to pipe, given flag. */
{
  if (!flag)
    glColor3f(WIRE_R,WIRE_G,WIRE_B);
  else
    glColor3f(0.0,0.0,0.0);
}

/* plane's border box colour... */
#define PLANE_BORDER_R 0.8
#define PLANE_BORDER_G 0.2
#define PLANE_BORDER_B 0.0

/* wire z-scale hack */
#define WIRE_SC 30

#define BACK_CLIP 1000.0
#define FRONT_CLIP 1.0

#define N_DISP_MODES 3
#define N_PLANE_MODES 4

void display(void)
{
  int x,y,z,voxn, flag;
  float a;
  double *f;
  char text[100],*p;

  /* update model spin */
  th += thdot;
  th = (th > 0.0) ? ((th < 2*PI) ? th : th-2*PI) : th+2*PI;
  ph += phdot;
  ph = (ph > 0.0) ? ((ph < 2*PI) ? ph : ph-2*PI) : ph+2*PI;

  /* clear window */
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  /* 3d projection */
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(atan(iris/2)*360.0/PI, width/(float)height,FRONT_CLIP,\
		 BACK_CLIP);

  /* eye location and angle */
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslatef(xof,yof,-dist);
  glRotatef(-rh*180.0/PI, 0.0,1.0,0.0);
  glRotatef(-ph*180.0/PI, 1.0,0.0,0.0);
  glRotatef(-th*180.0/PI, 0.0,0.0,1.0);

  /* set up antialiased lines if needed... */
  if (smoothmode) {
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_LINE_SMOOTH);
    glLineWidth(1.0);
  } else {
    glDisable(GL_BLEND);
    glDisable(GL_LINE_SMOOTH);
    glLineWidth(1.0);
  } 


 
  axes();

  glScalef(sys.dl,sys.dl,sys.dl);
  glTranslatef(-(nx-1)/2,-(ny-1)/2,-(nz-1)/2);

  /* sources, detectors, tissuetypes need to be relative to total region... */
  glPushMatrix();
  glTranslatef(1/2-ox, 1/2-oy, 1/2-oz);
  if (tissueshow<=sys.Ntissue+3)
     showtissue(tissueshow);
  glScalef(1/sys.dl,1/sys.dl,1/sys.dl);
  if (dettoggle)
    showSD();
  glPopMatrix();
  

  /* point to re part of 2pt data... */
  f = II;

  if (displaymode!=1) {
    /* ........................... colored planes of quads ............. */
    if (planemode==0) {
      /* -------------------- normal to x ---------- */
      x = ip;
      for (y=0; y<(ny-1); ++y) {
	glBegin(GL_QUAD_STRIP);
	for (z=0; z<nz; ++z) {
	  voxn = y + ny*(x + nx*z);
	  a = trans(zscale * f[voxn + nv*it],&flag);
	  glColor3f(rc(a),gc(a),bc(a));
	  glVertex3f((float)x,(float)y,(float)z);
	  a = trans(zscale * f[voxn + nv*it + 1],&flag);
	  glColor3f(rc(a),gc(a),bc(a));
	  glVertex3f((float)x,(float)(y+1),(float)z);
	}
	glEnd();
	glBegin(GL_LINE_LOOP);
	glColor3f(PLANE_BORDER_R,PLANE_BORDER_G,PLANE_BORDER_B);
	glVertex3f((float)x,0.0,0.0);
	glVertex3f((float)x,0.0,(float)(nz-1));
	glVertex3f((float)x,(float)(ny-1),(float)(nz-1));
	glVertex3f((float)x,(float)(ny-1),0.0);
	glEnd();
      }
    } else if (planemode==1) {
      /* -------------------- normal to y ---------- */
      y = ip;
      for (z=0; z<(nz-1); ++z) {
	glBegin(GL_QUAD_STRIP);
	for (x=0; x<nx; ++x) {
	  voxn = y + ny*(x + nx*z);
	  a = trans(zscale * f[voxn + nv*it],&flag);
	  glColor3f(rc(a),gc(a),bc(a));
	  glVertex3f((float)x,(float)y,(float)z);
	  a = trans(zscale * f[voxn + nv*it + ny*nx],&flag);
	  glColor3f(rc(a),gc(a),bc(a));
	  glVertex3f((float)x,(float)y,(float)(z+1));
	}
	glEnd();
	glBegin(GL_LINE_LOOP);
	glColor3f(PLANE_BORDER_R,PLANE_BORDER_G,PLANE_BORDER_B);
	glVertex3f((float)(nx-1),(float)y,0.0);
	glVertex3f((float)(nx-1),(float)y,(float)(nz-1));
	glVertex3f(0.0,(float)y,(float)(nz-1));
	glVertex3f(0.0,(float)y,0.0);
	glEnd();
      }
    } else if (planemode==2) {
      /* -------------------- normal to z ---------- */
      z = ip;
      for (x=0; x<(nx-1); ++x) {
	glBegin(GL_QUAD_STRIP);
	for (y=0; y<ny; ++y) {
	  voxn = y + ny*(x + nx*z);
	  a = trans(zscale * f[voxn + nv*it],&flag);
	  glColor3f(rc(a),gc(a),bc(a));
	  glVertex3f((float)x,(float)y,(float)z);
	  a = trans(zscale * f[voxn + nv*it + ny],&flag);
	  glColor3f(rc(a),gc(a),bc(a));
	  glVertex3f((float)(x+1),(float)y,(float)z);
	}
	glEnd();
	glBegin(GL_LINE_LOOP);
	glColor3f(PLANE_BORDER_R,PLANE_BORDER_G,PLANE_BORDER_B);
	glVertex3f(0.0,0.0,(float)z);
	glVertex3f(0.0,(float)(ny-1),(float)z);
	glVertex3f((float)(nx-1),(float)(ny-1),(float)z);
	glVertex3f((float)(nx-1),0.0,(float)z);
	glEnd();
      }
    }

  }
  if (displaymode>0) {
    /* ........................... normally-pointing z-mesh ............. */
    if (planemode==0) {
      /* -------------------- normal to x ---------- */
      x = ip;
      for (y=0; y<ny; ++y) {
	glBegin(GL_LINE_STRIP);
	for (z=0; z<nz; ++z) {
	  voxn = y + ny*(x + nx*z);
	  a = WIRE_SC* zscale * trans(f[voxn + nv*it],&flag);
	  set_col(flag);
	  glVertex3f((float)(x+a),(float)y,(float)z);
	}
	glEnd();
      }
      for (z=0; z<nz; ++z) {
	glBegin(GL_LINE_STRIP);
	for (y=0; y<ny; ++y) {
	  voxn = y + ny*(x + nx*z);
	  a = WIRE_SC*zscale * trans(f[voxn + nv*it],&flag);
	  set_col(flag);
	  glVertex3f((float)(x+a),(float)y,(float)z);
	}
	glEnd();
      }
    } else if (planemode==1) {
      /* -------------------- normal to y ---------- */
      y = ip;
      for (x=0; x<nx; ++x) {
	glBegin(GL_LINE_STRIP);
	for (z=0; z<nz; ++z) {
	  voxn = y + ny*(x + nx*z);
	  a = WIRE_SC* zscale * trans(f[voxn + nv*it],&flag);
	  set_col(flag);
	  glVertex3f((float)x,(float)(y+a),(float)z);
	}
	glEnd();
      }
      for (z=0; z<nz; ++z) {
	glBegin(GL_LINE_STRIP);
	for (x=0; x<nx; ++x) {
	  voxn = y + ny*(x + nx*z);
	  a = WIRE_SC*zscale * trans(f[voxn + nv*it],&flag);
	  set_col(flag);
	  glVertex3f((float)x,(float)(y+a),(float)z);
	}
	glEnd();
      }
    } else if (planemode==2) {
      /* -------------------- normal to z ---------- */
      z = ip;
      for (x=0; x<nx; ++x) {
	glBegin(GL_LINE_STRIP);
	for (y=0; y<ny; ++y) {
	  voxn = y + ny*(x + nx*z);
	  a = WIRE_SC* zscale * trans(f[voxn + nv*it],&flag);
	  set_col(flag);
	  glVertex3f((float)x,(float)y,(float)(z+a));
	}
	glEnd();
      }
      for (y=0; y<ny; ++y) {
	glBegin(GL_LINE_STRIP);
	for (x=0; x<nx; ++x) {
	  voxn = y + ny*(x + nx*z);
	  a = WIRE_SC*zscale * trans(f[voxn + nv*it],&flag);
	  set_col(flag);
	  glVertex3f((float)x,(float)y,(float)(z+a));
	}
	glEnd();
      }
    }

  }


  decorate();

  glutSwapBuffers();
}



/* ------------------------------- MENU options ---------------------------*/
void menu(int value)
{
  switch (value) {
  case 1:
    setviewcoords();
    break;
  case 9:
    helptoggle = 1 - helptoggle;
    break;
  case 10:
    exit(0);
    break;
  }
}


/* ----------------------------- KEYPRESS options ---------------------------*/
void keyfunc( unsigned char key, int x, int y)
{
  int i,j;
  double maxf;

  switch (key) {
  case 'C':
  case 'c':
    helptoggle = 0;
    break;
  case 'I':
  case 'i':
    printf("some output\n");
    break;
  case ' ':
    ++displaymode;
    if (displaymode==N_DISP_MODES) {
      displaymode = 0;
      smoothmode = 1-smoothmode;
    }
    break;
  case 'A':
  case 'a':
    ++axesmode;
    if (axesmode==N_AXES_MODES)
      axesmode = 0;
    break;
  case 'S':
  case 's':
    ++planemode;
    if (planemode==N_PLANE_MODES)
      planemode = 0;
    /* clip ip to allowed max... */
    if (planemode==0)
      i = nx;
    else if (planemode==1)
      i = ny;
    else
      i = nz;
    if (ip>=i)
      ip = i-1;
    break;
  case 'T':
  case 't':
    ++tissueshow;
    /* showing Ntissue+4 is showing none, after that cycle to 0. */
    if (tissueshow > sys.Ntissue+4)
      tissueshow = 0;
    break;
  case 'D':
  case 'd':
    dettoggle = 1-dettoggle;
    break;
  case 'L':
  case 'l':
    transmode = 1-transmode;
    break;
  case 'm':
  case 'M':
    /* make zscale sensible... */
    maxf = 0.0;
    for (i=0; i<nv; ++i)
      maxf = max(maxf, fabs(II[i + nv*it]));
    zscale = 1/maxf;
    break;
  case 'n':
  case 'N':
    /* step forward in time */
    ++it;
    if (it==nt)
      it = 0;
    break;
  case 'Z':
  case 'z':
    /* step fluence plane back */
    --ip;
    if (ip<0)
      ip = 0;
    break;
  case 'X':
  case 'x':
    /* step fluence plane forward, bump into end */
    ++ip;
    if (planemode==0)
      i = nx;
    else if (planemode==1)
      i = ny;
    else
      i = nz;
    if (ip>=i)
      ip = i-1;
    break;
  case '[':
  case '{':
    /* step tissue plane back : range checks done in showtissue */
    --tp;
    break;
  case ']':
  case '}':
    /* step tissue plane forward : range checks done in showtissue */
    ++tp;
    break;
  case 'Q':
  case 'q':
    exit(0);
    break;
  }
  /* chances are you'll need to redisplay... */
  glutPostRedisplay();
}

void specfunc( unsigned char key, int x, int y)
{

  switch (key) {
  case GLUT_KEY_LEFT:
    thdot += 0.001;
    break;
  case GLUT_KEY_RIGHT:
    thdot -= 0.001;
    break;
  case GLUT_KEY_DOWN:
    phdot -= 0.0005;
    break;
  case GLUT_KEY_UP:
    phdot += 0.0005;
    break;
  }
  if (thdot!=0.0 || phdot!=0.0) {
    freely_spinning = 1;
    glutIdleFunc(animate);
    glutPostRedisplay();
  }
}


/* ---------------------------------motion---------------------------------*/
void motion(int mx,int my)
     /* deals with mouse motion mx my are relative to window origin */
{
  int d;
  /* change origin from lower left (OpenGL, IrisGL) to upper left (GLUT) */
  my = height - my;

  if (dragging) {
    thdot = -(mx-omx)*2.5/(float)width;
    phdot = (my-omy)*2.5/(float)width;
    increment();
    /* trigger a single refresh cycle... */
    glutIdleFunc(animate);
  }
  if (scaling) {
    dist += (mx-omx)*250.0/(float)width;
    /* dist = (dist > -20.0) ? ((dist < 100.0) ? dist : 100.0) : -20.0;
    // iris += (my-omy)/400.0;
    // iris = (iris > 0.05)?((iris < 5.0) ? iris : 5.0) : 0.05; */
    zscale *= exp((my-omy)*3/(float)width);
    /* trigger a single refresh cycle... */
    glutIdleFunc(animate);
  }

  /* save test result of whether still 'moving' for mouse_button... */
  mouse_still_moving = (abs(mx-omx)>1 || abs(my-omy)>1);
  
  /* restart mouse relative counting here... */
  omx = mx;
  omy = my;
}



/* ---------------------------------------------------------------------- */
void mouse(int button, int state, int x, int y)
{
  y = height - y;

  if (button == GLUT_LEFT_BUTTON) {
    if (state == GLUT_DOWN) {
      freely_spinning = 0;
      dragging = 1;
      omx = x;
      omy = y;
    } else if (state == GLUT_UP) {
      dragging = 0;
      if (mouse_still_moving) {
	/* let freely spin if non-zero release velocity of mouse... */
	freely_spinning = 1;
      } else {
	/* zero the velocities... */
	thdot = phdot = 0.0;
      }
    }
  }
  if (button == GLUT_MIDDLE_BUTTON) {
    if (state == GLUT_DOWN) {
      scaling = 1;
	omx = x;
	omy = y;
    } else if (state == GLUT_UP) {
      scaling = 0;
    }
  }

  if (button == GLUT_LEFT_BUTTON || button == GLUT_MIDDLE_BUTTON)
    glutPostRedisplay();

  /* cursor appearance... */
  if (dragging || scaling)
    glutSetCursor(GLUT_CURSOR_INFO);
  else
    glutSetCursor(GLUT_CURSOR_CROSSHAIR);
}


/* ----------------------------------------------------------------------*/
void reshape(int w, int h)
{
  glViewport(0, 0, w, h);
  width = w;
  height = h;
}





/* ===================== Jon's format load ============================== */

/*********************************************************
             OPEN AND READ THE INPUT FILE

  Returns non-zero on error
    1 -> File error (can't open file, unexpected EOF, etc.)
    2 -> Invalid input parameter
    3 -> System error (out of memory, etc.)

*********************************************************/

int loadconfig(struct Config *p, int argc, char *argv[])
{
  char filenm[FILENMLEN];
  FILE *fp;
  int i, n;

  /* Initialize the memory */
  memset(p, 0, sizeof(struct Config));

  /* Set some constants */
  for (i = 0; i < MAXTISSUE; i++)
    {
      p->tmus[i] = MUS_AIR;
      p->tmua[i] = MUA_AIR;
      p->tg[i]   =   G_AIR;
      p->tn[i]   = IDX_AIR;
    }

  /* Open input file */
  snprintf(filenm, FILENMLEN, "%s.cfg", argv[1]);

  if ((fp = fopen(filenm, "rt")) == NULL) 
    {
      fprintf(stderr, 
              "input_file \"%s\" not found, assuming you meant \"%s\"\n",
              filenm, argv[1]);

      if ((fp = fopen(argv[1], "rt")) == NULL) 
        {
          fprintf(stderr, "usage: tmcv input_file (.cfg assumed)\n");
          fprintf(stderr, "input_file = %s does not exist.\n", filenm);
          return 1;
        } 
      else 
        {
          /* chop off the extension */
          char *p = rindex(argv[1], '.');

          if (p != NULL)
            *p = '\0';
        }
    }

  /* READ THE INPUT FILE */
  n = parseconfig(p, fp);

  /* CLOSE FILE AND RETURN */
  fclose(fp);
  return n;
}

/*********************************************************
        OPEN AND READ THE SEGMENTATION FILE

  Returns non-zero on error
    1 -> File error (can't open file, unexpected EOF, etc.)
    2 -> Invalid input parameter
    3 -> System error (out of memory, etc.)

*********************************************************/

int loadsegments(const struct Config *p, TISSUE ****tt)
{
  TISSUE ***tissueType;
  FILE *fp;
  int i, j, k;

  assert(p != NULL);
  assert(tt != NULL);

  if ((fp = fopen(p->segFile, "rb")) == NULL) 
    {
      fprintf(stderr,
              "ERROR: THe binary image file %s was not found!\n",
              p->segFile);
      return 1;
    }

  rewind(fp);			  /* Check actual vs expected file sizes */
  fseek(fp, 0, SEEK_END);

  if (ftell(fp) != p->nxstep * p->nystep * p->nzstep * sizeof(TISSUE)) 
    {
      fprintf(stderr, 
              "Size of segmentation file != nxstep*nystep*nzstep elements\n");
      return 1;
    } 
  else
    rewind(fp);

  if ((*tt = (TISSUE ***)malloc(p->nxstep * sizeof(TISSUE **))) == NULL) 
    {
      fprintf(stderr, "malloc() failed, can't allocate %d bytes\n",
              p->nxstep * sizeof(TISSUE **));

      return 3;
    }

  tissueType = *tt;

  for (i = 0; i < p->nxstep; i++) 
    {
      tissueType[i] = (TISSUE **)malloc(p->nystep * sizeof(TISSUE *));

      if (tissueType[i] == NULL) 
        {
          fprintf(stderr, "malloc() failed, can't allocate %d bytes\n",
                  p->nystep * sizeof(TISSUE *));
          return 3;
        } 
      else
        for (j = 0; j < p->nystep; j++) 
          {
            tissueType[i][j] = (TISSUE *)malloc(p->nzstep * sizeof(TISSUE));
            
            if (tissueType[i][j] == NULL) 
              {
                fprintf(stderr,
                        "malloc() failed, can't allocate %d bytes\n",
                        p->nzstep * sizeof(TISSUE));
                return 3;
              }
          }
    }

  /* Matlab stores data column [Y,X,Z], I use data row
   * [X,Y,Z] first.  Flip the data around as I read it in.
   */

  for (k = 0; k < p->nzstep; k++)
    {
      for (i = 0; i < p->nxstep; i++) /*  i, for Matlab */
        {
          for (j = 0; j < p->nystep; j++) /* j, for Matlab */
            {
              TISSUE s;

              if (fread(&s, sizeof(TISSUE), 1, fp) != 1) 
                { 
                  fprintf(stderr,
                          "Unexpected end of segmentation file %s", 
                          p->segFile);
                  return 1;
                }  
              else 
                tissueType[i][j][k] = s;
            }
        }

      printf(".");
      fflush(stdout);
    }

  printf("\n");

  fclose(fp);

  return 0;
}


/* ------------------------load_2pt----------------------------- */
int load_2pt(char *name)
{
  FILE *fp;
  int i;
  double a,b;

  if ((fp = fopen(name,"r")) == NULL) {
    printf("load_pt2 can't read %s\n",name);
    return 1;
  }

  fread(II, sizeof(double), nv * nt, fp);

  fclose(fp);

#if (ENDIAN)
  /* flip all doubles' byte-ordering */
  for (i=0;i<nv*nt;++i) {
    II[i] = endian_8byte(II + i);
  }
#endif

  return 0;
}




/* ================================================== MAIN =============== */
int main(int argc, char **argv)
{
  int i,j,k,t,voxn;
  char filehead[FILENMLEN], filesuffix[FILENMLEN], name[FILENMLEN], temp[FILENMLEN];
  FILE *fp;


  /* read command line */
  if (argc==2) {
	sscanf(argv[1],"%s",filehead);
  } else if (argc==3) {
	sscanf(argv[1],"%s",filehead);
	sscanf(argv[2],"%s",filesuffix);
  } else {
    printf("Usage:      tmcv filehead [2ptsuffix]\n");
    exit(1);
  }

  /* PARSE THE INPUT FILE (it gets filehead from argv) */
  if ((i = loadconfig(&sys, argc, argv)) != 0)
    exit(i);

  /* READ IN THE SEGMENTED DATA FILE */
  if ((i = loadsegments(&sys, &tistyp)) != 0)
    exit(i);

  /* own easy-to-use global versions of Jon variables... */
  nTx = sys.nxstep;
  nTy = sys.nystep;
  nTz = sys.nzstep;
  nTv = nTx*nTy*nTz;
  nx = sys.nIxstep;
  ny = sys.nIystep;
  nz = sys.nIzstep;
  nv = nx*ny*nz;
  nt = sys.nTstep;
  ox = sys.Ixmin;
  oy = sys.Iymin;
  oz = sys.Izmin;
  printf("full %d * %d * %d = %d voxels of size (%gmm)^3.\n",\
	 nTx,nTy,nTz,nTv,sys.dl);
  printf("region of interest %d * %d * %d = %d voxels, %d timesteps.\n",\
	 nx,ny,nz,nv,nt);
  printf("ROI offset (%d,%d,%d) voxels.\n",\
	 ox,oy,oz);

  /* 2pt filename */
  if (argc==2)
    sprintf(name,"%s.2pt",filehead);
  else
    sprintf(name,"%s.2pt.%s",filehead,filesuffix);
  if ((fp = fopen(name,"r")) == NULL) { /* test openability of file */
    printf("No fluence data file %s found.\n", name);
    nt = 1;    /* choose dummy number of timesteps for 2pt data */
    /* allocate 2pt */
    II = (double *)malloc(nv*nt*sizeof(double));
    if (II==NULL) {
      printf("can't allocate %d bytes.\n",nv*nt*sizeof(double));
      return 1;
    }
    /* zero the 2pt array */
    for (i=0;i<nx;++i)
      for (j=0;j<ny;++j)
	for (k=0;k<nz;++k) {
	  voxn = j + ny*(i + nx*k);
	  for (t=0;t<nt;++t)
	    II[voxn + nv*t] = 0.0;
	  /* (double)tistyp[i+ox][j+oy][k+oz]; */
	}
  } else {
    fclose(fp);  /* close so load_2pt can reopne it */
    /* allocate 2pt */
    II = (double *)malloc(nv*nt*sizeof(double));
    if (II==NULL) {
      printf("can't allocate %d bytes.\n",nv*nt*sizeof(double));
      return 1;
    }
    if ((i = load_2pt(name)) != 0)
      printf("Problem reading fluence data file %s\n", name);
  }

  /* initialize mouse-adjustable params */
  v = 1.0; 
  w = 1.0;   

  /* graphics setup:--------------------------------------------- */

  setviewcoords();

  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_ALPHA | GLUT_DEPTH);
  glutInitWindowSize(700, 600);
  sprintf(temp,"tmcv : %s",name);
  glutCreateWindow(temp);

  width  = glutGet(GLUT_WINDOW_WIDTH);
  height = glutGet(GLUT_WINDOW_HEIGHT);
  xpos   = glutGet(GLUT_WINDOW_X);
  ypos   = glutGet(GLUT_WINDOW_Y);

  /* z buffering... */
  glEnable(GL_DEPTH_TEST);

  glutMotionFunc(motion);
  glutMouseFunc(mouse);
  glutReshapeFunc(reshape);
  glutDisplayFunc(display);
  glutKeyboardFunc(keyfunc);
  glutSpecialFunc(specfunc);
  /* glutMenuStateFunc(menustat); */

  glutCreateMenu(menu);
  glutAddMenuEntry("reset view", 1);
  glutAddMenuEntry("help", 9);
  glutAddMenuEntry("quit", 10);
  glutAttachMenu(GLUT_RIGHT_BUTTON);

  /* pass control to GLUT */
  glutMainLoop();

  return 0;             /* ANSI C requires main to return int */
}

/* end */

